# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *


# Structure for a LogicalOperator block
# operators handled: NOT, AND, OR 
class SimulinkLogic(SimulinkBlock):
	index = 0
	operators_gen = {"AND": " \<otimes> ", "OR": " \<oplus> "}
	negation_gen = " \<ominus> "

	operators_logic = {"AND": " \<and> ", "OR": " \<or> "}
	negation_logic = " \<not> "

	def update_arities(self):
		self.set_outport_arity(0,1)
		return 0
	
	def update_const_params(self, opt_const, opt_generic):
		if opt_generic:
			p = "vl_" + str(SimulinkLogic.index)
			self.set_params(p)
			SimulinkLogic.index += 1
	
	def get_def_params(self, opt_fun, opt_generic):
		params = self.get_params()
		if params:
			return "(%s::'a::simulink)" % params
		else:
			return params

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
			
		# get logical operator
		op = get_property(self.xml, 'Operator')
		var = [x.get_var_name() for x in self.inport]
		expr = ["(%s \<noteq> 0)" % x.get_var_name() for x in self.inport]
		
		# build logical property
		if op in SimulinkLogic.operators_gen.keys():
			pred = SimulinkLogic.operators_logic[op].join(expr)
			if opt_generic:
				pred = "if %s then 1::'a else 0" %pred
		else:
			pred = "%s = 0" % self.inport[0].get_var_name()
			#pred = SimulinkLogic.negation_logic + expr[0]
			if opt_generic:
				pred = "if %s then 1::'a else 0" %pred

		return iFunc(var, pred, opt_fun)


# Structure for a RelationalOperator block 
# operators handled: ==, ~=, <, <=, >, >=
# operators not handled: isInf, isNaN, isFinite
class SimulinkRelationalOperator(SimulinkBlock):
	index = 0
	operators_gen = {"==":"is_eq_num ", "~=":"is_neq_num ", "<":"is_less_num ", "<=":"is_less_eq_num ", ">":"is_gt_num ", ">=":"is_ge_num "}
	operators_logic = {"==":" = ", "~=":"\<noteq>", "<":" < ", "<=":" <= ", ">":" > ", ">=":" >= "}
	
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		# set block's parameter
		self.op = get_property(self.xml, "Operator")
		if self.op not in self.operators_gen.keys():
			raise Exception('\nError:\n  The relational operator of the block ' + self.get_qualified_name() + ' is unknown.\n  The only hanlded operators are ==, ~=, <, <=, >, >=.\n')

	def update_const_params(self, opt_const, opt_generic):
		p = "vr_" + str(SimulinkRelationalOperator.index)
		self.set_params(p)
		SimulinkRelationalOperator.index += 1 

	def get_def_params(self, opt_fun, opt_generic):
		#params = self.get_params_list()
		params = self.get_params()
		if params:
			return "(%s::'a::simulink)" % params
		#if params:
		#	return "(%s::'a::MyNumber) (%s::'b::MyNumber)" % (params[0], params[1])
		else:
			return params

		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		invars = [p.get_var_name() for p in self.inport]
		
		lhs = "(%s::'a)" % invars[0] if self.get_params() else invars[0]
		
		if opt_generic:
			# prefix operator
			operator = self.operators_gen[self.op]
			expr = "%s %s %s" % (operator , invars[0], invars[1])
			expr = "((%s)::'a)" % expr if self.get_params() else expr
		else:
			operator = self.operators_logic[self.op]
			# infix operator
			expr = "%s %s %s" % (invars[0], operator, invars[1])
		
		return iFunc(invars, expr, opt_fun)

