# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *
from utilities.simulink_globals import *


# Structure for a DataTypeConversion block
class SimulinkDataTypeConversion(SimulinkBlock):
	index = 0
	#ConvTo = {"boolean" : "ToBool", "double" : "ToReal"}
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.out_type = get_property(self.xml, 'OutDataTypeStr')
		self.skip = False
	
	def update_const_params(self, opt_const, opt_generic):
		p = "vco_" + str(SimulinkDataTypeConversion.index)
		#SimulinkDataTypeConversion.index += 1 
		#q = "vco_" + str(SimulinkDataTypeConversion.index)
		#self.set_params([p,q])
		self.set_params(p)
		SimulinkDataTypeConversion.index += 1
	
		if opt_generic:
			p = "vco_" + str(SimulinkDataTypeConversion.index)
			self.set_params(p)
			SimulinkDataTypeConversion.index += 1


	def get_def_params(self, opt_fun, opt_generic):
		params = self.get_params_list()
		#params = self.get_params()
		#if params:
		#	return "(%s::'a::simulink)" % params
		if params and len(params)==2:
			return "(%s::'a::simulink) (%s::'b::simulink)" % (params[0], params[1])
		elif params and len(params)==1:
			return "(%s::'a::simulink)" % params[0]
		else:
			return self.get_params()
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
		
		inp = self.inport[0].var_name
		
		param = "(%s::'a)" % inp if self.get_params() else inp
		
		#if "Inherit" in self.out_type:  #newly added, please check
		#	return iSkip()
		
		
		if not opt_generic and self.out_type in conv_types.keys():
			expr = "(conversion %s)::%s" % (param, conv_types[self.out_type])
		elif not opt_generic:
			#print "Warning: %s - conversion to \"%s\" will be handled by Isabelle." % (self.get_qualified_name(),self.out_type)
			#expr = "((conversion %s)::'b::MyNumber)" % param
			expr = "conversion %s" % param
		else:
			#expr = "((conversion %s)::'b::MyNumber)" % param
			expr = "(conversion %s)::'b" % param
			
		return iFunc([inp], expr, opt_fun)		
		
# Structure for a Mux block
class SimulinkMux(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		in_arity = get_property(self.xml, "Inputs")		
		if in_arity:
			in_arity = in_arity.replace("[","").replace("]","").strip()
			in_arity = eval(in_arity.replace(" ",","))
			if not isinstance(in_arity,int):
				for i in in_arity:
					if i > 1:
						raise Exception("\nError: only Mux blocks with arity 1 for each inport are considered. Please modify your model before proceeding.\n")
		
	def update_arities(self):
		flag = 0
		# compute arity for output port based on the arities of input ports
		s = 0
		for p in self.inport:
			if p.get_arity() == 0:
				p.set_arity(1)
				flag = 1
			s += p.get_arity()
			
		# update output port arity if needed and flag it
		if s != self.get_outport_arity(0):
			flag = 1
			self.set_outport_arity(0, s)
		return flag

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		return iSkip()

	def split_block(self):
		n = len(self.inport)
		assert(n == len(self.outport))
		skip_blocks = []
		for i in range(n):
			skip_blocks.append(Skip(self.inport[i], self.outport[i]))
		return skip_blocks

class SimulinkSelector(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		index_options = get_property(self.xml, 'IndexOptions')
		if index_options != 'Index vector (dialog)':
			raise Exception('\nError:\n  Only Selector blocks - '+self.get_qualified_name()+' - with index vector (dialog) are handled.\n')
		self.indices = eval(get_property(self.xml, 'Indices').replace(" ",","))
		self.skip = False
			
	def update_arities(self):
		self.set_outport_arity(0,len(self.indices))
		return 0

	def split_block(self):
		skip_blocks = []
		for i in range(len(self.outport)):
			skip_blocks.append(Skip(self.inport[self.indices[i]-1], self.outport[i]))
		return skip_blocks

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		invars = [p.get_var_name() for p in self.inport]
		expr = ", ".join([invars[self.indices[i]-1] for i in range(len(self.indices))]) 
		return iFunc(invars, expr, opt_fun)

# Structure for a Switch block
class SimulinkSwitch(SimulinkBlock):
	def update_arities(self):
		self.set_outport_arity(0,1)
		return 0
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		criteria = get_property(self.xml, 'Criteria')
		if criteria.find('Threshold') != -1 :
			value = get_property(self.xml, 'Threshold')
			expr = criteria.replace('Threshold', value)
		else:
			expr = criteria.replace('~=', '\<noteq>')
			
		if_expr = "If (%s) Then u1 Else u3" % expr
		return iFunc(["u1", "u2", "u3"], if_expr, opt_fun)


# Structure for a MultiPortSwitch block
class SimulinkMultiPortSwitch(SimulinkBlock):
	def __init__(self, xml = None, baseswitch = None, inport = None, outport = None, parent = None, index = None):
		SimulinkBlock.__init__(self, xml, None, None, parent)
		if xml == None:
			# this is a split of the MultiPortSwitch
			# the ports are given in the inport and outport parameters
			self.inport = inport
			self.outport = outport
			self.sid = "%s:%d" % (baseswitch.sid, 0)
			self.type = "%s:%d" % (baseswitch.type, 0)

		self.size = len(self.inport)
		self.arities = None
		
		if (baseswitch is not None) and (index is not None):
				self.name = baseswitch.get_name() + str(index)
		
	def update_arities(self):
		if self.inport[0].get_arity() == 0:
			self.inport[0].set_arity(1)
			
		res = 0
		for p in self.inport:
			if p.get_arity() == 0:
				res = 1
				p.set_arity(1)

		max_arity = max([p.get_arity() for p in self.inport[1:]])
		
		if max_arity != self.outport[0].get_arity():
			self.outport[0].set_arity(max_arity)
			return 1
		return res

	def get_cases(self):
		l = []
		if self.arities == None:
			out_arity = self.outport[0].get_arity()
			x = [p for p in self.inport[1:] if p.get_arity() != 1 and p.get_arity() != out_arity]
			assert(x == []) # check if input ports have the same arity as outport or 1.
			for p in self.inport[1:]:
				if p.get_arity() > 1:
					l.append(mk_tuple_val(p.get_var_list()))
				else:
					l.append(mk_tuple_val([p.get_var_name()] * out_arity))
		else:
			# the case when the ports with arities are split.
			out_size = len(self.outport)
			x = [k for k in self.arities if k != 1 and k != out_size]
			assert(x == []) # check if input ports have the same arity as outport or 1.
			i = 1 # skip control input
			for k in self.arities:
				if k == 1:
					l.append(mk_tuple_val([self.inport[i].get_var_name()] * out_size))
					i = i + 1
				else:
					l.append(mk_tuple_val([p.get_var_name() for p in self.inport[i:i + out_size]]))
					i = i + out_size
		return l
			

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		flat_in_list_vars  = [x.get_var_list() for x in self.inport]
		flat_in_vars = [j for i in flat_in_list_vars for j in i]
		
		assert(self.inport[0].get_arity() == 1)
		
		
		ctrl_var = self.inport[0].get_var_name()
		
		prec = "1 <= %s \<and> %s < %d" % (ctrl_var, ctrl_var, self.size)
		
		cases = self.get_cases()
		assert(self.size == len(cases) + 1 and self.size > 1)
		
		expr = cases[-1]
		for i in range(2, self.size):
			expr = "If %s < %d Then %s Else %s" % (ctrl_var, self.size - i + 1, cases[-i], expr)

		return iFunc(flat_in_vars, expr, opt_fun, prec)
	
	def split_ports(self):
		if self.arities != None:
			raise Exception('\nError:\n  An error has occured while building the internal model.\n The operation split_ports has already been executed.\n')
		self.arities = [p.get_arity() for p in self.inport[1:]]
		SimulinkBlock.split_ports(self)
	
	def split_block(self):
		assert(self.arities != None) # the ports must be split already
		out_size = len(self.outport)
		expanded_inport = []
		i = 1 # skip control input
		for k in self.arities:
			if k == 1:
				expanded_inport.append([self.inport[i]] * out_size)
				i = i + 1
			else:
				expanded_inport.append(self.inport[i:i + out_size])
				i = i + out_size
				
		return [SimulinkMultiPortSwitch(xml = None, baseswitch = self,
			inport = [self.inport[0]] + [p[i] for p in expanded_inport], 
			outport = [self.outport[i]], parent = self.parent, index = i) for i in range(out_size)]


# Structure for a Demux block
# TODO: study better demuxes
class SimulinkDemux(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.skip = False
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		return iSkip()
		
		
		
# Structure for a GoTo block
class SimulinkGoto(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		if get_property(xml, "TagVisibility") != "local":
			raise Exception('\nError:\n  Only Goto blocks - '+self.get_qualified_name()+' - with local visibility are handled.\n')
		self.create_ports(1, 1)
		self.label = get_property(self.xml, "GotoTag")
	
	def update_arities(self):
		flag = 0
		a = self.get_inport_arity(0)
		for p in self.outport:
			if p.get_arity() != a:
				p.set_arity(a)
				flag = 1
		return flag

	def add_ports(self, index):
		n = len(self.outport)
		p = [Port(self.sid, i+1, "out", self.get_inport_arity(0)) for i in range(n, index)]
		self.outport += p
	
		
	def split_block(self):
		n = len(self.inport)
		m = len(self.outport)
					
		k = m / n

		skip_blocks = []
		x = 0
		for j in range(k):
			for i in range(n):
				skip_blocks.append(Skip(self.inport[i], self.outport[x + i]))
			x = x + n
		return skip_blocks


	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		n = len(self.inport)
		m = len(self.outport)
		return iSwitchSplit([self.inport], [self.inport * (m / n)], opt_fun)


# Structure for a From block
class SimulinkFrom(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		if get_property(xml, "TagVisibility") != "local":
			raise Exception('\nError:\n  Only From blocks - '+self.get_qualified_name()+' - with local visibility are handled.\n')
		self.create_ports(1, 1)
		self.label = get_property(self.xml, "GotoTag")
	
	def update_arities(self):
		flag = 0
		self.set_outport_arity(0, 1)
		return flag

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		return iSkip()
