# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *


class SimulinkAssertion(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		inpt = self.inport[0].get_var_name()

		#prec = "%s \<noteq> (0::'a::MyNumber)" %inpt
		prec = "%s" %inpt
		expr = "()"

		return iFunc([inpt], expr, opt_fun, prec)

class SimulinkCheckStaticRange(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.min = get_instance_data_property(xml, "min")
		self.is_min_included = True if get_instance_data_property(xml, "min_included") else False
		self.max = get_instance_data_property(xml, "max")
		self.is_max_included = True if get_instance_data_property(xml, "max_included") else False
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())

		inpt = self.inport[0].get_var_name()

		op = "<=" if self.is_min_included else "<"
		prec1 = "%s %s %s" %(self.min, op, inpt)
		op = "<=" if self.is_max_included else "<"
		prec2 = "%s %s %s" %(inpt, op, self.max)

		prec = " \<and> ".join([prec1, prec2])
		expr = "()"

		return iFunc([inpt], expr, opt_fun, prec)


class SimulinkCheckStaticGap(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.min = get_instance_data_property(xml, "min")
		self.is_min_included = True if get_instance_data_property(xml, "min_included") else False
		self.max = get_instance_data_property(xml, "max")
		self.is_max_included = True if get_instance_data_property(xml, "max_included") else False
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		inpt = self.inport[0].get_var_name()
		
		op = "<=" if self.is_min_included else "<"
		prec1 = "%s %s %s" %(inpt, op, self.min)
		op = "<=" if self.is_max_included else "<"
		prec2 = "%s %s %s" %(self.max, op, inpt)
		
		prec = " \<or> ".join([prec1, prec2])
		expr = "()"
		
		return iFunc([inpt], expr, opt_fun, prec)


class SimulinkCheckStaticLowerBound(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.min = get_instance_data_property(xml, "min")
		self.is_min_included = True if get_instance_data_property(xml, "min_included") else False
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		inpt = self.inport[0].get_var_name()
		
		op = "<=" if self.is_min_included else "<"
		prec = "%s %s %s" %(self.min, op, inpt)
		expr = "()"
		
		return iFunc([inpt], expr, opt_fun, prec)


class SimulinkCheckStaticUpperBound(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.max = get_instance_data_property(xml, "max")
		self.is_max_included = True if get_instance_data_property(xml, "max_included") else False
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		inpt = self.inport[0].get_var_name()
		
		op = "<=" if self.is_max_included else "<"
		prec = "%s %s %s" %(inpt, op, self.max)
		expr = "()"
		
		return iFunc([inpt], expr, opt_fun, prec)


class SimulinkCheckDynamicRange(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())

		max = self.inport[0].get_var_name()
		min = self.inport[1].get_var_name()
		inpt = self.inport[2].get_var_name()

		prec1 = "%s <= %s" %(min, inpt)
		prec2 = "%s <= %s" %(inpt, max)

		prec = " \<and> ".join([prec1, prec2])
		expr = "()"

		return iFunc([max, min, inpt], expr, opt_fun, prec)


class SimulinkCheckDynamicGap(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())

		max = self.inport[0].get_var_name()
		min = self.inport[1].get_var_name()
		inpt = self.inport[2].get_var_name()
		
		prec1 = "%s < %s" %(inpt, min)
		prec2 = "%s < %s" %(max, inpt)
		
		prec = " \<or> ".join([prec1, prec2])
		expr = "()"
		
		return iFunc([max, min, inpt], expr, opt_fun, prec)



class SimulinkCheckDynamicLowerBound(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		min = self.inport[0].get_var_name()
		inpt = self.inport[1].get_var_name()
		
		prec = "%s < %s" %(min, inpt)
		expr = "()"
		
		return iFunc([min, inpt], expr, opt_fun, prec)


class SimulinkCheckDynamicUpperBound(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#self.outport = [Port(self.get_id(), 1, 'out', 0)]
		#self.outport = []
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		
		max = self.inport[0].get_var_name()
		inpt = self.inport[1].get_var_name()
		
		prec = "%s < %s" %(inpt, max)
		expr = "()"
		
		return iFunc([max, inpt], expr, opt_fun, prec)

