# -*- coding: utf-8 -*-

import os
from utilities.simulink_utilities import *
from blocks.simulink_system import *
from utilities.simulink_theories_path import *
import fileinput
import fractions


# Method printing the theory declaration and import command
def write_theory_title(filename, currentFolder, opt_comp, opt_sim, opt_iter):
	isa_sim_theory = get_sim_theory_path(filename, currentFolder)
	isa_py_theory = get_py_theory_path(filename, currentFolder)
	isa_iter_theory = get_iter_theory_path(filename, currentFolder)
	
	x = {"-fp":"feedback-parallel translation", "-ic":"incremental translation", "-ici":"incremental translation with intermediate outputs", "-nfb":"feedbackless translation"}
	op = x[opt_comp]
	
	outputFile = open(filename, "w")
	print >> outputFile, 'theory %s imports\n  \"%s\"' %(os.path.split(filename)[1].replace(".thy",""), isa_sim_theory)
	
	if opt_iter:
		print >> outputFile, '  \"%s\"' % isa_iter_theory
	
	if opt_sim:
		print >> outputFile, '  \"%s\"' % isa_py_theory
	
	print >> outputFile, 'begin\n' 		
	print >> outputFile, '  (* theory obtained with the %s *)\n' %op
	print >> outputFile, '  (*declare [[show_sorts]]*) \n'
	outputFile.close()

# Method printing the theory header (simplification lemmas)
def write_theory_header(filename):
	outputFile = open(filename, "a")
	print >> outputFile, '  named_theorems cblock'
	print >> outputFile, '  named_theorems bblock'
	print >> outputFile, '  named_theorems fblock'
	print >> outputFile, '  named_theorems basic_simps'
	#print >> outputFile, '  named_theorems feedback_simps'	
	#print >> outputFile, '  named_theorems comp_simps'
	print >> outputFile, '  named_theorems iter_simps\n\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln " ";\n    val start_time_comp = Time.now();\n  *}\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln "Performing expansion/simplification";\n  *}\n\n'
	#print >> outputFile, '  declare [[coercion "of_bool :: bool \<Rightarrow> real"]]\n'
	print >> outputFile, '  declare [[eta_contract = false]]\n\n'
	outputFile.close()


# Method printing the model constants defined in a textbox annotation
def write_constants(consts, filename):
	if len(consts) == 0:
		return
	#print consts
	expr = ""
	for k, v in consts.iteritems():
		expr = expr + "  definition [cblock]: \""+k+" = "+v+"\"\n"
	outputFile = open(filename, "a")
	print >> outputFile, "  (* Begin definition model constants *)\n"
	print >> outputFile, expr
	print >> outputFile, "  (* End definition model constants *)\n"
	outputFile.close()	

# Method printing the MPT definitions for the basic blocks
def write_atomic_blocks(blocks, opt_fun, opt_generic, filename):
	if len(blocks) == 0:
		return
		
	declBlocks = ''
	for b in blocks:
		declBlocks += "  %s\n\n" % b.get_definition(opt_fun, opt_generic)
	
	outputFile = open(filename, "a")
	print >> outputFile, "  (* Begin definition basic block instances *)\n"
	print >> outputFile, "  " + declBlocks.strip() + "\n" 
	print >> outputFile, "  (* End definition basic block instances *)\n\n"
	outputFile.close()

# Method printing the simplification lemmas used by Isabelle
def write_simp_theorems(filename, opt_fun, opt_iter):
	outputFile = open(filename, "a")
	if opt_fun:
		print >> outputFile, '  lemmas [basic_simps] = basic_block_func_simps cblock bblock fblock comp_func_simps\n'
		'''
		print >> outputFile, '  lemmas [basic_block_simps] = basic_block_func_simps cblock bblock fblock\n'
		print >> outputFile, '  lemmas [feedback_simps] = feedback_func_simps id_def\n'
		print >> outputFile, '  lemmas [comp_simps] = comp_func_simps assert_o_def update_o_def If_prod If_eq split_def fst_conv snd_conv convert\n'
		'''
	else:
		print >> outputFile, '  lemmas [basic_block_simps] = basic_block_rel_simps cblock bblock fblock\n'
		print >> outputFile, '  lemmas [feedback_simps] = feedback_rel_simps\n'		
		print >> outputFile, '  lemmas [comp_simps] = comp_rel_simps\n'

	if opt_iter:
		print >> outputFile, '  lemmas [iter_simps] = DelayFeedbackInit_update_simp_a DelayFeedbackInit_update_simp_b DelayFeedbackInit_bot\n'

	print >> outputFile, ''
	
	outputFile.close()

# Method computing the CPT to be printed
def produce_formula(formula, opt_fun, opt_flat, opt_def, opt_comp, defs):
	name = formula.get_qualified_name()
	params = formula.get_params()
	
	isa_form = formula.isabelle(opt_fun, False, True)
	
	if params == "":
		isa_name = name
	else:
		isa_name = name + " " + params
	
	if opt_def:
		op = 'simulink_unfold'
	else:
		op = 'simplify_RCRS'

	isa_def = "  %s \"%s = %s\"" %(op, isa_name, str(isa_form))

	# obtain lists of inputs and outputs

	flat_in_list_vars  = map(Port.get_var_list, formula.inport)
	flat_out_list_vars = map(Port.get_var_list, formula.outport)
	inpt_vars  = [j for i in flat_in_list_vars for j in i]
	outpt_vars = [j for i in flat_out_list_vars for j in i]
	
	inpt_def = mk_tuple_var(inpt_vars)
	outpt_def = mk_tuple_var(outpt_vars)
	
	isa_def = '''%s
  \"%s\"
  \"%s\"\n''' %(isa_def, inpt_def, outpt_def)
  
	# obtain dependencies

	dep_bl_def = formula.get_dependent_blocks()
	dep_bl_simp = ""
	for b in dep_bl_def:
		dep_bl_simp = dep_bl_simp + b + "_simp "


	if len(dep_bl_simp) > 0:
		isa_def = isa_def + "  use (" + dep_bl_simp.strip() + ")\n\n"
	else:
		isa_def = isa_def + "\n"


	isa_def = isa_def + "  (* thm " + name + "_simp *)\n"

	defs.append(isa_def)

# Method computing (for printing) the CPTs of a system having intermediate outputs (for option -ici)
def get_formula_inc(formulaB, opt_fun, opt_flat, opt_def, opt_comp, defs):
	if not opt_flat:
		for b in formulaB.blocks:
			if isinstance(b, SimulinkSubsystem):
				defs.append("  (* Begin definition block instance " + b.get_qualified_name() + " *)\n")
				
				get_formula_inc(b, opt_fun, opt_flat, opt_def, opt_comp, defs)
				
				defs.append("  (* End definition block instance " + b.get_qualified_name() + " *)\n")


	for b in formulaB.definition_blocks:
		produce_formula(b, opt_fun, opt_flat, opt_def, opt_comp, defs)

# Method computing (for printing) the CPTs of a system recursively (for option -fp, -ic)
def get_formula(formulaB, opt_fun, opt_flat, opt_def, opt_comp, opt_type, defs, flag):
	if not opt_flat:
		for b in formulaB.blocks:
			if isinstance(b, SimulinkSubsystem) and flag:
				defs.append("  (* Begin definition block instance " + b.get_qualified_name() + " *)\n")
				get_formula(b, opt_fun, opt_flat, opt_def, opt_comp, opt_type, defs, flag)
				defs.append("  (* End definition block instance " + b.get_qualified_name() + " *)\n")
	
	formula = formulaB.formula
	produce_formula(formula, opt_fun, opt_flat, opt_def, opt_comp, defs)

# Method printing the CPTs of a system (for option -fp, -ic, -ici)
def write_formula(system, opt_fun, opt_flat, opt_def, opt_generic, opt_type, opt_sim, opt_iter, opt_comp, filename):
	
	defs = []
	
	outputFile = open(filename, 'a')
	print >> outputFile, "  (* Begin definition model " + system.formula.get_name() + " *)\n"
	outputFile.close()
	
	if isinstance(system, SimulinkModel) and len(system.definition_blocks) > 0:
		get_formula_inc(system, opt_fun, opt_flat, opt_def, opt_comp, defs)
		get_formula(system, opt_fun, opt_flat, opt_def, opt_comp, opt_type, defs, False)
	else:
		get_formula(system, opt_fun, opt_flat, opt_def, opt_comp, opt_type, defs, True)
	
	outputFile = open(filename, 'a')
	for f in defs:
		print >> outputFile, f
	print >> outputFile, "  (* End definition model " + system.formula.get_name() + " *)\n\n"
	outputFile.close()

# Method printing the CPTs of a system (for option -nfb)
def write_formula_nfb(system, opt_fun, opt_def, filename):
	outputFile = open(filename, 'a')
	print >> outputFile, "  (* Begin definition model " + system.get_name() + " *)\n"
	outputFile.close()

	defs = []
	for b in system.display_blocks:
		if not isinstance(b,SimulinkFunction):
			produce_formula(b, opt_fun, True, opt_def, "-nfb", defs)

	outputFile = open(filename, 'a')
	for f in defs:
		print >> outputFile, f
	outputFile.close()

	outputFile = open(filename, 'a')
	print >> outputFile, "  (* End definition model " + system.get_name() + " *)\n\n"
	outputFile.close()

# Method printing the CPTs of a system
def write_isabelle(system, opt_comp, opt_flat, opt_fun, opt_consts, opt_generic, opt_type, opt_iter, opt_sim, filename):
	if system.display_blocks == None:
		write_formula(system, opt_fun, opt_flat, False, opt_generic, opt_type, opt_sim, opt_iter, opt_comp, filename)
		b = system
	else:
		write_formula_nfb(system, opt_fun, False, filename)
		b = system.display_blocks[len(system.display_blocks)-1]
		
	if opt_type:
		write_formula_type(b, opt_type, filename)
		
	if opt_sim:
		write_formula_sim(b, opt_type, filename)

# Method computing (for printing) the CPTs of a system with type specified
def get_formula_type(formulaB, opt_type):
	formula = formulaB.formula
	name = formula.get_name()
	params = formula.get_params()
	
	lhs = "%s_type %s" %(name, params)
	rhs = "%s %s" %(name, params)
	
	# map ports with arities to lists of variable names
	flat_in_list_vars  = map(Port.get_var_list, formula.inport)
	flat_out_list_vars = map(Port.get_var_list, formula.outport)
	inpt_vars  = [j for i in flat_in_list_vars for j in i]
	outpt_vars = [j for i in flat_out_list_vars for j in i]
	
	inpt_def = mk_tuple_var(inpt_vars)
	outpt_def = mk_tuple_var(outpt_vars)
	
	isa_def = '''  simplify_RCRS \"%s = %s\"
  \"%s\"
  \"%s\"
  use (%s_simp if_simps)
  type \"%s\"\n\n''' %(lhs, rhs, inpt_def, outpt_def, name, opt_type)
	
	return isa_def

# Method printing the CPTs of a system with type specified
def write_formula_type(formula, opt_type, filename):
	r = get_formula_type(formula, opt_type)
	outputFile = open(filename, 'a')
	print >> outputFile, r
	outputFile.close()


# Method computing the system's CPT prepared for simulation
def get_formula_sim(formulaB, opt_type):
	formula = formulaB.formula
	name = formula.get_name()
	params = formula.get_params()
	
	if opt_type:
		name_aux = "%s_type" %name
	else:
		name_aux = name
	
	lhs = "%s_py %s" %(name, params)
	rhs = "%s %s" %(name_aux, params)
	usef = "%s_simp" % name_aux
	
	# map ports with arities to lists of variable names
	flat_in_list_vars  = map(Port.get_var_list, formula.inport)
	flat_out_list_vars = map(Port.get_var_list, formula.outport)
	inpt_vars  = [j for i in flat_in_list_vars for j in i]
	outpt_vars = [j for i in flat_out_list_vars for j in i]
	
	inpt_def = mk_tuple_var(inpt_vars)
	outpt_def = mk_tuple_var(outpt_vars)
	
	isa_def = '''  simplify_RCRS \" %s = %s \"
  \"%s\"
  \"%s\"
  use (%s python_simps)\n\n''' %(lhs, rhs, inpt_def, outpt_def, usef)

	return isa_def

# Method printing the system's CPT for simulation
def write_formula_sim(formula, opt_type, filename):
	r = get_formula_sim(formula, opt_type)
	outputFile = open(filename, 'a')
	print >> outputFile, r
	outputFile.close()


# Method computing the CPT of the system prepared for iteration
def get_formula_iter_split(formulaB):
	formula = formulaB.formula
	
	name = formula.get_name() + "A"
	params = formula.get_params()
	
	state_inpts_aux = [i for i in formula.inport if i.direction == "sin"]
	state_inpts = sorted(state_inpts_aux, key = Port.get_var_name)
	val_inpts = [i for i in formula.inport if i not in state_inpts]
	
	state_outpts_aux = [i for i in formula.outport if i.direction == "sout"]
	
	assert(len(state_inpts)==len(state_outpts_aux))
	
	state_outpts = sorted(state_outpts_aux, key = Port.get_var_name)
	val_outpts = [i for i in formula.outport if i not in state_outpts]
	
	inpt_def = iSwitch([state_inpts, val_inpts], [formula.inport], True)
	outpt_def = iSwitch([formula.outport],[state_outpts, val_outpts], True)
	
	lhs = "%s %s" %(name, params)
	rhs = "%s \<circ> %s %s \<circ> %s" %(inpt_def, formula.get_name(), params, outpt_def)
	
	
	# map ports with arities to lists of variable names
	flat_in_list_vars  = [map(Port.get_var_list, x) for x in [state_inpts, val_inpts]]
	flat_out_list_vars = [map(Port.get_var_list, x) for x in [state_outpts, val_outpts]]
	f = lambda x: [j for i in x for j in i]
	
	# concatenate the list of variables from the same component
	flat_in_vars   = map(f,  flat_in_list_vars)
	flat_out_vars  = map(f,  flat_out_list_vars)
	
	inpts = mk_tuple_var(map(mk_tuple_var, flat_in_vars))
	outpts = mk_tuple_var(map(mk_tuple_var, flat_out_vars))
	
	isa_def = '''simplify_RCRS \"%s = %s\"
  \"%s\"
  \"%s\"
  use (%s_simp)\n\n''' %(lhs, rhs, inpts, outpts, formula.get_name())
	
	return isa_def

# Method computing the iteration of the system CPT
def get_formula_iter(system):
	
	if system.display_blocks == None:
		b = system
	else:
		b = system.display_blocks[len(system.display_blocks)-1]
	
	r = get_formula_iter_split(b)
	
	b = b.formula

	init_vals = []
	aux = sorted([p for p in b.inport if p.direction == "sin"], key = Port.get_var_name)
	for p in aux:
		val = p.get_value()
		init_vals.append(str(val))

	name_aux = b.get_name()
	params = b.get_params()
	
	lhs = "%s_iter %s" %(name_aux, params)
	rhs = "DelayFeedbackInit init_vals (%sA %s)" %(name_aux, params)
	
	state_inpts_aux = [i for i in b.inport if i.direction == "sin"]
	state_inpts = sorted(state_inpts_aux, key = Port.get_var_name)
	val_inpts = [i for i in b.inport if i not in state_inpts]
	
	state_outpts_aux = [i for i in b.outport if i.direction == "sout"]
	
	assert(len(state_inpts)==len(state_outpts_aux))
	
	state_outpts = sorted(state_outpts_aux, key = Port.get_var_name)
	val_outpts = [i for i in b.outport if i not in state_outpts]
	
	#flat_in_list_vars  = [map(Port.get_var_list, x) for x in [val_inpts]]
	#flat_out_list_vars = [map(Port.get_var_list, x) for x in [val_outpts]]
	#f = lambda x: [j for i in x for j in i]
	
	# concatenate the list of variables from the same component
	#flat_in_vars   = map(f,  flat_in_list_vars)
	#flat_out_vars  = map(f,  flat_out_list_vars)
	
	#inpts = mk_tuple_var(map(mk_tuple_var, flat_in_vars))
	#outpts = mk_tuple_var(map(mk_tuple_var, flat_out_vars))
	
	# inputs and outputs are functions from nat to tuples
	inpts = "(x)"
	outpts = "(y)"

	
	isa_def = '''  definition \"init_vals = (%s)\"\n
  %s
  simplify_RCRS  \"%s = %s\"
  \"%s\"
  \"%s\"
  use(%sA_simp init_vals_def iter_simps)

  (* This is the basic simplified iterated formula (property transformer) of the top-level model *)

  thm %s_iter_simp\n\n''' %(",".join(init_vals), r, lhs, rhs, inpts, outpts, name_aux, name_aux)
	  
	return isa_def

# Method printing the system CPT's iteration
def write_formula_iter(formula, filename):
	r = get_formula_iter(formula)
	outputFile = open(filename, 'a')
	print >> outputFile, r
	outputFile.close()

# Method computing the iteration of the system's CPT for a specified type of variables
def get_formula_iter_type(system, opt_type):
	
	if system.display_blocks == None:
		formula = system.formula
	else:
		formula = system.display_blocks[len(system.display_blocks)-1].formula

	name = formula.get_name() + "_iter_type"
	params = formula.get_params()

	lhs = "%s %s" %(name, params)
	rhs = "%s %s" %(formula.get_name() + "_iter", params)
	
	
	state_inpts_aux = [i for i in formula.inport if i.direction == "sin"]
	state_inpts = sorted(state_inpts_aux, key = Port.get_var_name)
	val_inpts = [i for i in formula.inport if i not in state_inpts]
	
	state_outpts_aux = [i for i in formula.outport if i.direction == "sout"]
	
	assert(len(state_inpts)==len(state_outpts_aux))
	
	#state_outpts = sorted(state_outpts_aux, key = Port.get_var_name)
	#val_outpts = [i for i in formula.outport if i not in state_outpts]
	
	#flat_in_list_vars  = [map(Port.get_var_list, x) for x in [val_inpts]]
	#flat_out_list_vars = [map(Port.get_var_list, x) for x in [val_outpts]]
	#f = lambda x: [j for i in x for j in i]
	
	# concatenate the list of variables from the same component
	#flat_in_vars   = map(f,  flat_in_list_vars)
	#flat_out_vars  = map(f,  flat_out_list_vars)
	
	#inpts = mk_tuple_var(map(mk_tuple_var, flat_in_vars))
	#outpts = mk_tuple_var(map(mk_tuple_var, flat_out_vars))
	
	# inputs and outputs are functions from nat to tuples
	inpts = "(x)"
	outpts = "(y)"

	
	isa_def = '''  simplify_RCRS \"%s = %s\"
  \"%s\"
  \"%s\"
  use (%s_iter_simp if_simps)
  type \"%s\"

  (* This is the simplified iterated formula (property transformer) of the top-level model where the type of variables has been set to %s as specified with the -type option *)
  
  thm %s_iter_type_simp\n''' %(lhs, rhs, inpts, outpts, formula.get_name(),opt_type,opt_type,formula.get_name())
	
	return isa_def

# Method printing the iteration of the system's CPT for a specified type of variables
def write_formula_iter_type(formula, opt_type, filename):
	r = get_formula_iter_type(formula, opt_type)
	outputFile = open(filename, 'a')
	print >> outputFile, r
	outputFile.close()

# Method printing intermediary ML code to separate computations from result
def write_theory_inter_body(filename):
	outputFile = open(filename, "a")
	print >> outputFile, '  ML{* \n    val end_time_comp = Time.now();\n  *}\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln " ";\n  *}\n'
	print >> outputFile, '  ML{* \n    val start_time_print = Time.now();\n  *}\n\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln " ";\n  *}\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln "Expansion/simplification result";\n  *}\n'
	outputFile.close()

# Method printing the system's simplified MPT(s)
def write_theory_footer(filename, system, opt_type, opt_sim):
	
	if system.display_blocks == None:
		simp_term_name = system.formula.get_name()
	else:
		simp_term_name = system.display_blocks[len(system.display_blocks)-1].formula.get_name()
	
	simp_term_expr = '  thm %s_simp\n' %simp_term_name
	simp_type_term_expr = '  thm %s_type_simp\n' %simp_term_name
	simp_sim_term_expr = '  thm %s_py_simp\n' %simp_term_name


	outputFile = open(filename, "a")

	print >> outputFile, '  (* The following theorems state equivalence of different versions of the top-level model and the corresponding simplified terms: *)\n'

	print >> outputFile, '  (* This is the basic simplified formula of the top-level model *)\n'
	print >> outputFile, simp_term_expr
	
	if opt_type:
		print >> outputFile, '  (* This is the simplified formula of the top-level model where the types of variables have been set to %s as specified with the -type option *)\n' %opt_type
		print >> outputFile, simp_type_term_expr

	if opt_sim:
		print >> outputFile, '  (* This is the simplified formula of the top-level model used within the simulation script *)\n'
		print >> outputFile, simp_sim_term_expr
	
	print >> outputFile, '  ML{* \n    val _ = pwriteln " ";\n  *}\n'
	print >> outputFile, '  ML{* \n    val end_time_print = Time.now();\n  *}\n'
	print >> outputFile, '  ML{* \n    val _ = pwriteln " ";\n  *}\n\n'
	outputFile.close()

# Method printing the script for generating the simplified formula as Python code (for simulation)
def write_sim_step_file(system, filename):
	formula = system.formula
	name = formula.get_name() + "_py_simp"
	params = formula.get_params()
	
	
	lhs = "%s_A" %(name)
	rhs = "%s [of %s]" %(name, params)
	def_py = "  lemmas %s = %s\n" %(lhs, rhs)
	
	flat_in_list_vars  = map(Port.get_var_list, formula.inport)
	inpt_vars  = [j for i in flat_in_list_vars for j in i]	
	inpt_def = mk_tuple_var(inpt_vars)
	
	file_py = os.path.split(filename)[1].replace(".thy","_step.py")
	code_py = '''  ML{*
    val _ = write_thm_python \"%s\" @{context} @{thm %s}
	  @{term \"%s\"};
	*}\n''' %(file_py, lhs, inpt_def)
	
	outputFile = open(filename, "a")
	print >> outputFile, def_py
	print >> outputFile, code_py
	outputFile.close()


# Method printing the theory eof
def write_theory_end(filename):
	outputFile = open(filename, "a")
	print >> outputFile, 'end'
	outputFile.close()

# Method printing the theory with the equivalence lemmas
def write_equiv_theory(ofilename, formulas):
	filename = ofilename.replace(".thy","_equiv.thy")	
	
	#get theory dependencies
	fimport = ''
	for o in ["fp", "ic", "ici", "nfb"]:
		fimport += '"'
		#fimport +=  ofilename.replace(".thy","_")
		fimport += os.path.split(ofilename)[1].replace(".thy","_")
		fimport += o
		fimport += '" '
	
	outputFile = open(filename, "w")
	print >> outputFile, 'theory %s imports %s \nbegin\n\n  (*declare [[show_sorts]]*) \n' % (os.path.split(filename)[1].replace(".thy",""), fimport)
	print >> outputFile, '  lemmas comp_simps = comp_func_simps assert_o_def update_o_def If_prod If_eq convert mult_ac'
	print >> outputFile,'  lemmas equiv_pt_simps = MyIf_def\n'
	
	main = formulas[0].formula
	
	s_in = sorted([a for x in main.inport for a in x.get_var_list()])
	s_out = sorted([a for x in main.outport for a in x.get_var_list()])
	

	#print main
	for i in range(1, len(formulas)):
		if i != len(formulas):
			alt = formulas[i].formula
		else:
			alt = formulas[i]

		#check separately for identical inputs and outputs
		
		sa_in = sorted([a for x in alt.inport for a in x.get_var_list()])
		sa_out = sorted([a for x in alt.outport for a in x.get_var_list()])

		'''
		if s_in != sa_in or s_out != sa_out:
			raise Exception("The inputs/outputs are different between the two formula. The equivalence lemma cannot be generated.")
		'''

		sout_aux = [s for s in s_out if not (s in sa_out)]

		if s_in != sa_in or len(sout_aux)>0:
			raise Exception("\nError:\n  An internal error has occured.\n  The inputs/outputs are different between the two formula.\n  The equivalence lemma cannot be generated.\n")

		inp = iSwitchSplit([main.inport], [alt.inport], True)
		out = iSwitchSplit([alt.outport], [main.outport], True)

		lhs = "(%s %s)" %(main.get_name(), main.get_params())
		rhs = "(%s \<circ> (%s %s) \<circ> %s)" %(inp, alt.get_name(), alt.get_params(), out)
		strategy = '''    apply (simp only: %s_simp %s_simp comp_simps additional_simps fst_conv snd_conv split)
    by (unfold_safe_auto)\n\n''' %(main.get_name(), alt.get_name())

		print >> outputFile, "  lemma \"%s = %s\"" %(lhs,rhs)
		print >> outputFile, strategy
	print >> outputFile, 'end'
	
	outputFile.close()

# Method printing the simulation script
def write_sim_file(res, oFilename, currentFolder, consts):
	filename = oFilename.replace(".thy", ".py")
	aux = oFilename.split("/")[-1]
	#print aux
	file_step = aux.replace(".thy", "_step")
	file_res = aux.replace(".thy","_res.m")

	if res.display_blocks == None:
		formula = res.formula
	else:
		formula = res.display_blocks[len(res.display_blocks)-1]

	flat_in_list_vars  = map(Port.get_var_list, formula.inport)
	inpt_vars  = [j for i in flat_in_list_vars for j in i]
	
	vals = {}
	res.get_sample_dt(vals)
	port_st = {}
	for b in vals.keys():
		for p in b.inport:
			if p.get_var_name() in inpt_vars:
				try:
					x = float(vals[b])
					if x == -1:
						continue
				except:
					pass
				'''
				if float(vals[b]) == -1:
					continue
				'''
				port_st[p.get_var_name()] = vals[b]
				try:
					port_st[p.get_var_name()] = consts[vals[b]]
				except:
					pass
	st_list = list(set([port_st[b] for b in port_st.keys()]))
	delay_var = []
	for i in st_list:
		aux = []
		for b in port_st.keys():
			if port_st[b] == i:
				aux.append(b)
		delay_var.append(aux)

	if port_st:
		max_mul = 0
		for e in port_st.keys():
			m = len(port_st[e].split('.')[1])
			if m > max_mul:
				max_mul = m
		aux = [float(port_st[b])*(10**max_mul) for b in port_st.keys() ]
		
		if aux:
			sample_time = reduce(fractions.gcd, aux)/(10**max_mul)
		else:
			sample_time = 0.01
	else:
		sample_time = 0.01

	inpts = ", ".join(inpt_vars)
	inpt_st_vars = [i for i in inpt_vars if i.startswith("si_")]

	ids2keep = [i for i, dt in enumerate(st_list) if dt != str(sample_time)]
	st_list = [e for i, e in enumerate(st_list) if i in ids2keep]
	delay_var = [e for i, e in enumerate(delay_var) if i in ids2keep]

	delay_var_flat_set = reduce(lambda a, b: a | b, map(set, delay_var), set())
	
	for e in delay_var_flat_set:
		idx = inpt_st_vars.index(e)
		del inpt_st_vars[idx]

	inpts_state = mk_tuple_var(inpt_st_vars)
	
	init_vals_aux = [str(p.get_value()) for p in formula.inport]
	
	init_vals_aux = []
	for p in formula.inport:
		val = p.get_value()
		try:
			init_vals_aux.append(consts[val])
		except:
			init_vals_aux.append(str(val))
	init_vals = mk_tuple_var(init_vals_aux)
	
	flat_out_list_vars = map(Port.get_var_list, formula.outport)  
	outpt_vars = [j for i in flat_out_list_vars for j in i]
	outpts = mk_tuple_var(outpt_vars)
	outpt_st_vars = [i.replace("si_", "so_") for i in inpt_st_vars]
	outpts_state = mk_tuple_var(outpt_st_vars)	
	
	outpts_orig_aux = [i for i in outpt_vars if i not in outpt_st_vars]
	outpts_orig = ", ".join(outpts_orig_aux)
	
	if 'dt' in formula.get_params_list():
		sig = inpts.replace("(","").replace(")","") + ', dt'
	else:
		sig = inpts.replace("(","").replace(")","")
	
	xx = get_sim_rel_path(filename, currentFolder)

	header_import = '''#!/usr/bin/python
# -*- coding: utf-8 -*-

import os, sys
lib_path = os.path.abspath(\"%s\")
sys.path.append(lib_path)

from utilities.simulink_options import *
from utilities.simulink_theories_path import *
		
currentFolder = os.path.dirname(rnpath(__file__))\n''' %xx

	#	header = '''use_mpmath = 0
	#if use_mpmath:
	#	from utilities.simulink_sim_mpmath import *
	#else:
	#	from utilities.simulink_sim_math import *\n'''

	#	fcn = '''def step(%s):
	#	#copy here the simplified predicate transformer computed by Isabelle
	#	assert(True)
	#	return ()\n''' %sig

	filecheck = '''if not os.path.isfile(currentFolder + "/%s.py"):
	print
	print "The simplified CPT is not yet defined."
	print "Please run Isabelle on the theory file in order to obtain it."
	print
	exit(1)\n''' %file_step

	initargs = '''args = sys.argv[1:]
opts = OptionsSimulation(args,os.path.realpath(os.path.normpath(__file__)).split("/")[-1])
	
if opts.mpmath:
	try:
		from utilities.simulink_sim_mpmath import *
	except ImportError as e:
		print
		print "The mpmath library is not installed/activated in the current session.\\nContinuing with the standard python math library."
		from utilities.simulink_sim_math import *
else:
	from utilities.simulink_sim_math import *
	
if opts.time:
	sim_time = float(opts.time)
else:
	sim_time = 50.0
	
if opts.dt:
	dt = float(opts.dt)
else:
	dt = %s\n''' % sample_time

	fcn = '''execfile(currentFolder + '/%s.py')\n''' %(file_step)

	initv = '(%s) = %s\n' %(inpts, init_vals)

	initc = '''nb_steps = sim_time / dt
		
delay_dt = %s
t = %s
		
iter_nr = 0

print
print "Running simulation from 0 to", sim_time, "with a time step of", str(dt)+"."
print "Please wait..."
print

oFile = open("%s", "w")
print >> oFile, "True = true"
print >> oFile, "False = false"
print >> oFile, "R = ["

while iter_nr <= nb_steps:\n''' %("["+",".join(st_list)+"]",
								[0.0]*len(st_list), file_res)

	main = '''\ttry:
		%s = step(%s)
	except AssertionError as e:
		oFile.close()
		print "The predicate transformer's assert is violated for this time step.\\nPossible causes are a too coarse time step or not enough precision in float representation.\\nSolutions include using a different time step (-dt option) or using the mpmath library (-mpmath option).\\nTo view the syntax for these options use -help."
		print
		exit(2)
	except OverflowError as e:
		oFile.close()
		print "A numerical error (overflow, etc.) has occurred during the simulation.\\nPossible causes are a too coarse time step or not enough precision in float representation.\\nSolutions include using a different time step (-dt option) or using the mpmath library (-mpmath option).\\nTo view the syntax for these options use -help."
		print
		exit(3)
	except BaseException as e:
		oFile.close()
		print "The following error has occured during simulation:", e,".\\nPlease check the simulation options with -help."
		print
		exit(1)\n''' %(outpts, sig)
	

	if inpt_st_vars:
		main = main + "\n\t"+inpts_state +" = " + outpts_state 
				
	delayed_var_update_tmpl = '''
	if t[%s] > delay_dt[%s]:
		%s
		t[%s] = t[%s] - delay_dt[%s]
'''
	
	delayed_var_update = '\n\t'.join([
		delayed_var_update_tmpl % (i, i,
		   '\n\t\t'.join(['%s = %s' % (si_var, si_var.replace('si_', 'so_'))
			for si_var in delay_var[i]]),
		i, i, i)
		for i, _ in enumerate(st_list)
	])
	
	end = '''\tprint >> oFile, %s, ";"
	
	iter_nr += 1

print >> oFile, "]\\n"
#print >> oFile, "plot(R)\\n"
print >> oFile, \'''for i=1:%s\n  subplot(%s,1,i);\n  plot(R(:,i));\nend\'''
oFile.close()

print "Simulation finished."
print
print "The numerical results are in file %s."
print "To view the simulation plot, run this file in Matlab."
print ""''' %(outpts_orig, len(outpts_orig_aux), len(outpts_orig_aux), file_res)
	
	outputFile = open(filename, "w")
	print >> outputFile, header_import
	#print >> outputFile, header
	print >> outputFile, filecheck
	print >> outputFile, initargs
	print >> outputFile, fcn
	print >> outputFile, initv
	print >> outputFile, initc
	print >> outputFile, main
	print >> outputFile, delayed_var_update
	#print >> outputFile, var_update
	print >> outputFile, end
	outputFile.close()	





