# -*- coding: utf-8 -*-

class Options:
	options = [
			([("fp", 0,  [], "feedback-parallel translation"),
			  ("ic", 0,  [], "incremental translation"),
			  ("ici", 0, [], "incremental translation with intermediate outputs"),
			  ("nfb", 0,  [], "feedbackless translation")],                         "nfb", "comp"),
			([("flat", 0, [],  "flatten diagram for translation")],                  False, "flat"),
			([("const", 0, [], "generates parameter types for constants (to be used in case of \"Additional types\" error given by Isabelle)")], False, "const"), #newly added
			([("type", 1, ["<Isabelle type>"], "Instantiate all generic types with the parameter of this option. Parameter can be any valid Isabelle type, e.g., bool, int, real, int \<times> real, etc.")], None, "type"),
			#([("default-types", 0, [], "uses the real type for integrators and other similar blocks and the bool type for Boolean blocks")],  False, "default_types"),
			([("generic", 0, [], "Use generic types for integrators, ..., Boolean blocks")],  False, "generic"),
			([("iter", 0, [], "generates additionally the system iteration definition (requires -nfb)")],  False, "iter"),
			([("equiv", 0, [], "runs and outputs the result for all translation algorithms, together with an Isabelle theory for proving semantical equivalence ")],  False, "equiv"),
			([("sim", 0, [], "generates a template file to be used for simulating the system represented by the simplified formula over a lapse of time")],  False, "sim"),
			([("blocks", 0, [], "outputs the list of blocks that are currently handled by the translator")],  False, "sim")
			]

	def __init__(self, argv):
		self.arity = {}
		for (opts, default, name) in self.options:
			self.__dict__[name] = default
			self.__dict__[name + "_count"] = 0
			for opt in opts:
				self.arity[opt[0]] = opt[1] 
			#print name
		copts = ["-"+x[0] for (opts, default, name) in self.options for x in opts]
		files = []
		while argv:
			opt = argv[0]
			argv = argv[1:]
			if opt[:1] == "-":
				opt = opt[1:]
				if opt == "blocks":
					self.print_handled_blocks()
				ok = False
				for (opts, default, name) in self.options:
					labels = [(x[0]) for x in opts]
					if opt in labels:
						arity = self.arity[opt]
						ok = True
						if self.__dict__[name + "_count"] > 0:
							print
							print "At most one of the options %s must be given" % (", ".join(labels))
							self.usage()
						self.__dict__[name + "_count"] = 1
						if len(argv) < arity:
							print
							print "Option -%s must have %d parameter(s)" % (opt, arity)
							self.usage()
						if arity == 0:
							self.__dict__[name] = opt if len(labels) > 1 else True
						elif arity == 1:
							if argv[0] in copts:
								print
								print "The parameter of option -%s is not valid." % opt
								self.usage()
							self.__dict__[name] = argv[0]
							argv = argv[1:]
						else:
							self.__dict__[name] = argv[:arity]
							argv = argv[arity:]
				if not ok:
					print
					print "Unknown option:", opt
					self.usage()
			else:
				files.append(opt)
		self.update_files(files)
				
	def update_files(self, files):
		if len(files) > 2:
			print
			print "Too many arguments."
			self.usage()
	
		if len(files) == 0:
			print
			print "Missing Simulink filename."
			self.usage()

		self.input_file = files[0]
				
		if not self.input_file.endswith(".slx"):
			print
			print "Simulink filename must have slx extension."
			self.usage()
		
		# Output file optional - if missing use the same name as the input file

		if len(files) == 1: 
			self.output_file = self.input_file.replace(".slx",".thy")
		elif len(files) == 2:
			self.output_file = files[1]
			if not self.output_file.endswith(".thy"):
				print
				print "Isabelle filename must have thy extension."
				self.usage()
				
	def usage_opt(self, opt):
		if opt[1] > 0:
			return " ".join([opt[0]] + opt[2])
		else:
			return opt[0]
			
	#Method printing compiler options
	def usage(self):
		opt = " ".join(["[-%s]" % o for o in ["|".join([self.usage_opt(y) for y in x[0]]) for x in self.options]])
		print
		print "Usage: simulink2isabelle.py %s <filename.slx> [<filename.thy>]" % opt
		
		for o in self.options:
			for op in o[0]:
				default = " (default)" if op[0] == o[1] else ""
				print "     -%s:\t%s%s" % (op[0], op[3], default)
				
		raise Exception("")
		
	def print_options(self,opt_sim):
		sim_file = self.output_file.replace(".thy",".py")
		print
		print "Input file:\t%s" % self.input_file
		print "Output files:\t%s %s" % (self.output_file, sim_file if opt_sim else "")
		for (opts, default, name) in self.options:
			i = 1 if (len(name) + 1) % 8 == 0 else 2
			print "    %s:%s" % (name, 1 * "\t"), self.__dict__[name]
		print
		
	def print_handled_blocks(self):
		handled_blocks = ["Constant", "PulseGenerator", "Step", "SignalGenerator", "Inport", "Scope", "Terminator", "Outport", "Add", "Sum", "Subtract", "UnaryMinus", "Product", "Divide", "Gain", "Sqrt", "MathFunction - only log, exp, square, pow, 10^u", "MinMax", "Fcn", "SinWave", "Logic - only and, or, not", "RelationalOperator", "Integrator", "TransferFcn ", "Relay", "Saturation" , "UnitDelay", "Mux", "Demux - only with 1 output per each branch and without the -nfb option", "Switch", "MultiportSwitch", "From", "Goto", "DataTypeConversion", "ActionPort", "SwitchCase", "Subsystem", "Enabled subsystem - only for reset", "Switch case action subsystem", "Assertion", "Check Static Range", "Check Static Gap", "Check Static Lower Bound", "Check Static Upper Bound", "Check Dynamic Bound", "Check Dynamic Gap", "Check Dynamic Lower Bound", "Check Dynamic Upper Bound"]
		print
		print "The list of accepted blocks (alphabetically ordered):"
		for x in sorted(handled_blocks):
			print "\t%s" %x
		print
		raise Exception("")

class Options2(Options):
	options = [
			([("fp", 0,  [], "feedback-parallel translation"),
			  ("ic", 0,  [], "incremental translation"),
			  ("ici", 0, [], "incremental translation with intermediate outputs"),
			  ("nfb", 0,  [], "feedbackless translation")],                         "nfb", "comp"),
			([("flat", 0, [],  "flatten diagram for translation")],                  False, "flat"),
			([("const", 0, [], "generates parameter types for constants (to be used in case of \"Additional types\" error given by Isabelle)")], False, "const"), #newly added
			([("typeA", 1, ["<Isabelle type>"], "Instantiate all generic types with the parameter of this option")], None, "typeA"),
			#([("default-types", 0, [], "uses the real type for integrators and other similar blocks and the bool type for Boolean blocks")],  False, "default_types"),
			([("generic", 0, [], "Use generic types for integrators, ..., Boolean blocks")],  False, "generic"),
			([("iter", 0, [], "generates additionally the system iteration definition (requires -nfb)")],  False, "iter"),
			([("equiv", 0, [], "runs and outputs the result for all translation algorithms, together with an Isabelle theory for proving semantical equivalence ")],  False, "equiv"),
			([("sim", 0, [], "generates a template file to be used for simulating the system represented by the simplified formula over a lapse of time")],  False, "sim")
			]
			
class OptionsSimulation(Options):
	options = [([("dt", 1, ["<value>"], "sets the time step to the specified value (the default value is computed and reported upon running this program)")], None, "dt"),
		([("time", 1, ["<value>"], "sets the simulation time (length) to the specified value (default value: 50)")], None, "time"),
		([("mpmath", 0, [], "uses the mpmath library functions for the numerical computations with arbitrary float precision, instead of the standard python math library. This library must be installed beforehand.")],  False, "mpmath"),
		([("help", 0, [], "prints the usage command line")],  False, "help")
		]

	def __init__(self, argv, filename):
		self.arity = {}
		for (opts, default, name) in self.options:
			self.__dict__[name] = default
			self.__dict__[name + "_count"] = 0
			for opt in opts:
				self.arity[opt[0]] = opt[1] 
			#print name
		files = []
		while argv:
			opt = argv[0]
			argv = argv[1:]
			#print opt
			if opt[:1] == "-":
				opt = opt[1:]
				ok = False
				for (opts, default, name) in self.options:
					labels = [(x[0]) for x in opts]
					#print opts
					if opt in labels:
						arity = self.arity[opt]
						ok = True
						if self.__dict__[name + "_count"] > 0:
							print
							print "At most one of the options %s must be given" % (", ".join(labels))
							self.usage(filename)
						self.__dict__[name + "_count"] = 1
						if len(argv) < arity:
							print
							print "Option -%s must have %d parameter(s)" % (opt, arity)
							self.usage(filename)
						if arity == 0:
							self.__dict__[name] = opt if len(labels) > 1 else True
						elif arity == 1:
							self.__dict__[name] = argv[0]
							argv = argv[1:]
						else:
							self.__dict__[name] = argv[:arity]
							argv = argv[arity:]
				if not ok:
					print
					print "Unknown option:", opt
					self.usage(filename)

		x = self.__dict__["help"]
		if x:
			self.usage(filename)

	def usage_opt(self, opt):
		if opt[1] > 0:
			return " ".join([opt[0]] + opt[2])
		else:
			return opt[0]
					
	def usage(self,filename):
		opt = " ".join(["[-%s]" % o for o in ["|".join([self.usage_opt(y) for y in x[0]]) for x in self.options]])
		print
		print "Usage: %s %s" % (filename+".py", opt)
		
		for o in self.options:
			for op in o[0]:
				default = " (default)" if op[0] == o[1] else ""
				print "     -%s:\t%s%s" % (op[0], op[3], default)
		print
		
		exit(1)
