subsection{*\label{DelayFeedback}Iterate Operators*}

theory IterateOperators imports "../RefinementReactive/RefinementReactive"
begin

definition append_inf :: "'a list \<Rightarrow> (nat \<Rightarrow> 'a) \<Rightarrow> (nat \<Rightarrow> 'a)" (infixr ".." 65) where
    "(xs..s) i = (if i < length xs then xs ! i else s (i - (length xs)))"

lemma [simp]: "[x 0] .. x[Suc 0..] = x"
  by (simp add: fun_eq_iff append_inf_def)

lemma [simp]: "([a] .. x)[Suc 0 ..] = x"
  by (simp add: fun_eq_iff append_inf_def)

lemma [simp]: "([a] .. x) 0 = a"
  by (simp add: fun_eq_iff append_inf_def)

definition "SkipNext S = [: x \<leadsto> a, b . a = x \<and> b = x[Suc 0..] :] o (Prod Skip S) o [: a, b \<leadsto> x .  x = cat (Suc 0) a b :]"

definition "Next S = [:x \<leadsto> y . y = x[Suc 0..]:] o S o [:y \<leadsto> x . y = x[Suc 0..]:]"
    
definition "NextAngelic S = {: x \<leadsto> y . y = x[Suc 0..]:} o S o {:y \<leadsto> x . y = x[Suc 0..] :}"

definition "SkipTop n = [:eqtop n:]"

lemma SkipNext_Next: "SkipNext S = Next S \<parallel> SkipTop (Suc 0)"
proof (simp add: SkipNext_def Next_def Prod_def fusion_def demonic_def le_fun_def prod_pred_def Skip_def SkipTop_def fun_eq_iff, safe)
  fix x :: "'a \<Rightarrow> bool" fix xa :: 'a fix p :: "'a \<Rightarrow> bool" fix p' :: "'a \<Rightarrow> bool"
  assume "\<forall>a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)"
  from this have A: "\<And> a b .  p a \<Longrightarrow> p' b \<Longrightarrow> x (cat (Suc 0) a b)"
    by simp
  assume [simp]: "p xa"
  assume [simp]: "S p' (xa[Suc 0 ..])"
  let ?p = "\<lambda> x . p' (x[Suc 0..])"
  let ?p' = "\<lambda> x . \<exists> y .p (cat (Suc 0) x y)"
  have [simp]: "(\<And> xa y. ?p xa \<Longrightarrow> p (cat (Suc 0) xa y) \<Longrightarrow> x xa)"
    by (drule A, simp, simp) 
  have [simp]: "\<And>x. eqtop (Suc 0) xa x \<Longrightarrow> p (cat (Suc 0) x (xa[Suc 0..]))"
    by (cut_tac x = "cat (Suc 0) x (xa[Suc 0 ..])" and y = xa and n = "Suc 0" in eqtop_tail, simp_all add: eqtop_sym)     
  from this have [simp]: "\<forall>x. eqtop (Suc 0) xa x \<longrightarrow> ?p' x"
    by blast
  have [simp]: "[: \<lambda>y x. y = x[Suc 0 ..] :] (\<lambda>x. p' (x[Suc 0 ..])) = p'"
    apply (simp add: fun_eq_iff demonic_def le_fun_def, safe)
    by (drule_tac x = "cons x" in spec, simp)
  show "\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> S ([: \<lambda>y x. y = x[Suc 0 ..] :] p) (xa[Suc 0 ..]) \<and> (\<forall>x. eqtop (Suc 0) xa x \<longrightarrow> p' x)"
    by (rule_tac x = ?p in exI, rule_tac x = ?p' in exI, auto)
next
  fix x :: "'a \<Rightarrow> bool" fix xa :: 'a fix p :: "'a \<Rightarrow> bool" fix p' :: "'a \<Rightarrow> bool"
  assume [simp]: "\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa"
  assume [simp]: " S ([: \<lambda>y x. y = x[Suc 0 ..] :] p) (xa[Suc 0 ..])"
  assume [simp]: " \<forall>x. eqtop (Suc 0) xa x \<longrightarrow> p' x"
  show "\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> p xa \<and> S p' (xa[Suc 0 ..])" 
    apply (rule_tac x = "\<lambda> x . \<forall> y .p' (cat (Suc 0) x y)" in exI)
    apply (rule_tac x = "[: \<lambda>y x. y = x[Suc 0 ..] :] p" in exI)
    by (auto simp add: demonic_def le_fun_def)
qed

lemma [simp]: "SkipTop 0 = Havoc"
  by (simp add: SkipTop_def Havoc_def)

lemma proj_skip [simp]: "[: y \<leadsto> x. y = x[Suc 0 ..] :] \<circ> [: x \<leadsto> y. y = x[Suc 0 ..] :] = Skip"
proof - 
  have [simp]: "[: \<lambda>y x. y = x[Suc 0 ..] :] \<circ> [: \<lambda>x y. y = x[Suc 0 ..] :] = [: (\<lambda>y x. y = x[Suc 0 ..]) OO (\<lambda>x y. y = x[Suc 0 ..]) :]" 
    by (simp add: demonic_demonic)
  have [simp]:" (\<lambda>y x. y = x[Suc 0 ..]) OO (\<lambda>x y. y = x[Suc 0 ..]) = op ="
    by (simp add: fun_eq_iff OO_def, auto)
  show ?thesis by (simp add: demonic_eq_skip)
qed

 lemma Next_comp: "Next (S o T) = Next S o Next T"
  proof -
    have "Next (S o T) = [: \<lambda>x y. y = x[Suc 0 ..] :] \<circ> S \<circ> ([: \<lambda>y x. y = x[Suc 0 ..] :] \<circ> [: \<lambda>x y. y = x[Suc 0 ..] :]) \<circ> T \<circ> [: \<lambda>y x. y = x[Suc 0 ..] :]"
      by (simp add: Next_def comp_assoc)
    also have "... =  Next S o Next T"
      by (unfold Next_def comp_assoc, simp)
    finally show ?thesis by simp
  qed
      
lemma transp_ref_comp: "transp r \<Longrightarrow> [:r:] \<le> [:r:] o [:r:]"
  apply (simp add: le_fun_def demonic_def)
  by (unfold transp_def, blast)

lemma fusion_comp_demonic: "transp r \<Longrightarrow> (S o T) \<parallel> [:r:] \<le> (S \<parallel> [:r:]) o (T \<parallel> [:r:])"
proof (simp add: fusion_def  le_fun_def, safe)
  fix x :: "'a \<Rightarrow> bool" fix xa :: 'a fix p :: "'a \<Rightarrow> bool" fix p' :: "'a \<Rightarrow> bool"
  assume A: "transp r"
  assume [simp]: "\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa"
  assume [simp]: "S (T p) xa"
  assume "[: r :] p' xa"
  from A and this have [simp]: "[: r :] ([: r :] p') xa"
    by (simp add: demonic_def le_fun_def, metis transp_def)
  have "\<And> xa . T p xa \<Longrightarrow> [: r :] p' xa \<Longrightarrow> (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> T p xa \<and> [: r :] p' xa"
    by simp
  from this have [simp]:"\<And> xa . T p xa \<Longrightarrow> [: r :] p' xa \<Longrightarrow> \<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> T p xa \<and> [: r :] p' xa"
    by blast
  have "(\<forall>xa. T p xa \<and> [:r:] p' xa \<longrightarrow> (\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> T p xa \<and> [: r :] p' xa)) \<and> S (T p) xa \<and> [: r :] ([:r:] p') xa"
    by simp
  from this show "\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> (\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> T p xa \<and> [: r :] p' xa)) \<and> S p xa \<and> [: r :] p' xa"
    by blast
qed

lemma fusion_comp_eqtop: "(S o T) \<parallel> [:eqtop n:] \<le> (S \<parallel> [:eqtop n:]) o (T \<parallel> [: eqtop n:])"
  by (metis eqtop_tran fusion_comp_demonic transpI)

lemma SkipNext_comp_a[simp]: "SkipNext (S o T) \<le> (SkipNext S) o (SkipNext T)"
  by (simp add: SkipNext_Next SkipTop_def Next_comp, rule fusion_comp_eqtop)

definition "auxfun p' T x xa = (SUPREMUM {b. p' b} (\<lambda> b . (Sup {p' . (\<exists> p . (\<forall> a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> p xa \<and> T p' b)})))"

lemma SkipNext_comp_b[simp]: "mono S \<Longrightarrow> mono T\<Longrightarrow> SkipNext (S o T) \<ge> (SkipNext S) o (SkipNext T)"
proof (simp add: SkipNext_def le_fun_def Prod_def prod_pred_def Skip_def demonic_def, safe)
  fix x :: "'a \<Rightarrow> bool" fix xa :: 'a fix p :: "'a \<Rightarrow> bool"  fix p':: "'a \<Rightarrow> bool"
  assume [simp]: "p xa"
  assume "\<forall>a b. p a \<and> p' b \<longrightarrow> (\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> p (cat (Suc 0) a b) \<and> T p' b)"
  from this have D:"\<And> b . p' b \<Longrightarrow> (\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> p (cat (Suc 0) xa b) \<and> T p' b)"
    by (drule_tac x = "xa" in spec, simp)
  from this have  E: "\<And>b xb f p . p' xb \<Longrightarrow> f b \<Longrightarrow> \<forall>a b. p a \<and> f b \<longrightarrow> x (cat (Suc 0) a b) \<Longrightarrow> p xa \<Longrightarrow> T f xb \<Longrightarrow> x (cat (Suc 0) xa b)"
    by simp
  from this have [simp]: "\<And>b. auxfun p' T x xa b \<Longrightarrow> x (cat (Suc 0) xa b)"
    by (auto simp add: auxfun_def)
  have [simp]: "\<And>xb p p'a. p' xb \<Longrightarrow> \<forall>a b. p a \<and> p'a b \<longrightarrow> x (cat (Suc 0) a b) \<Longrightarrow> p (cat (Suc 0) xa xb) \<Longrightarrow> T p'a xb \<Longrightarrow>  p'a \<le> auxfun p' T x xa"
    apply (simp add: le_fun_def auxfun_def)
    apply (cut_tac b = xb in D, auto)
    apply (rule_tac x = xb in exI, simp)
    apply (rule_tac x = p'a in exI, safe, simp)
    apply (rule_tac x = "\<lambda> x . x = xa" in exI, simp, safe)
    by (drule_tac x = "cat (Suc 0) xa xb" in spec, simp_all)
  assume F: "mono T"
  have [simp]: "\<And>xb p p'a. p' xb \<Longrightarrow> \<forall>a b. p a \<and> p'a b \<longrightarrow> x (cat (Suc 0) a b) \<Longrightarrow> p (cat (Suc 0) xa xb) \<Longrightarrow> T p'a xb \<Longrightarrow> T (auxfun p' T x xa) xb"
    apply (cut_tac F)
    apply (drule_tac x = p'a and y = "auxfun p' T x xa" in monoD, simp_all)
    by (simp add: le_fun_def)

  have "\<And>xb. p' xb \<Longrightarrow> T (auxfun p' T x xa) xb"
    by (cut_tac b = xb in D, auto)
  from this have A: "p' \<le> T (auxfun p' T x xa)"
    by (simp_all add: le_fun_def)
  assume B: "S p' (xa[Suc 0 ..])"
  assume "mono S"
  from A and B and this have [simp]: "S (T (auxfun p' T x xa)) (xa[Suc 0 ..])"
    by (metis monoD predicate1D)
  let ?p =  "\<lambda> x . x = xa"
  let ?p' = "auxfun p' T x xa"
  have "(\<forall>a b. ?p a \<and> ?p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> ?p xa \<and> S (T ?p') (xa[Suc 0 ..])"
    by simp
  from this show "\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> x (cat (Suc 0) a b)) \<and> p xa \<and> S (T p') (xa[Suc 0 ..])"
    by (rule_tac x = ?p in exI, rule_tac x = ?p' in exI, simp)
qed

lemma SkipNext_comp: "mono S \<Longrightarrow> mono T\<Longrightarrow> SkipNext (S o T) = (SkipNext S) o (SkipNext T)"
  by (rule antisym, simp_all)

lemma Next_fusion: "Next (S \<parallel> T) = (Next S) \<parallel> (Next T)"
proof (simp add: Next_def fun_eq_iff fusion_def demonic_def le_fun_def, safe)
  fix x :: "'a \<Rightarrow> bool" fix xa::'b fix p::"'a \<Rightarrow> bool" fix p'::"'a \<Rightarrow> bool"
  assume [simp]: "\<forall>xa. p xa \<and> p' xa \<longrightarrow> (\<forall>xb. xa = xb[Suc 0 ..] \<longrightarrow> x xb)"
  assume [simp]: "S p (xa[Suc 0 ..])"
  assume [simp]: "T p' (xa[Suc 0 ..])"
  let ?p = "\<lambda>x. p (x[Suc 0 ..])"
  let ?p' = "\<lambda>x. p' (x[Suc 0 ..])"
  have [simp]: "\<And> p' . [: \<lambda>y x. y = x[Suc 0 ..] :] (\<lambda>x. p' (x[Suc 0 ..])) = p'"
    apply (simp add: fun_eq_iff demonic_def le_fun_def, safe)
    by (drule_tac x = "cons x" in spec, simp)
  show "\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa) \<and> S ([: \<lambda>y x. y = x[Suc 0 ..] :] p) (xa[Suc 0 ..]) \<and> T ([: \<lambda>y x. y = x[Suc 0 ..] :] p') (xa[Suc 0 ..])"
    by (rule_tac x = ?p in exI, rule_tac x = ?p' in exI, simp)
next
  fix x :: "'a \<Rightarrow> bool" fix xa::'b fix p::"'a \<Rightarrow> bool" fix p'::"'a \<Rightarrow> bool"
  assume [simp]: "\<forall>xa. p xa \<and> p' xa \<longrightarrow> x xa"
  assume [simp]: "S ([: \<lambda>y x. y = x[Suc 0 ..] :] p) (xa[Suc 0 ..])"
  assume [simp]: "T ([: \<lambda>y x. y = x[Suc 0 ..] :] p') (xa[Suc 0 ..])"
  show "\<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> (\<forall>xb. xa = xb[Suc 0 ..] \<longrightarrow> x xb)) \<and> S p (xa[Suc 0 ..]) \<and> T p' (xa[Suc 0 ..])"
    apply (rule_tac x = "[: \<lambda>y x. y = x[Suc 0 ..] :] p" in exI, simp)
    apply (rule_tac x = "[: \<lambda>y x. y = x[Suc 0 ..] :] p'" in exI, simp)
    by (auto simp add: demonic_def le_fun_def)
qed


lemma fusion_SkipTop_idemp [simp]: "SkipTop n \<parallel> SkipTop n = SkipTop n"
  by (simp add: SkipTop_def)

lemma SkipNext_fusion: "SkipNext (S \<parallel> T) = (SkipNext S) \<parallel> (SkipNext T)"
proof -
  have "(SkipNext S) \<parallel> (SkipNext T) = (Next S \<parallel> SkipTop (Suc 0)) \<parallel> (Next T \<parallel> SkipTop (Suc 0))"
    by (simp add: SkipNext_Next)
  also have "... = Next S \<parallel> (SkipTop (Suc 0) \<parallel> Next T) \<parallel> (SkipTop (Suc 0))"
    by (simp add: fusion_assoc)
  also have "... = Next S \<parallel> (Next T \<parallel> SkipTop (Suc 0)) \<parallel> SkipTop (Suc 0)"
    by (simp add: fusion_comute)
  also have "... = Next S \<parallel> Next T \<parallel> SkipTop (Suc 0)"
    by (simp add: fusion_assoc [THEN sym])
  also have "... = SkipNext (S \<parallel> T)"
    by (simp add: SkipNext_Next Next_fusion)
  finally show ?thesis by simp
qed

lemma SkipNext_SkipTop: "SkipNext (SkipTop n) = SkipTop (Suc n)"
proof -
  have [simp]: "\<And>x xa . (\<exists>v::'a. eqtop n (x[Suc (0::nat) ..]) v \<and> xa = cat (Suc (0::nat)) x v) = eqtop (Suc n) x xa"
    apply (subst eqtop_Suc_a, auto)
    apply (rule_tac x = "xa[Suc 0..]" in exI, auto)
    apply (rule_tac n = "Suc 0" in eqtop_tail, simp_all)
    by (simp add: eqtop_sym)
  have [simp]: "((\<lambda>(x::'a) (u::'a, v::'a). x = u \<and> eqtop n (x[Suc (0::nat) ..]) v) OO (\<lambda>(a::'a, b::'a) x::'a. x = cat (Suc (0::nat)) a b))
      = (\<lambda> x y . \<exists> v . eqtop n (x[Suc (0::nat) ..]) v \<and> y = cat (Suc (0::nat)) x v)"
    by (simp add: fun_eq_iff OO_def)
  have [simp]: "\<And> f g r . (\<lambda> x (a, b) . a = f x \<and> b = g x) OO r = (\<lambda> x y . r (f x, g x) y)"
    by (auto simp add: fun_eq_iff)
  show ?thesis
    by (simp add: SkipNext_def SkipTop_def Prod_skip_demonic demonic_demonic)
qed
  
lemma SkipTop_SkipNext: "SkipTop n = (SkipNext ^^ n) Havoc"
  by (induction n, simp_all add: SkipNext_SkipTop [THEN sym])

lemma SkipNext_power: "(SkipNext ^^ (Suc n)) S = (Next ^^ (Suc n)) S \<parallel> SkipTop (Suc n)"
proof (induction n)
  case 0
  show ?case by (simp add: SkipNext_Next SkipTop_def)
  case (Suc n)
  from this show ?case
    apply (simp add: SkipNext_Next Next_fusion)
    apply (simp add: fusion_assoc [THEN sym])
    apply (unfold SkipNext_Next [THEN sym])
    by (simp add: SkipNext_SkipTop)
qed

lemma Next_demonic: "Next [:r:] = [:\<Odot> r:]"
  by (auto simp add: fun_eq_iff demonic_def Next_def le_fun_def next_def at_fun_def)

lemma SkipNext_demonic: "SkipNext {.p.} = {.\<Odot> p.}"
  apply (simp add: fun_eq_iff demonic_def SkipNext_def le_fun_def next_def at_fun_def Prod_def assert_def prod_pred_def)
  apply safe
  apply (drule_tac x = "xa" in spec)
  apply (drule_tac x = "xa[Suc 0 ..]" in spec)
  apply simp
  apply (rule_tac x = "\<lambda> x . x = xa" in exI)
  apply (rule_tac x = "\<lambda> x . x = xa[Suc 0..]" in exI)
  by auto 

lemma NextAngelic_angelic: "NextAngelic ({:r::(nat \<Rightarrow> 'a) \<Rightarrow> (nat \<Rightarrow> 'a) \<Rightarrow> bool:}) = {:\<Odot> r:}"
  apply (simp add: fun_eq_iff)
  apply (simp add:  angelic_def NextAngelic_def le_fun_def next_def at_fun_def assert_def bot_fun_def fun_eq_iff)
  apply safe
  apply (rule_tac x = "xd" in exI, simp)
  apply (subgoal_tac "(xb = xa[Suc 0 ..]) \<and> (xc = xd[Suc 0 ..])")
  apply simp
  apply (simp add: append_inf_def)
  apply (rule_tac x = "xa[Suc 0..]" in exI)
  apply simp
  apply (rule_tac x = "xb[Suc 0..]" in exI)
  by auto
    
lemma Next_assert_demonic: "Next ({.p.} o [:r:]) = {.\<Odot> p.} o [:\<Odot> r:]"
  apply (simp add: fun_eq_iff demonic_def Next_def le_fun_def next_def at_fun_def assert_def)
  by blast

lemma Next_angelic_demonic: "Next ({:r:} o [:r':]) = {:\<Odot> r:} o [:\<Odot> r':]"
  apply (simp add: fun_eq_iff demonic_def Next_def le_fun_def next_def at_fun_def angelic_def)
  apply safe
  apply (rule_tac x = "cons xb" in exI, simp)
  by metis

lemma eqtop_Suc_zero: "eqtop (Suc 0) = (\<lambda> x y . x 0 = y 0)"
  apply (simp add: fun_eq_iff)
  apply (subst (3) zero_Nat_def)
  by (subst (4) zero_Nat_def, simp)

definition "idnext r = \<Odot> r \<sqinter> eqtop (Suc 0)"

lemma SkipNext_assert_demonic: "SkipNext ({.p.} o [:r:]) = {. \<Odot> p .} \<circ> [: idnext r :]"
  apply (simp add: SkipNext_Next Next_assert_demonic idnext_def SkipTop_def)
  apply (cut_tac p = "\<Odot> p" and r = "\<Odot> r" and p' = \<top> and r' = "eqtop (Suc 0)" in fusion_spec)
  by (simp add: assert_true_skip)

lemma Next_assert_demonic2: "Next (\<lambda> q . {.p.} ([:r:] q)) = {.\<Odot> p.} o [:\<Odot> r:]"
  apply (simp add: fun_eq_iff demonic_def Next_def le_fun_def next_def at_fun_def assert_def)
  by blast

lemma Iterate_Next_assert_demonic: "(Next^^n) ({.p.} o [:r:]) = {.(next^^n)p.} o [:(next^^n) r:]"
   apply (induction n)
   apply simp
   by (simp add: Next_assert_demonic2)

lemma power_SkipNext_assert_demonic: "(SkipNext^^n) ({.p.} o [:r:]) = {.(next^^n)p.} o [:(idnext^^n) r:]"
   apply (induction n)
   apply simp
   apply (simp add: SkipNext_assert_demonic del: comp_apply)
   apply simp
   apply (unfold comp_def [symmetric], simp add: SkipNext_assert_demonic)
   by (simp add: fun_eq_iff)

lemma Iterate_Next_demonic: "(Next ^^ n) [:r:] = [:(next ^^ n) r:]"
   apply (induction n)
   apply simp_all
   by (simp add: Next_demonic)

definition "Always S = Fusion (\<lambda> n . (Next ^^ n) S)"
    
lemma Always_demonic: "Always [:r:] = [:\<box> r:]"
  by (simp add: Always_def Fusion_demonic Iterate_Next_demonic iterate_next always_def)
    
lemma Always_assert_demonic: "Always ({.p.} o [:r:]) = {.\<box> p.} o [:\<box> r:]"
  by (simp add: Always_def Fusion_spec Iterate_Next_assert_demonic iterate_next always_def)
   
lemma SkipNext_simp: "SkipNext S Q x = 
    (\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> Q (cat (Suc 0) a b)) \<and> p x \<and> S p' (x[Suc 0..]))"
  by (simp add: SkipNext_def Prod_def prod_pred_def Skip_def demonic_def le_fun_def)

type_synonym ('a, 'b) trans = "('b \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool)"

primrec Iterate :: "(('a, 'a) trans \<Rightarrow> ('a, 'a) trans) \<Rightarrow> ('a, 'a) trans \<Rightarrow> nat \<Rightarrow> ('a, 'a) trans" where
  "Iterate F S 0 = Skip" |
  "Iterate F S (Suc n) = (Iterate F S n) o ((F ^^ n) S)"

  
definition "Mask n S = S o (SkipTop n)"
  
definition "IterateNextMask S n = Mask n (Iterate Next S n)"
  
lemma IterateNextMask_simp: "IterateNextMask S = (\<lambda> n . Mask n (Iterate Next S n))"
  by (simp add: fun_eq_iff IterateNextMask_def)

definition "IterateSkipNextMask S n = Mask n (Iterate SkipNext S n)"
  
lemma IterateSkipNextMask_simp: "IterateSkipNextMask S = (\<lambda> n . Mask n (Iterate SkipNext S n))"
  by (simp add: fun_eq_iff IterateSkipNextMask_def)

definition "IterateOmegaNextMask S = Fusion (IterateNextMask S)"

definition "IterateOmegaSkipNextMask S = Fusion (IterateSkipNextMask S)"

definition "AddUnitDelay S = ([:u, x, y \<leadsto> a, b . a = u (0::nat) \<and> b = x (0::nat):] o S o [:c, d \<leadsto> u', x', y' . u' (Suc 0) = c \<and> y' (0::nat) = d:]) 
      \<parallel> [:u, x, (y::nat \<Rightarrow> 'a) \<leadsto> u', x', (y'::nat \<Rightarrow>'a) . u' (0::nat) = u (0::nat) \<and> x' = x:]"
 
lemma AddUnitDelay_spec: "AddUnitDelay ({.u, x . p u x.} o [:u, x \<leadsto> u', y . r u u' x y:]) = 
   {.u, x, y . p (u 0) (x 0).} o [:u, x, y \<leadsto> u', x', y' . r (u 0) (u' (Suc 0)) (x 0) (y' 0) \<and> x = x' \<and> u 0 = u' 0:]"
   (is "?L = ?R")
  proof (unfold AddUnitDelay_def)
    have "([: u::nat \<Rightarrow> 'a, x::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c \<leadsto> a::'a, b::'b . a = u (0::nat) \<and> b = x (0::nat) :] 
           \<circ> ({.x::'a, y::'b. p x y.} \<circ> [: u::'a, x::'b \<leadsto> u'::'a, y::'c . r u u' x y :]) \<circ> [: c::'a, d::'c \<leadsto> u'::nat \<Rightarrow> 'a,
            x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c . u' (Suc (0::nat)) = c \<and> y' (0::nat) = d :])
       = {.u ,x, y. p (u (0::nat)) (x (0::nat)).} o [: u, x, y \<leadsto> a, b . a = u (0::nat) \<and> b = x (0::nat) :] 
           \<circ> [: u, x \<leadsto> u', y . r u u' x y :] \<circ> [: c, d \<leadsto> u', x', y' . u' (Suc 0) = c \<and> y' (0::nat) = d :]"
       (is "?A = ?B")
     apply (simp add: comp_assoc [THEN sym] demonic_assert_comp)
     apply (subgoal_tac "(\<lambda>x::(nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c). \<forall>(a::'a) b::'b. (case x of (u::nat \<Rightarrow> 'a, xa::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) \<Rightarrow> \<lambda>(a::'a, b::'b). a = u (0::nat) \<and> b = xa (0::nat)) (a, b) \<longrightarrow> p a b ) 
        = (\<lambda>(u, x, y). p (u (0::nat)) (x (0::nat)))")
     apply simp_all
     by (simp add: fun_eq_iff)
   also have "... = {.u ,x, y. p (u (0::nat)) (x (0::nat)).} o [: (u::nat \<Rightarrow> 'a), (x::nat \<Rightarrow> 'b), (y::nat \<Rightarrow> 'c) \<leadsto> u', x', y' . r (u (0::nat)) (u' (Suc 0)) (x (0::nat)) (y' (0::nat)) :]"
       (is "?C = ?D")
    apply (simp add: comp_assoc demonic_demonic)
    apply (subgoal_tac "(((\<lambda>(u::nat \<Rightarrow> 'a, x::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) (a::'a, b::'b). a = u (0::nat) \<and> b = x (0::nat)) OO (\<lambda>(u::'a, x::'b) (u'::'a, y::'c). r u u' x y)) OO
     (\<lambda>(c::'a, d::'c) (u'::nat \<Rightarrow> 'a, ab::(nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c)). u' (Suc (0::nat)) = c \<and> (case ab of (x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c) \<Rightarrow> y' (0::nat) = d)))
           = (\<lambda>(u::nat \<Rightarrow> 'a, x::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) (u'::nat \<Rightarrow> 'a, x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c). r (u (0::nat)) (u' (Suc (0::nat))) (x (0::nat)) (y' (0::nat)) )")
    apply (simp_all)
    by (simp add: fun_eq_iff OO_def)
  finally have "?A  = ?D" by simp
  from this have "?A \<parallel> [:u, x, y \<leadsto> u', x', y' . u' 0 = u 0 \<and> x' = x :] = ?D \<parallel> ({.\<top>.} o [:u, x, y \<leadsto> u', x', y' . u' 0 = u 0 \<and> x' = x :])"
    by (simp add: assert_true_skip)
  also have "... = ?R"
    apply (unfold fusion_spec, simp add: inf_fun_def)
    apply (subgoal_tac "(\<lambda>x::(nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c). case x of (u::nat \<Rightarrow> 'a, xa::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) \<Rightarrow> p (u (0::nat)) (xa (0::nat)))
        = (\<lambda>(u::nat \<Rightarrow> 'a, x::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c). p (u (0::nat)) (x (0::nat)))")
    apply simp
    apply (subgoal_tac "( \<lambda>(x::(nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c)) xa::(nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c).
        (case x of (u::nat \<Rightarrow> 'a, xa::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) \<Rightarrow> \<lambda>(u'::nat \<Rightarrow> 'a, x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c). r (u (0::nat)) (u' (Suc (0::nat))) (xa (0::nat)) (y' (0::nat))) xa \<and>
        (case x of (u::nat \<Rightarrow> 'a, xa::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) \<Rightarrow> \<lambda>(u'::nat \<Rightarrow> 'a, ab::(nat \<Rightarrow> 'b) \<times> (nat \<Rightarrow> 'c)). u' (0::nat) = u (0::nat) \<and> (case ab of (x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c) \<Rightarrow> x' = xa)) xa)
        = (\<lambda>(u::nat \<Rightarrow> 'a, x::nat \<Rightarrow> 'b, y::nat \<Rightarrow> 'c) (u'::nat \<Rightarrow> 'a, x'::nat \<Rightarrow> 'b, y'::nat \<Rightarrow> 'c). r (u (0::nat)) (u' (Suc (0::nat))) (x (0::nat)) (y' (0::nat)) \<and> x = x' \<and> u (0::nat) = u' (0::nat))")
    apply simp
    by (auto simp add: fun_eq_iff)
  finally show "?A \<parallel> [:u, x, y \<leadsto> u', x', y' . u' 0 = u 0 \<and> x' = x :] = ?R" by simp
  qed

definition "DelayFeedback init S = [:x \<leadsto> u, x',y . init (u (0::nat)) \<and> x = x':]
         o IterateOmegaSkipNextMask (AddUnitDelay S) o  [:u, x, y \<leadsto> y' . y = y':]"
    
lemma SkipNext_refinement: "S \<le> T \<Longrightarrow> SkipNext S \<le> SkipNext T"
  apply (simp add: SkipNext_def)
  by (simp add: le_fun_def demonic_def Prod_def prod_pred_def, auto)
    
lemma SkipNext_pow_refinement: "S \<le> T \<Longrightarrow> (SkipNext ^^ n) S \<le> (SkipNext ^^ n) T"
  proof (induction n)
    case 0
    then show ?case by simp
  next
    case (Suc n)
    then show ?case
      apply simp
      by (rule SkipNext_refinement, simp)
  qed
    
lemma Mask_refinement: "S \<le> T \<Longrightarrow> Mask i S \<le> Mask i T"
  by (simp add: Mask_def le_fun_def)

lemma mono_SkipNext[simp]: "mono (SkipNext S)"
  by (simp add: SkipNext_def)

lemma mono_SkipNext_pow [simp]: "mono S \<Longrightarrow> mono ((SkipNext ^^ n) S)"
  by (induction n, simp_all)
  
lemma mono_Iterate_SkipNext[simp]: "mono S \<Longrightarrow> mono (Iterate SkipNext S n)"
  by (induction n, simp_all del: comp_apply)
    
lemma  Iterate_SkipNext_refinement: "\<And> S T . mono S \<Longrightarrow> S \<le> T \<Longrightarrow> Iterate SkipNext S n \<le> Iterate SkipNext T n"
proof (induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  have A: "Iterate SkipNext S (Suc n) \<le> Iterate SkipNext T (Suc n)"
    apply simp
    apply (rule mono_comp)
      apply (rule mono_Iterate_SkipNext)
      apply (simp add: Suc.prems(1))
     apply (simp add: Suc.IH Suc.prems(1) Suc.prems(2))
    using SkipNext_pow_refinement Suc.prems(2) by auto
  from this show ?case
    by (simp add: le_fun_def)
qed

    
lemma IterateSkipNextMask_refinemnt: "mono S \<Longrightarrow> S \<le> T \<Longrightarrow> IterateSkipNextMask S i \<le> IterateSkipNextMask T i"
  apply (simp add: IterateSkipNextMask_def)
  apply (rule Mask_refinement)
  apply (rule Iterate_SkipNext_refinement)
    by simp_all
  
lemma IterateOmegaSkipNextMask_refinement: "mono S \<Longrightarrow> S \<le> T \<Longrightarrow> IterateOmegaSkipNextMask S \<le> IterateOmegaSkipNextMask T"
  apply (simp add: IterateOmegaSkipNextMask_def)
  apply (rule Fusion_refinement)
  by (rule IterateSkipNextMask_refinemnt, simp)
    
lemma AddUnitDelay_refinement: "S \<le> T \<Longrightarrow> AddUnitDelay S \<le> AddUnitDelay T"
  apply (simp add: AddUnitDelay_def)
  apply (simp add: le_fun_def demonic_def fusion_def, safe)
    apply (rule_tac x = p in exI)
  apply (rule_tac x = p' in exI)
  by auto
    
lemma mono_IterateOmegaSkipNextMask: "mono (IterateOmegaSkipNextMask S)"
  by (simp add: IterateOmegaSkipNextMask_def mono_Fusion)
    
lemma mono_AddUnitDelay: "mono (AddUnitDelay S)"
  by (simp add: AddUnitDelay_def)
    
theorem DelayFeedback_refinement: "init' \<le> init \<Longrightarrow> S \<le> T \<Longrightarrow> DelayFeedback init S \<le> DelayFeedback init' T"
  apply (simp add: DelayFeedback_def)
  apply (rule mono_comp)
    apply (rule mono_comp_a, simp_all)
    apply (rule mono_IterateOmegaSkipNextMask)
  apply (rule mono_comp, simp_all)
   apply (rule demonic_refinement)
   apply (simp add: le_fun_def)
  apply (rule IterateOmegaSkipNextMask_refinement)
   apply (simp add: mono_AddUnitDelay)
  apply (rule AddUnitDelay_refinement)
    by simp
    
lemma [simp]: "mono (SkipTop n)"
  by (simp add: SkipTop_def)

lemma [simp]: "SkipNext Skip = Skip"
  apply (simp add: SkipNext_def demonic_demonic)
  by (simp add: fun_eq_iff Skip_def OO_def demonic_def le_fun_def)


lemma Iterate_SkipNextA: "mono S \<Longrightarrow> S \<circ> (SkipNext (Iterate SkipNext S n)) = Iterate SkipNext S (Suc n)"
  apply (induction n, simp)
  apply (simp del: comp_apply)
  apply (simp add: SkipNext_comp del: comp_apply)
  apply (simp add: comp_assoc [THEN sym] del: comp_apply)
  by simp

lemma skiptop_simp: "SkipTop n p = (\<lambda> x . \<forall> y . eqtop n x y \<longrightarrow> p y)"
  by (simp add: SkipTop_def demonic_def fun_eq_iff le_fun_def)

definition "HavocTop n = [:x \<leadsto> y . x[n..] = y[n..]:]"
  
  
lemma HavocTop_Next: "HavocTop (Suc n) = Next (HavocTop n)"
  apply (simp add: HavocTop_def Next_def demonic_demonic)
  apply (subgoal_tac "( \<lambda>(x::'a) y. x[Suc n ..] = y[Suc n ..]) = ( ((\<lambda>x y. y = x[Suc 0 ..]) OO (\<lambda>x y. x[n ..] = y[n ..])) OO (\<lambda>y x. y = x[Suc 0 ..])  )")
  apply simp_all
  apply (simp add: fun_eq_iff, safe)
  apply (rule relcomppI, simp_all)
  apply (rule relcomppI, simp_all)
  by auto

lemma [simp]: "HavocTop 0 = Skip"
  by (simp add: HavocTop_def demonic_eq_skip)

lemma "HavocTop n = (Next ^^ n) Skip"
  by (induction n, simp_all add: HavocTop_Next)

lemma Next_NextSkip_aux: "[: \<lambda>y x. \<forall>xa. y xa = x (Suc xa) :] (\<lambda>a. \<forall>b. a[Suc 0 ..] = b[Suc 0 ..] \<longrightarrow> x b) = [: \<lambda>y x. \<forall>xa. y xa = x (Suc xa) :] x"
  by (auto simp add: fun_eq_iff demonic_def le_fun_def)
 
lemma demonic_apply_pred: "[: \<lambda>x y. r x y :] p  = (\<lambda>x. \<forall> y . r x y \<longrightarrow> p y)"
  by (simp add: fun_eq_iff demonic_def le_fun_def)

  lemma Next_SkipNext_HavocTop: "mono S \<Longrightarrow> Next S = SkipNext S o HavocTop (Suc 0)"
    proof (simp add:  SkipNext_Next SkipTop_def Next_def fun_eq_iff fusion_def HavocTop_def demonic_def le_fun_def demonic_apply_pred, safe)
      fix x xa
      assume [simp]: "S (\<lambda>y. \<forall>xa. y = xa[Suc 0 ..] \<longrightarrow> x xa) (xa[Suc 0 ..])"
      have [simp]: " (\<lambda>y. \<forall>xa. y = xa[Suc 0 ..] \<longrightarrow> (\<forall>b. b[Suc 0 ..] = xa[Suc 0 ..] \<longrightarrow> x b)) = (\<lambda>y. \<forall>xa. y = xa[Suc 0 ..] \<longrightarrow> x xa)"
        by (auto simp add: fun_eq_iff)
      show " \<exists>p p'. (\<forall>xa. p xa \<and> p' xa \<longrightarrow> (\<forall>y. xa[Suc 0 ..] = y[Suc 0 ..] \<longrightarrow> x y)) \<and>
                   S (\<lambda>y. \<forall>x. y = x[Suc 0 ..] \<longrightarrow> p x) (xa[Suc 0 ..]) \<and> (\<forall>y. eqtop (Suc 0) xa y \<longrightarrow> p' y)"
        apply (rule_tac x = "(\<lambda> a . \<forall> b .  b[Suc 0..] = a[Suc 0..] \<longrightarrow> x b)" in exI)
        by auto
   next
     fix x::"'a \<Rightarrow> bool" fix xa::'a fix p p':: "'a \<Rightarrow> bool"
     assume [simp]: "S (\<lambda>y. \<forall>x. y = x[Suc 0 ..] \<longrightarrow> p x) (xa[Suc 0 ..])"
     assume A: "\<forall>xa. p xa \<and> p' xa \<longrightarrow> (\<forall>y. xa[Suc 0 ..] = y[Suc 0 ..] \<longrightarrow> x y)"
     assume B: "\<forall>y. eqtop (Suc 0) xa y \<longrightarrow> p' y"
     from A and B have C: "\<And> xa . (\<forall>x. xa[Suc 0 ..] = x[Suc 0 ..] \<longrightarrow> p x) \<longrightarrow> x xa"
      apply safe
      apply (drule_tac x = "cat (Suc 0) xa (xaa[Suc 0..])" in spec)
      by auto 
     assume "mono S"
     from this have "S (\<lambda>y. \<forall>x. y = x[Suc 0 ..] \<longrightarrow> p x) \<le>  S (\<lambda>y. \<forall>xa. y = xa[Suc 0 ..] \<longrightarrow> x xa)"
       apply (rule monoD, auto)
       by (cut_tac C, auto)
     from this show "S (\<lambda>y. \<forall>xa. y = xa[Suc 0 ..] \<longrightarrow> x xa) (xa[Suc 0 ..])"
       by (simp add: le_fun_def)
    qed
 
  lemma HavocTop_Next_power: "HavocTop n \<circ> Next ((Next ^^ n) S) =  Next ((Next ^^ n) S)"
    proof (induction n)
      case 0
      show ?case by simp
      case (Suc n)
      from this have [simp]: "HavocTop n \<circ> Next ((Next ^^ n) S) =  Next ((Next ^^ n) S)" by (simp del: comp_apply)
      have [simp]: "(Next \<circ> Next ^^ n) S =  Next ((Next ^^ n) S)" by simp
      show ?case
        by (simp_all add: HavocTop_Next Next_comp [THEN sym] Suc del: comp_apply)
    qed

  lemma Next_SkipNext: "mono S \<Longrightarrow> (Next ^^ n) S = (SkipNext ^^ n) S \<circ> HavocTop n" (is "?Q \<Longrightarrow> ?A n = ?B n")
    proof (induction n)
      case 0 show ?case by simp
      case (Suc n)
        from this
        have [simp]: "mono S" and [simp]: "?A n = ?B n" by simp_all
        have "(Next ^^ Suc n) S = Next  ((Next ^^ n) S)" by simp
        also have "... = Next  ((SkipNext ^^ n) S \<circ> HavocTop n)" by (simp del: comp_apply)
        also have "... = SkipNext ((SkipNext ^^ n) S) \<circ> HavocTop (Suc (0::nat)) \<circ> Next (HavocTop n)"  by (simp add: Next_comp Next_SkipNext_HavocTop)
        also have "... = ?B (Suc n)"  by (simp add: HavocTop_Next Next_comp [THEN sym] comp_assoc del: comp_apply, simp add: fun_eq_iff)
        finally show ?case by simp
    qed

  lemma Iterate_Next_SkipNext_aux: "mono S \<Longrightarrow> HavocTop n \<circ> (Next ^^ (Suc n)) S = (SkipNext ^^ (Suc n)) S \<circ> HavocTop (Suc n)" (is "?P \<Longrightarrow> ?A = ?B")
    proof -
      assume [simp]: "?P"
      have "?A =  (Next ^^ (Suc n)) S" by (simp add: HavocTop_Next_power)
      also have "... = ?B" by (subst Next_SkipNext, auto)
      finally show ?thesis by simp
    qed

  lemma Iterate_Next_SkipNext_Suc: "mono S \<Longrightarrow> Iterate Next S (Suc n) = (Iterate SkipNext S (Suc n)) o (HavocTop n)" (is "?P \<Longrightarrow> ?A n = ?B n")
    proof (induction n)
      case 0 show ?case by simp
      case (Suc n)
      from this have
        [simp]: "HavocTop n \<circ> (Next o (Next ^^ n)) S = SkipNext ((SkipNext ^^ n) S) \<circ> HavocTop (Suc n)" 
        and [simp]: " Iterate Next S n \<circ> (Next ^^ n) S = ?B n" 
          by (cut_tac Iterate_Next_SkipNext_aux, simp, simp_all del: comp_apply)
        show ?case 
          by (simp_all add: comp_assoc del: comp_apply, simp)
   qed

  lemma Iterate_Next_SkipNext: "mono S \<Longrightarrow> Iterate Next S n = (Iterate SkipNext S n) o (HavocTop (n - 1))"
    by (case_tac n, simp_all, cut_tac Iterate_Next_SkipNext_Suc, simp_all)

  lemma "HavocTop n \<le> Skip"
    by (simp add: HavocTop_def Skip_def le_fun_def demonic_def)

  lemma  mono_Iterate_NextSkip: "mono S \<Longrightarrow> mono (Iterate SkipNext S n)"
    by (induction n, simp_all)

  lemma "(Havoc (X::'a::complete_lattice) \<noteq> \<bottom>) = (X = \<top>)"
    apply (simp add: Havoc_def fun_eq_iff demonic_def le_fun_def top_fun_def, safe)
    by (rule antisym, simp_all)

  type_synonym ('a, 'b) rel = "('a \<Rightarrow> 'b \<Rightarrow> bool)"

  primrec IterateRel :: "(('a, 'a) rel \<Rightarrow> ('a, 'a) rel) \<Rightarrow> ('a, 'a) rel \<Rightarrow> nat \<Rightarrow> ('a, 'a) rel" where
    "IterateRel F r 0 = (\<lambda> a b . a = b)" |
    "IterateRel F r (Suc n) = IterateRel F r n OO ((F ^^ n) r)"

lemma IterateRel_init: "(\<forall> r r' . F (r OO r') = F r OO F r') \<Longrightarrow> F (op =) = (op =) \<Longrightarrow> IterateRel F r (Suc n) = r OO F (IterateRel F r n)"  (is "?P \<Longrightarrow> ?Q \<Longrightarrow> ?R n")
proof (induction n)
  case 0 from this show ?case by (simp add: eq_OO OO_eq) 
  case (Suc n)
    assume [simp]: "?P"
    assume [simp]: "?Q"
    assume "?P \<Longrightarrow> ?Q \<Longrightarrow> ?R n"
    from this have A: "?R n" by simp
    have "IterateRel F r (Suc (Suc n)) = IterateRel F r (Suc n) OO (F ^^ (Suc n)) r"
      by simp
    also have "... = r OO F (IterateRel F r (Suc n))"
      by (subst A , simp add: relcompp_assoc)
    finally show "?R (Suc n)" by simp
qed

lemma [simp]: "idnext (op =) = (op =)"
  apply (simp add: idnext_def fun_eq_iff next_def at_fun_def, safe, simp_all)
  by (drule eqtop_tail, simp_all)

lemma [simp]: "idnext (r OO r') = (idnext r) OO idnext r'"
  apply (auto simp add: idnext_def fun_eq_iff next_def at_fun_def OO_def)
  apply (rule_tac x = "cat (Suc 0) x y" in exI, simp)
  by (rule_tac b = y in eqtop_tran, simp)

lemma IterateRel_idnext_init: "IterateRel idnext r (Suc n) = r OO idnext (IterateRel idnext r n)"
  by (rule IterateRel_init, simp_all)

lemma [simp]: "(\<And> (p::'a\<Rightarrow>bool) (r::'a\<Rightarrow>'b\<Rightarrow>bool) . F ({.p.} o [:r:]) = {.A p.} o [:(B::('a\<Rightarrow>'b\<Rightarrow>bool) \<Rightarrow> ('a\<Rightarrow>'b\<Rightarrow>bool)) r:]) 
  \<Longrightarrow> ((F ^^ n) ({.p.} o [:r:])) = {.(A^^n) p.} o [:(B^^n) r:]"
  by (induction n, auto)

lemma Iterate_id: "Iterate id S n = S ^^ n"
  apply (induction n, simp_all)
  by (metis comp_id fun.map_ident id_funpow funpow_swap1)

lemma IterateRel_id: "IterateRel id r n = (r ^^ n)"
  apply (induction n)
  apply (simp_all)
  by (metis comp_id fun.map_ident id_funpow)

lemma Iterate_IterateRel: "(\<And> p r . F ({.p.} o [:r:]) = {.A p.} o [:B r:]) \<Longrightarrow> Iterate F ({.p.} o [:r:]) n 
    = {.x . (\<forall> i < n .(\<forall> y . IterateRel B r i x y \<longrightarrow> (A^^i) p y)).} o [:IterateRel B r n:]"
  proof (induction n)
   case 0
    show ?case by (simp add: demonic_eq_skip assert_true_skip_a)
  next
    case (Suc n)
      assume [simp]: "\<And>p r. F ({. p .} \<circ> [: r :]) = {. A p .} \<circ> [: B r :]"
      assume "((\<And>p r. F ({. p .} \<circ> [: r :]) = {. A p .} \<circ> [: B r :]) \<Longrightarrow>
        Iterate F ({. p .} \<circ> [: r :]) n = {.x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} \<circ> [: IterateRel B r n :])"
      from this have [simp]: "Iterate F ({. p .} \<circ> [: r :]) n = {.x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} \<circ> [: IterateRel B r n :]"
        by simp
      have [simp]: " (\<lambda>x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y) \<sqinter> (\<lambda>x. \<forall>y. IterateRel B r n x y \<longrightarrow> (A ^^ n) p y)
            =  (\<lambda>x. \<forall>i < Suc n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y)"
         by (simp add: fun_eq_iff, auto, case_tac "i = n", auto)
      have "Iterate F ({. p .} \<circ> [: r :]) (Suc n) = {.x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} \<circ> [: IterateRel B r n :] o ({. (A ^^ n) p .} \<circ> [: (B ^^ n)r :])"
        by simp
      also have "... = {.x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} \<circ> ([: IterateRel B r n :] o {. (A ^^ n) p .}) \<circ> [: (B ^^ n)r :]"
        by (simp add: comp_assoc)
      also have "... = {.x. \<forall>i<n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} o {.x. \<forall>y. IterateRel B r n x y \<longrightarrow> (A ^^ n) p y.} o [: IterateRel B r n :] \<circ> [: (B ^^ n)r :]"
        by (simp add: demonic_assert_comp, simp add: comp_assoc)
      also have "... = {.x. \<forall>i < Suc n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} o [: IterateRel B r n :] \<circ> [: (B ^^ n)r :]"
        by (simp add: assert_assert_comp)
      also have "... =  {.x. \<forall>i < Suc n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} o [: IterateRel B r n OO (B ^^ n)r :]"
        by (simp add: comp_assoc demonic_demonic)
      also have "... =  {.x. \<forall>i < Suc n. \<forall>y. IterateRel B r i x y \<longrightarrow> (A ^^ i) p y.} o [: IterateRel B r (Suc n) :]"
        by simp
      finally show ?case by (simp del: comp_apply)
     qed

lemma IterateRel_app: "\<And> y . IterateRel next r n x y = (\<exists> a . a 0 = x \<and> a n = y \<and> (\<forall> i < n. r ((a i)[i..]) ((a (Suc i))[i..])))"
  apply (induction n)
  apply (simp)
  apply safe [1]
  apply (rule_tac x = "\<lambda> n . x" in exI, simp)
  apply simp
  apply safe
  apply simp
  apply safe
  apply (simp add: iterate_next at_fun_def)
  apply (rule_tac x = "\<lambda> i . if i \<le> n then a i else y" in exI)
  apply simp
  apply auto [1]
  apply (case_tac "i = n", simp_all)
  apply (rule relcomppI)
  apply simp_all
  apply (rule_tac x = a in exI, simp_all)
    by (simp add: iterate_next at_fun_def)

lemma Iterate_Next_IterateRel: "Iterate Next ({.p.} o [:r:]) n 
  = {. x . (\<forall> k < n . (\<forall> y . IterateRel next r k x y \<longrightarrow> (next ^^ k) p y) ).} o [:IterateRel next r n:]"
  apply (rule Iterate_IterateRel)
  by (simp add: Next_assert_demonic)

lemma IterateOmegaNextMask_spec_aux: "IterateOmegaNextMask ({.p.} o [:r:]) 
  = {. INF x. (\<lambda>xa. \<forall>k<x. \<forall>y. IterateRel next r k xa y \<longrightarrow> (next ^^ k) p y) .} \<circ> [: INF n. IterateRel next r n OO eqtop n :]"
  apply (simp add: IterateOmegaNextMask_def IterateNextMask_simp Mask_def Iterate_Next_IterateRel SkipTop_def comp_assoc demonic_demonic)
  by (simp add: Fusion_spec)

lemma IterateOmegaNextMask_spec: "IterateOmegaNextMask ({.p.} o [:r:]) 
        =  {. INF k . (\<lambda>xa. \<forall>y. IterateRel next r k xa y \<longrightarrow> (next ^^ k) p y) .} \<circ> [: INF n. IterateRel next r n OO eqtop n :]"
  apply (simp add: IterateOmegaNextMask_spec_aux)
  apply (subgoal_tac "(INF x. (\<lambda>xa. \<forall>k<x. \<forall>y. IterateRel next r k xa y \<longrightarrow> (next ^^ k) p y)) = (INF k. (\<lambda>xa. \<forall>y. IterateRel next r k xa y \<longrightarrow> (next ^^ k) p y))")
  by auto

lemma power_spec: "({.p.} o [:r:]) ^^ n 
    = {.x . (\<forall> i < n .(\<forall> y . (r ^^ i) x y \<longrightarrow> p y)).} o [:r ^^ n:]"
  apply (cut_tac A = id and B = id and F = id in Iterate_IterateRel)
  apply simp
  by (simp add: Iterate_id IterateRel_id)


lemma Iterate_SkipNext_IterateSkipRel: "Iterate SkipNext ({.p.} o [:r:]) n 
      = {. x . (\<forall> k < n . (\<forall> y . IterateRel idnext r k x y \<longrightarrow> (next ^^ k) p y) ).} o [:IterateRel idnext r n:]"
  apply (rule Iterate_IterateRel)
  by (simp add: SkipNext_assert_demonic)

lemma IterateOmegaSkipNextMask_spec: "IterateOmegaSkipNextMask ({.p.} o [:r:]) 
  =  {. (\<lambda>x . \<forall> n . \<forall>y. IterateRel idnext r n x y \<longrightarrow> (next ^^ n) p y) .} 
     \<circ> [: INF n. IterateRel idnext r n OO eqtop n :]"
  apply (simp add: IterateOmegaSkipNextMask_def IterateSkipNextMask_simp Mask_def 
      Iterate_SkipNext_IterateSkipRel SkipTop_def comp_assoc demonic_demonic)
  apply (simp add: Fusion_spec)
  apply (subgoal_tac "(INF x. (\<lambda>xa. \<forall>k<x. \<forall>y. IterateRel idnext r k xa y \<longrightarrow> (next ^^ k) p y)) 
         = (\<lambda>x . \<forall>n y. IterateRel idnext r n x y \<longrightarrow> (next ^^ n) p y)")
  by auto

  

lemma IterateOmegaSkipNextMask_demonic: "IterateOmegaSkipNextMask [:r:] =  [: INF n. IterateRel idnext r n OO eqtop n :]"
  by (cut_tac p = \<top> and r = r in IterateOmegaSkipNextMask_spec, simp add: assert_true_skip assert_true_skip_a)

lemma [simp]: "(next ^^ n) \<top> x"
  by (simp add: iterate_next)

lemma power_idnext: "(idnext ^^ n) r = ((next ^^ n) r \<sqinter> eqtop n)"
proof (induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case
    apply (simp add: fun_eq_iff idnext_def next_def at_fun_def, safe)
    apply (cut_tac n = "Suc 0" and m = n and x = x and y = xa in  eqtop_plus, simp_all)
     apply (cut_tac n = "Suc 0" and m = n and x = x and y = xa in  eqtop_plus, simp_all)
    by (cut_tac n = "Suc 0" and m = n and x = x and y = xa in  eqtop_plus, simp_all)
qed

lemma example_feedback_delay_a: " \<forall>xb. \<exists>z. IterateRel idnext (\<lambda>x y. \<forall>xa. y xa = ([0] .. x) xa) xb x z \<and> (\<forall>i<xb. z i = xa i) \<Longrightarrow> xa n = 0"
    apply (induction n)
    apply (drule_tac x = 1 in spec)
    apply auto [1]
    apply (drule_tac x = "Suc (Suc n)" in spec)
    apply simp
    apply auto
    apply (simp add: idnext_def next_def at_fun_def append_inf_def power_idnext iterate_next)
    apply auto
    apply (drule_tac x = "Suc n" in spec, simp)
    by (drule_tac x = "0" in spec, simp)

lemma example_feedback_delay_b: "\<forall>x. xa x = 0 \<Longrightarrow> \<exists>z. IterateRel idnext (\<lambda>x y. \<forall>xa. y xa = ([0] .. x) xa) n x z \<and> (\<forall>i < n. z i = xa i)"
  apply (induction n)
  apply auto [1]
  apply simp
  apply safe
  apply (rule_tac x = "[0] .. z" in exI, safe)
  apply (simp add: OO_def)
  apply (rule_tac x= z in exI, simp_all)
  apply (simp add: idnext_def next_def at_fun_def append_inf_def power_idnext iterate_next)
  by (simp add: append_inf_def)

lemma example_feedback_delay: "IterateOmegaSkipNextMask [:x \<leadsto> y . y = [0::nat] .. x:] = [:x\<leadsto> y . y = (\<lambda> i . 0):]"
  apply (cut_tac p = \<top> and r = "\<lambda> x  y . y = [0::nat] .. x" in IterateOmegaSkipNextMask_spec, (simp add: assert_true_skip_a))
  apply (subgoal_tac "( INF x::nat. IterateRel idnext (\<lambda>(x::nat \<Rightarrow> nat) y::nat \<Rightarrow> nat. y = [0::nat] .. x) x OO eqtop x) 
          = (\<lambda>(x::nat \<Rightarrow> nat) y::nat \<Rightarrow> nat. y = (\<lambda>i::nat. 0::nat))", simp_all)
  apply (simp_all add: fun_eq_iff OO_def assert_true_skip, safe)
  apply (drule example_feedback_delay_a, simp)
  apply (rule example_feedback_delay_b)
  by simp

lemma next_simp: "next (r::(nat \<Rightarrow> 'a) \<Rightarrow>(nat \<Rightarrow> 'b) \<Rightarrow>bool) x y = r (x[Suc 0..]) (y[Suc 0..])"
  by (simp add: fun_eq_iff next_def at_fun_def)

lemma idnext_simp: "idnext (r::(nat \<Rightarrow> 'a) \<Rightarrow>(nat \<Rightarrow> 'a) \<Rightarrow>bool) x y = (r (x[Suc 0..]) (y[Suc 0..]) \<and> x 0 = y 0)"
  by (simp add: idnext_def fun_eq_iff next_def at_fun_def)

lemma idnext_next_eqtop: "\<And> (x::nat\<Rightarrow>'a) y . (idnext ^^ n) r x y = ((next ^^ n) r x y \<and> eqtop n x y)"
  apply (induction n)
  apply simp
  apply simp
  apply (subst next_simp)
  apply (subst idnext_simp)
  apply safe
  apply metis
  apply (metis AbsNat_nat_def add_Suc at_trace_def id_apply less_Suc_eq_0_disj monoid_add_class.add.left_neutral)
  apply auto [1]
  by auto

lemma IrerateRel_IterateSkipRel_aux: "\<forall> x y . IterateRel next (r::(nat \<Rightarrow> 'a) \<Rightarrow>(nat \<Rightarrow> 'a) \<Rightarrow>bool) n x y \<longrightarrow> (\<exists> z . y[(n::nat)..] = z[n..] \<and> IterateRel idnext r n x z)"
  apply (induction n)
  apply simp
  apply simp
  apply (simp add: OO_def idnext_next_eqtop, safe)
  apply (drule_tac x = x in spec)
  apply (drule_tac x = ya in spec)
  apply simp
  apply safe
  apply (rule_tac x = "\<lambda>i . (if i < n then z i else y i)" in exI, simp)
  apply (rule_tac x = z in exI, simp)
  apply (simp add: iterate_next at_fun_def)
  apply (subgoal_tac "(ya[n ..]) = (z[n ..])", simp)
  apply (subgoal_tac "(y[n ..]) = ((\<lambda>i. if i < n then z i else y i)[n ..])")
  apply simp_all
  by (simp add: fun_eq_iff)

lemma IrerateRel_IterateSkipRel: "IterateRel next (r::(nat \<Rightarrow> 'a) \<Rightarrow>(nat \<Rightarrow> 'a) \<Rightarrow>bool) n x y \<Longrightarrow> (\<exists> z . y[(n::nat)..] = z[n..] \<and> IterateRel idnext r n x z)"
  by (cut_tac IrerateRel_IterateSkipRel_aux, auto)

(*
definition "prec_loc_st p r = -((lft_pred_st (inpt_st r)) until -(lft_pred_loc_st p))"
  

lemma prec_loc_st_simp: "prec_loc_st p r u x = (\<forall> n . (\<forall> i < n . inpt_st r (u i) (u (Suc i)) (x i)) \<longrightarrow> p (u n) (x n))"
  by (simp add: prec_loc_st_def until_def lft_pred_st_def inpt_st_def at_fun_def lft_pred_loc_st_def, metis)
    
*)

lemma next_eq: "\<forall>i<k. (\<forall>x. fst (snd (ab i)) (i + x) = fst (snd (ab (Suc i))) (i + x)) \<Longrightarrow>
     i \<le> k \<Longrightarrow>  (\<forall> j . fst (snd (ab i)) (i + j) =  fst (snd (ab 0)) (i + j))"
  apply (induction i)
  apply simp_all
  apply safe
  by (metis add_Suc_right)

(*
  definition "SymbolicSystem init p r = {.x . \<forall> u . init (u 0) \<longrightarrow> prec_loc_st p r u x.} o [:x \<leadsto> y . \<exists> u . init (u 0) \<and> (\<box> (lft_rel_st r)) u x y:]"
*)
  lemma IterateSkipRel_SymRel_zero: "\<And> u' x' y' . (IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u 0) (u' (Suc 0)) (x 0) (y' 0) \<and> (x = x') \<and> (u 0 = u' 0)) 0) (u, x, y) (u', x', y') 
    = (u = u' \<and> x = x' \<and> y = y')"
    by auto

  lemma IterateSkipRel_SymRel_Suc: "\<And> u' x' y' . (IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u 0) (u' (Suc 0)) (x 0) (y' 0) \<and> (x = x') \<and> (u 0 = u' 0)) (Suc n)) (u, x, y) (u', x', y') 
    = ((u' 0 = u 0) \<and> (\<forall> i < (Suc n) . r (u' i) (u' (Suc i)) (x i) (y' i)) \<and> x = x')"
    apply (induction n)
    apply auto [1]
    apply (simp add: idnext_def next_def at_fun_def append_inf_def power_idnext iterate_next OO_def eqtop_prod_def at_prod_def)
    apply safe
    apply auto [1]
    apply (case_tac "i < Suc n", simp_all)
    apply (drule_tac x = i in spec, simp)
    apply (subgoal_tac "a i = u' i \<and> a (Suc i) = u' (Suc i) \<and> b i = y' i", simp_all)
    apply (case_tac i, simp_all)
    apply auto [1]
    apply auto [1]
    apply (case_tac "Suc nat < n")
    apply auto [1]
    apply (case_tac "Suc nat = n")
    apply auto [1]
    apply auto [1]
    apply (case_tac "i = Suc n")
    apply auto [1]
    apply auto [1]
    apply (simp add: fun_eq_iff at_fun_def, safe)
    apply (case_tac xa, simp_all)
    apply (case_tac "nat < n", simp_all)
    apply (metis le_add_diff_inverse not_less)
    apply (rule_tac x = u' in exI, simp_all)
    by (rule_tac x = y' in exI, simp_all)

  lemma IterateSkipRel_SymRel: "\<And> u' x' y' . (IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u 0) (u' (Suc 0)) (x 0) (y' 0) \<and> (x = x') \<and> (u 0 = u' 0)) n) (u, x, y) (u', x', y') 
    = ((u' 0 = u 0) \<and> (\<forall> i <  n . r (u' i) (u' (Suc i)) (x i) (y' i)) \<and> x = x' \<and> (n = 0 \<longrightarrow> (u = u' \<and> y = y')))"
    apply (case_tac n)
    apply (cut_tac IterateSkipRel_SymRel_zero)
    apply auto [1]
    apply (cut_tac IterateSkipRel_SymRel_Suc)
    by auto

  lemma IterateSkipRel_SymRel_eqtop: "(IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u (0::nat)) (u' (Suc 0)) (x (0::nat)) (y' (0::nat)) \<and> (x = x') \<and> (u 0 = u' 0)) n OO (eqtop n)) (u, x, y) (u', x', y') 
    = (\<exists> v .  (v 0 = u 0 ) \<and> (\<forall> i < n . r (v i) (v (Suc i)) (x i) (y' i) \<and> v i = u' i \<and> (x i = x' i)))"
    apply (case_tac n)
    apply auto [1]
    apply (unfold OO_def)
    apply safe
    apply (unfold IterateSkipRel_SymRel)
    apply (rule_tac x = a in exI, simp)
    apply (simp add: eqtop_prod_def at_prod_def)
    apply auto [1]
    apply (rule_tac x = "(v, x, y')" in exI)
    apply (unfold IterateSkipRel_SymRel)
    apply (simp add: eqtop_prod_def at_prod_def)
    by auto [1]


  lemma INF_IterateSkipRel_SymRel_eqtop: "(INF n. IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u (0::nat)) (u' (Suc 0)) (x (0::nat)) (y' (0::nat)) \<and> x = x' \<and> u 0 = u' 0) n OO eqtop n) (u, x, y) (u', x', y')
       = (u' 0 = u 0 \<and> x = x' \<and> (\<box> (\<lambda> (u, x, y) . r (u 0) (u (Suc 0)) (x 0) (y 0))) (u', x, y'))"
    apply (simp add: Inf_fun_def IterateSkipRel_SymRel_eqtop) 
    apply safe
    apply (metis zero_less_Suc)
    apply (metis ext lessI)
    apply (simp add: always_def at_fun_def at_prod_def)
    apply (metis Suc_lessD lessI)
    apply (simp add: always_def at_fun_def at_prod_def)
    by (rule_tac x = "u'" in exI, simp)

  lemma INF_IterateSkipRel_SymRel_eqtop_abs: "(INF n. IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u (0::nat)) (u' (Suc 0)) (x (0::nat)) (y' (0::nat)) \<and> x = x' \<and> u 0 = u' 0) n OO eqtop n)
       = (\<lambda> (u, x, y) (u', x', y') . (u' 0 = u 0 \<and> x = x' \<and> (\<box> (\<lambda> (u, x, y) . r (u 0) (u (Suc 0)) (x 0) (y 0))) (u', x, y')))"
    
    apply (subst (2) fun_eq_iff)
    apply (subst (2) fun_eq_iff)
    apply (rule allI)
    apply (rule allI)
    apply (case_tac x)
    apply (case_tac xa)
    by (cut_tac INF_IterateSkipRel_SymRel_eqtop, auto)

  lemma move_down: "p \<Longrightarrow> p"
    by simp

  lemma IterateSkipRel_prec_loc_st: "(\<lambda>x. \<forall>a. init (a 0) \<longrightarrow>
                (\<forall>b n aa aaa ba.
                    IterateRel idnext (\<lambda>(u, x, y) (u', x', y'). r (u (0::nat)) (u' (Suc 0)) (x (0::nat)) (y' (0::nat)) \<and> x = x' \<and> u 0 = u' 0) n (a, x, b) (aa, aaa, ba) \<longrightarrow>
                    (next ^^ n) (\<lambda>(u, x, y). p (u 0) (x 0)) (aa, aaa, ba)))
         = prec_pre_sts init (\<lambda> (u, x) . p u x) (\<lambda> (u,x) (u', y) . r u u' x y)"
  apply (simp add: prec_pre_sts_simp iterate_next at_fun_def at_prod_def IterateSkipRel_SymRel)
  apply (simp add: fun_eq_iff)
    apply (safe)
      apply (unfold prec_pre_sts_simp, safe)
  apply (drule_tac x = "u" in spec, simp)
  apply (drule move_down)
  apply (drule move_down)
  apply (drule_tac x = "y" in spec)
  apply (drule move_down)
  apply (drule move_down)
  apply (drule_tac x = n in spec)
  apply (drule_tac x = u in spec)
  apply (drule_tac x = x in spec)
  apply safe
  apply (case_tac n, simp)
  apply auto [1]
  apply simp
  apply (rule_tac x = "\<lambda> i . SOME y . r (u i) (u (Suc i)) (x i) y" in exI)
  apply safe [1]
  apply (drule_tac x = i in spec, simp)
  apply (metis someI)
  apply (drule_tac x = aa in spec)
  apply metis
  by metis

    
theorem DelayFeedback_SymbolicSystem_aux: "DelayFeedback init ({.(x, y).p x y.} \<circ> [:(u, x)\<leadsto>(u', y).r u u' x y:])  
      = LocalSystem init (\<lambda> (u, x) . p u x) (\<lambda> (u,x) (u', y) . r u u' x y)"
    apply (unfold  AddUnitDelay_spec DelayFeedback_def)
    apply (simp add: IterateOmegaSkipNextMask_spec comp_assoc [THEN sym] INF_IterateSkipRel_SymRel_eqtop_abs)
    apply (simp add: demonic_assert_comp)
    apply (simp add: comp_assoc demonic_demonic)
  apply (simp add: LocalSystem_simp IterateSkipRel_prec_loc_st)
    apply (rule prec_rel_eq, simp)
  apply (simp add: fun_eq_iff always_def rel_pre_sts_simp at_fun_def at_prod_def OO_def)
  apply safe
   apply (rule_tac x = a in exI)
   apply auto [1]
   by (rule_tac x = u in exI, auto)

(*it was DelayFeedback_SymbolicSystem*)
theorem DelayFeedback_LocalSystem: "DelayFeedback init ({.p.} \<circ> [:r:])  
      = LocalSystem init p r"
  apply (cut_tac p = "\<lambda> u x . p (u, x)" and r = "\<lambda> u u' x y . r (u,x) (u',y)" in DelayFeedback_SymbolicSystem_aux)
  by simp
    

lemma DelayFeedback_simp: "DelayFeedback init ({.p.} o [:r:]) = {.prec_pre_sts init p r.} o [:rel_pre_sts init r:]"
  by(simp add: DelayFeedback_LocalSystem LocalSystem_simp)

lemma prec_pre_sts_prec_rel: "(\<And> s s' x y . p (s, x) \<Longrightarrow> r (s, x) (s', y) = r' (s, x) (s', y)) \<Longrightarrow> prec_pre_sts init p r = prec_pre_sts init p r'"
proof -
  assume [simp]: "\<And> s s' x y . p (s, x) \<Longrightarrow> r (s, x) (s', y) = r' (s, x) (s', y)"
  have "{.p.} o [:r:] = {.p.} o [:r':]"
    by (simp add: fun_eq_iff demonic_def assert_def le_fun_def, auto)
  from this have "prec (DelayFeedback init ({.p.} o [:r:])) = prec (DelayFeedback init ({.p.} o [:r':]))"
    by simp
  from this show "prec_pre_sts init p r = prec_pre_sts init p r'"
    by (simp add: DelayFeedback_simp)
qed


theorem DelayFeedback_a_simp: "DelayFeedback init ({.p.} o [:r:]) =
     {.x .(\<forall> u y . init (u 0) \<longrightarrow> (\<forall>n . (\<forall> i < n . r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))).} 
      o [:x \<leadsto> y . (\<exists> u . init (u 0) \<and> (\<forall> i . r (u i, x i) (u (Suc i), y i))):]"
proof (simp add: DelayFeedback_simp)
  have [simp]: "prec_pre_sts init p r =
      (\<lambda> x . \<forall> u y . init (u 0) \<longrightarrow> (\<forall>n . (\<forall> i < n . r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n)))"
    by (simp add: fun_eq_iff prec_pre_sts_simp)
  have [simp]: "rel_pre_sts init r = (\<lambda> x y . \<exists>u. init (u 0) \<and> (\<forall>i. r (u i, x i) (u (Suc i), y i)))"
    by (simp add: fun_eq_iff rel_pre_sts_simp)
  show " {. prec_pre_sts init p r .} \<circ> [: rel_pre_sts init r :] =
    {.x.\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n)).} 
      \<circ> [:x\<leadsto>y.\<exists>u. init (u 0) \<and> (\<forall>i. r (u i, x i) (u (Suc i), y i)):]"
    by simp
qed
    
theorem DelayFeedback_b_simp: "DelayFeedback init ([:r:]) 
      = [:rel_pre_sts init r:]"
  apply (cut_tac init = init and p = "\<top>" and r = "r" in  DelayFeedback_simp)
  by (simp add: always_def until_def at_fun_def  assert_true_skip_a assert_true_skip)


lemma DelayFeedback_comp: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> init' b \<Longrightarrow> 
     DelayFeedback init ({.p.} o [:r:]) o DelayFeedback init' ({.p'.} o [:r':]) =
        DelayFeedback (prod_pred init init') ({.prec_comp_sts p r p'.} o [:rel_comp_sts r r':])"
  by (simp add: DelayFeedback_LocalSystem sts_comp)
    

lemma DelayFeedback_empty_init[simp]: "DelayFeedback \<bottom> S' = \<top>"
  by (simp add: DelayFeedback_def fun_eq_iff demonic_def le_fun_def)


lemma assert_bot: "{.\<bottom>::'a::boolean_algebra.} = Fail"
  by (simp add: fun_eq_iff Fail_def assert_def)
    
lemma Fail_comp: "Fail o S = Fail"
  by (simp add: fun_eq_iff Fail_def assert_def)


lemma DelayFeedback_Fail[simp]: "init a \<Longrightarrow> DelayFeedback init (Fail:: ('a \<times> 'b \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c \<Rightarrow> bool)) = Fail"
  proof -
    assume A[simp]: "init a"
    have "DelayFeedback init (Fail:: ('a \<times> 'b \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c \<Rightarrow> bool)) = DelayFeedback init ({.\<bottom>.} o [:\<top>:])"
      by (simp add: assert_bot Fail_comp)
    also have "... = {. prec_pre_sts init (\<bottom>::('a \<times> 'c \<Rightarrow> bool)) (\<top>::('a \<times> 'c \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool)) .} \<circ> [: rel_pre_sts init (\<top>::('a \<times> 'c \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool)) :]"
      by (simp add: DelayFeedback_simp)
    also have "... = Fail"
      thm prec_pre_sts_bot
      using A apply (unfold prec_pre_sts_bot)
      by (simp add: assert_bot Fail_comp)
    finally show ?thesis by simp
  qed
 
lemma prod_empty [simp]: "prod_pred X \<bottom> = \<bottom>"
  by (simp add: fun_eq_iff prod_pred_def)

lemma sts_serial_comp_empty_init: "DelayFeedback (prod_pred \<top> \<bottom>) (sts_comp Fail S') \<noteq> DelayFeedback \<top> Fail o DelayFeedback \<bottom> S'"
  apply (simp add: Fail_comp)
  by (simp add: fun_eq_iff Fail_def)

thm DelayFeedback_LocalSystem
  
theorem sts_serial_comp: "implementable S \<Longrightarrow> implementable S' \<Longrightarrow> init' b \<Longrightarrow> 
  DelayFeedback (prod_pred init init') (sts_comp S S') = DelayFeedback init S o DelayFeedback init' S'"
  apply (drule implementable_spec, safe)
  apply (drule implementable_spec, safe)
  by (simp add: DelayFeedback_LocalSystem sts_comp_prec_rel sts_comp)

theorem implementableI: "p \<le> inpt r \<Longrightarrow> implementable ({.p.} o [:r:])"
  by (simp add: implementable_def non_magic_def fun_eq_iff le_fun_def inpt_def demonic_def assert_def)

lemma implementable_inpt[simp]: "implementable ({.inpt r.} o [:r:])"
  by (simp add: implementableI)

theorem implementable_DelayFeedback: "implementable S \<Longrightarrow> init a \<Longrightarrow> implementable (DelayFeedback init S)"
  apply (drule implementable_spec, safe)
  apply (simp add: DelayFeedback_LocalSystem LocalSystem_simp)
  apply (rule implementableI, safe)
  apply (subst inpt_def)
  apply (rule_tac x = "yy r a x" in exI)
    apply (simp add: rel_pre_sts_simp)
  apply (rule_tac x = "uu r a x" in exI)
  apply safe
   apply (simp add: uu_def)
  by (rule_tac p = p  and init = init in sts_exists, simp_all)

theorem LocalSystem_impt_implementable: "init a \<Longrightarrow> implementable (LocalSystem init (inpt r) r)"
  apply (simp add:  DelayFeedback_LocalSystem [symmetric])
  by (rule implementable_DelayFeedback, simp_all)
    
lemma prec_pre_sts_inpt: "init a \<Longrightarrow> prec_pre_sts init (inpt r) r \<le> inpt (rel_pre_sts init r)"
  apply (cut_tac r = r and init = init in LocalSystem_impt_implementable, simp_all)
  apply (simp add: LocalSystem_simp)
  apply (unfold implementable_def, safe)
  by (unfold non_magic_spec, simp add: le_fun_def)

lemma comp_middle: "A o B o C o D = A o (B o C) o D"
  by (simp add: comp_assoc)

lemma fun_eq: "(\<forall>x. f x = g x) = (f = g)"
  by auto

lemma [simp]: "SkipNext \<bottom> = \<bottom>"
  by (simp add: fun_eq_iff SkipNext_simp)

lemma "SkipNext \<bottom> = \<bottom>"
  by (simp add: SkipNext_simp le_fun_def fun_eq_iff Prod_def Skip_def demonic_def prod_pred_def)

lemma "SkipNext \<top> \<bottom> = \<top>"
  by (auto simp add: SkipNext_simp le_fun_def fun_eq_iff Prod_def Skip_def demonic_def prod_pred_def)

lemma "SkipNext \<top> = \<top>"
  by (auto simp add: SkipNext_simp le_fun_def fun_eq_iff Prod_def Skip_def demonic_def prod_pred_def)

subsection{*Examples*}
  

  definition "PREC_ID = \<top>"
  definition "REL_ID = (\<lambda> (u,x) (u', y) .(u = u') \<and> (u = y))"
  definition "INIT_ID u = (u = 0)"

lemma all_eq: "\<forall>x. u x = u (Suc x) \<Longrightarrow> u x = u 0"
  apply (induction x)
  by auto

lemma "LocalSystem INIT_ID PREC_ID REL_ID = [:x \<leadsto> y . \<forall> i . y i = 0:]"
  apply (simp add:  LocalSystem_simp PREC_ID_def assert_true_skip)
  apply (rule_tac f = "demonic" in arg_cong)
  apply (simp add: rel_pre_sts_simp INIT_ID_def REL_ID_def fun_eq_iff, safe, simp_all)
   apply (cut_tac  u = u and x = i in all_eq, simp_all, metis)
    by auto
  
  definition "PREC_COUNTER = \<top>"
  definition "REL_COUNTER = (\<lambda> (u, x) (u', y) . (u' = u + 1) \<and> (u = y))"
  definition "INIT_COUNTER u = (u = 0)"

  lemma add_suc: "\<forall>x. u (Suc x) = Suc (u x) \<Longrightarrow> u x = x + u 0"
    apply (induction x)
    by auto

  lemma "LocalSystem INIT_COUNTER PREC_COUNTER REL_COUNTER = [:x \<leadsto> y . \<forall> i . y i = i:]"
  apply (simp add:  LocalSystem_simp PREC_COUNTER_def assert_true_skip)
  apply (rule_tac f = "demonic" in arg_cong)
  apply (simp add: rel_pre_sts_simp INIT_COUNTER_def REL_COUNTER_def fun_eq_iff, safe, simp_all)
   apply (cut_tac  u = u and x = i in add_suc, simp_all, metis)
    by auto
    
definition "PREC_SUM = \<top>"
definition "REL_SUM = (\<lambda> (u, x) (u', y) . (u' = u + x) \<and> (u = y))"
definition "INIT_SUM u = (u = 0)"

primrec Summ :: "(nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat" where
  "Summ x 0 = 0" |
  "Summ x (Suc n) = Summ x n + x n"

lemma sum_suc: "\<forall>n. u (Suc n) = u n + x n \<Longrightarrow> u n = Summ x n + u 0"
  apply (induction n)
  by auto
  
lemma "LocalSystem INIT_SUM PREC_SUM REL_SUM = [:x \<leadsto> y . y = Summ x:]"
  apply (simp add:  LocalSystem_simp PREC_SUM_def assert_true_skip)
  apply (rule_tac f = "demonic" in arg_cong)
  apply (simp add: rel_pre_sts_simp INIT_SUM_def REL_SUM_def fun_eq_iff, safe, simp_all)
     apply (cut_tac  u = xa and n = xb in sum_suc, simp_all, metis)
    by (rule_tac x = xa in exI, simp)

definition "PREC_A = \<top>"
definition "REL_A = (\<lambda> (u, x) (u', y) . (u' = x) \<and> (u = y))"
definition "INIT_A u = (u = 0)"

lemma "LocalSystem INIT_A PREC_A REL_A = [:x \<leadsto> y . y = [0] .. x:]"
  apply (simp add:  LocalSystem_simp PREC_A_def assert_true_skip)
  apply (rule_tac f = "demonic" in arg_cong)
  apply (simp add: rel_pre_sts_simp INIT_A_def REL_A_def fun_eq_iff append_inf_def, safe)
    apply metis
    apply (metis Suc_pred)
    by (metis diff_Suc_Suc diff_zero zero_less_Suc)

definition "PREC_SUM_A = (\<lambda> (u, x) . u \<le> 100)"
definition "REL_SUM_A = (\<lambda> (u, x) (u', y) . (u' = u + x) \<and> (u = y))"
definition "INIT_SUM_A u = (u = 0)"
  
lemma sum_suc_le: "\<forall>n < k . u (Suc n) = u n + x n \<Longrightarrow> u k = Summ x k + u 0"
  apply (induction k)
  by auto

lemma "LocalSystem INIT_SUM_A PREC_SUM_A REL_SUM_A = {.x . \<forall> i . Summ x i \<le> 100.} o [:x \<leadsto> y . y = Summ x:]"
  apply (simp add:  LocalSystem_simp PREC_SUM_def assert_true_skip)
  apply (rule prec_rel_eq)
   apply (simp add: fun_eq_iff prec_pre_sts_simp INIT_SUM_A_def REL_SUM_A_def PREC_SUM_A_def, safe)
    apply (drule_tac x = "Summ x" in spec, simp_all, blast)
   apply (cut_tac  u = u and k = n and x = x in sum_suc_le, simp_all)

  apply (simp add: rel_pre_sts_simp INIT_SUM_A_def REL_SUM_A_def fun_eq_iff, safe, simp_all)
     apply (cut_tac  u = xa and n = xb in sum_suc, simp_all, metis)
    by (rule_tac x = xa in exI, simp)


  definition "PREC_SUM_B = (\<lambda> (u, x) . u \<le> 100)"
  definition "REL_SUM_B = (\<lambda> (u, x) (u', y) . (u' = u + x \<or> u' = x) \<and> (u = y))"
  definition "INIT_SUM_B u = (u = 0)"

lemma le_sum_suc: "\<forall>n < k . u (Suc n) = u n + x n \<or> u (Suc n) = x n \<Longrightarrow> u k \<le> Summ x k + u 0"
  apply (induction k)
  by auto

lemma "LocalSystem INIT_SUM_B PREC_SUM_B REL_SUM_B 
      = {.x . \<forall> i . Summ x i \<le> 100.} o [:x \<leadsto> y . y 0 = 0 \<and> (\<forall> i . y (Suc i) = y i + x i \<or> y (Suc i) = x i):]"
  apply (simp add:  LocalSystem_simp assert_true_skip)
  apply (rule prec_rel_eq)
  apply (simp add: fun_eq_iff prec_pre_sts_simp INIT_SUM_B_def PREC_SUM_B_def REL_SUM_B_def, safe)
  apply (drule_tac x = "Summ x" in spec, simp_all, metis)
  apply (cut_tac  u = u and k = n and x = x in le_sum_suc, simp_all)
  apply auto [1]
  using order_trans apply blast
  apply (simp add: rel_pre_sts_simp INIT_SUM_B_def REL_SUM_B_def fun_eq_iff, safe, simp_all)
  apply metis
  by (rule_tac x = xa in exI, simp)
    

lemma prod_comp_spec[simp]: "pa \<le> inpt ra \<Longrightarrow> pb \<le> inpt rb \<Longrightarrow> (({.pa.} \<circ> [:ra:]) ** ({.pb.} \<circ> [:rb:])) \<circ> (({.pc.} \<circ> [:rc:]) ** ({.pd.} \<circ> [:rd:])) 
  = ({.pa.} o [:ra:] o {.pc.} o [:rc:]) ** ({.pb.} o [:rb:] o {.pd.} o [:rd:])"
  apply (simp add: Prod_spec assert_demonic_comp)
  apply (simp add: comp_assoc [THEN sym] assert_demonic_comp)
  apply (rule prec_rel_eq)
  apply (simp add: fun_eq_iff le_fun_def OO_def inpt_def)
  apply safe
  by auto

lemma prod_comp_implem: "implementable S \<Longrightarrow> implementable S' \<Longrightarrow> sconjunctive T \<Longrightarrow> sconjunctive T' \<Longrightarrow> (S ** S') o (T ** T') = (S o T) ** (S' o T')"
  apply (drule implementable_spec)
  apply (drule implementable_spec)
  apply (drule sconjunctiveE)
  apply (drule sconjunctiveE, safe)
  apply (drule_tac pa = p and pb = pa and pc = pb and pd = pc and ra = r and rb = ra and rc = rb and rd = rc in prod_comp_spec)
  by (simp_all add: comp_assoc)
    
definition "ang_rel S s q = S q s"
definition "dem_rel S q s' = q s'"

lemma mono_rep: "mono S \<Longrightarrow> S = {:ang_rel S:} o [:dem_rel S:]"
  apply (simp add: demonic_def le_fun_def fun_eq_iff ang_rel_def dem_rel_def angelic_def, safe)
  apply (rule_tac x = x in exI, simp)
  apply (drule_tac x = xb and y = x in monoE, simp_all)
  by (simp_all add: le_fun_def)

lemma mono_repE: "mono (S::('a \<Rightarrow> bool) \<Rightarrow> ('b \<Rightarrow> bool)) \<Longrightarrow> \<exists> (r::'b \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> bool) (r') . S = {:r:} o [:r':]"
  by (drule_tac S = S in mono_rep, auto)

lemma prod_comp_a: "(S o T) ** (S' o T') \<le> (S ** S') o (T ** T')"
  apply (simp add: fun_eq_iff  Prod_demonic Prod_def prod_pred_def le_fun_def demonic_def, safe)
  by auto

lemma prod_comp_angelic_demonic: "({:r::'a\<Rightarrow>'b\<Rightarrow>bool:} ** {:r'::'c\<Rightarrow>'d\<Rightarrow>bool:}) o ([:t:] ** [:t':]) = ({:r:} o [:t:]) ** ({:r':} o [:t':])"
  apply (rule antisym)
  apply (simp_all add: prod_comp_a)
  proof safe
    fix  q s s'
    assume "({: r :} ** {: r' :} \<circ> [: t :] ** [: t' :]) q (s, s') "
    from this have "\<exists>p p'. prod_pred p p' \<le> ([: t :] ** [: t' :]) q \<and> {: r :} p s \<and> {: r' :} p' s'"
      by (simp add: Prod_def)
    from this have A: "\<exists>p p'. (\<forall>a b. p a \<and> p' b \<longrightarrow> (\<exists>pa pa'. prod_pred pa pa' \<le> q \<and> t a \<le> pa \<and> t' b \<le> pa')) \<and> (\<exists>x. r s x \<and> p x) \<and> (\<exists>x. r' s' x \<and> p' x)"
      by (simp add: Prod_def  demonic_def angelic_def le_fun_def fun_eq_iff prod_pred_def)
    from this obtain p p' where B: "(\<forall>a b. p a \<and> p' b \<longrightarrow> (\<exists>pa pa'. prod_pred pa pa' \<le> q \<and> t a \<le> pa \<and> t' b \<le> pa')) \<and> (\<exists>x. r s x \<and> p x) \<and> (\<exists>x. r' s' x \<and> p' x)"
      by blast
   from B have "\<exists>pa pa'. (( prod_pred pa pa' \<le> q )) \<and> (\<exists>x. r s x \<and> t x \<le> pa) \<and> (\<exists>x. r' s' x \<and> t' x \<le> pa')"
    apply safe
    apply (drule_tac x = x in spec)
    apply (drule_tac x = xa in spec)
    apply simp
    apply safe
    by auto
  from this show "(({: r :} \<circ> [: t :]) ** ({: r' :} \<circ> [: t' :])) q (s, s')"
    by (simp add: Prod_def demonic_def angelic_def fun_eq_iff)
 qed

definition "prod_rel r r' = (\<lambda> (x, y) (u, v) . r x u \<and> r' y v)"        

lemma Prod_angelic: "{:r:} ** {:r':} = {: prod_rel r r' :}"
  apply (simp add: fun_eq_iff angelic_def prod_pred_def prod_rel_def Prod_def le_fun_def, auto)
  apply (rule_tac x = "\<lambda> x . x = aa" in exI, simp)
  by (rule_tac x = "\<lambda> x . x = ba" in exI, simp)

lemma Prod_demonic_rel: "[:r:] ** [:r':] = [: prod_rel r r' :]"
  apply (simp add: Prod_demonic)
  apply (rule_tac f = demonic in arg_cong)
  by (simp add: fun_eq_iff prod_rel_def)

lemma prod_rel_comp: "prod_rel r r' OO prod_rel t t' = prod_rel (r OO t) (r' OO t')"
  apply (simp add: fun_eq_iff OO_def prod_rel_def)
  by auto
  
lemma prod_comp_angelic_demonic_demonic: "(({:ra:} o [:rd:]) ** ({:ra':} o [:rd':])) o ([:r:] ** [:r':]) = ({:ra:} o [:rd:] o [:r:]) ** (({:ra':} o [:rd':]) o [:r':])"
  proof -
    have "(({:ra:} o [:rd:]) ** ({:ra':} o [:rd':])) o ([:r:] ** [:r':]) = ({:ra:} ** {:ra':}) o ([:rd:] ** [:rd':]) o ([:r:] ** [:r':])"
      by (simp add: prod_comp_angelic_demonic)
    also have "... = ({:ra:} ** {:ra':}) o [:prod_rel rd rd':] o [:prod_rel r r':]"
      by (simp add: Prod_demonic_rel)
    also have "... = ({:ra:} ** {:ra':}) o [:prod_rel rd rd' OO prod_rel r r':]"
      by (simp add: comp_assoc demonic_demonic)

    also have "... = ({:ra:} ** {:ra':}) o [:prod_rel (rd OO r) (rd' OO r'):]"
      by (simp add:  prod_rel_comp)
    also have "... = ({:ra:} ** {:ra':}) o ([:(rd OO r):] ** [: (rd' OO r'):])"
      by (simp add: Prod_demonic_rel)
    also have "... = ({:ra:} o [:(rd OO r):]) ** ({:ra':} o [: (rd' OO r'):])"
      by (simp add: prod_comp_angelic_demonic)
    also have "... = ({:ra:} o [:rd:] o [:r:]) ** (({:ra':} o [:rd':]) o [:r':])"
      by (simp add: demonic_demonic [THEN sym] comp_assoc)
    finally show ?thesis by simp
 qed
      
lemma prod_comp_demonic: "mono (S::('a \<Rightarrow> bool) \<Rightarrow> ('b \<Rightarrow> bool)) \<Longrightarrow> mono (S':: ('c \<Rightarrow> bool) \<Rightarrow> ('d \<Rightarrow> bool)) \<Longrightarrow> 
    (S ** S') o ([:r:] ** [:r':]) = (S o [:r:]) ** (S' o [:r':])"
    apply (cut_tac ra  = "ang_rel S" and rd = "dem_rel S" and 
      ra' = "ang_rel S'" and rd' = "dem_rel S'" and r = r and r' = r' in prod_comp_angelic_demonic_demonic)
    by (simp add: mono_rep [THEN sym])
      
theorem DelayFeedback_prod: "init a \<Longrightarrow> init' a' \<Longrightarrow> implementable S \<Longrightarrow> implementable S' \<Longrightarrow> DelayFeedback init S ** DelayFeedback init' S' = 
   [- (x, y) \<leadsto> x || y -] o DelayFeedback (prod_pred init init') (prod_sts S S') o [-\<lambda> x . (fst o x, snd o x) -]"
  apply (drule implementable_spec, safe)
  apply (drule implementable_spec, safe)
  by (simp add: DelayFeedback_LocalSystem prod_sts sts_prod)
    

(* Definition of feedback for systems that already contain a unit delay *)

lemma rel_fun_power: "((\<lambda>x y. y = (f::'a \<Rightarrow> 'a) x) ^^ n) = (\<lambda> x y . (y = (f ^^ n) x))"
  apply (induction n)
  by auto

  lemma [simp]: "[:\<bottom>:] = Magic"
    by (simp add: Magic_def)



  definition "IterateMask S n = Mask n ((S::('a::trace\<Rightarrow>bool) \<Rightarrow> ('a\<Rightarrow>bool)) ^^ n)"

  lemma IterateMask_simp: "IterateMask S = (\<lambda> n. Mask n (S ^^ n))"
    by (simp add: IterateMask_def)


  definition "IterateOmega S = Fusion (IterateMask S)"

  definition "IterateMaskA S n = Mask (n - 1) ((S::('a::trace\<Rightarrow>bool) \<Rightarrow> ('a\<Rightarrow>bool)) ^^ n)"

  lemma IterateMaskA_simp: "IterateMaskA S = (\<lambda> n. Mask (n-1) (S ^^ n))"
    by (simp add: IterateMaskA_def)


  definition "IterateOmegaA S = Fusion (IterateMaskA S)"

  lemma "IterateMaskA S n = (S ^^ n)o [:x \<leadsto> y . \<forall> (i::nat) < n - 1 . ((y i)::'a) = x i:]"
    apply (simp add: IterateMaskA_def Mask_def SkipTop_def)
    apply (subgoal_tac "eqtop (n - Suc 0) = (\<lambda> x y . \<forall> (i::nat) < n - 1 . ((y i) ::'a) = x i)")
    apply auto [1]
    by (simp add: fun_eq_iff, auto)

lemma power_refin: "mono S \<Longrightarrow> (S::'a::order\<Rightarrow>'a) \<le> T \<Longrightarrow> S ^^ n \<le> T ^^ n"
proof (induction n)
  case 0
  then show ?case
    by simp
next
  case (Suc n)
  have "\<And> x . S ((S ^^ n) x) \<le> T ((T ^^ n) x)"
    apply (rule_tac y = "S ((T ^^ n) x)" in order_trans)
    apply (simp add: Suc.IH Suc.prems(1) Suc.prems(2) le_funD monoD)
    by (simp add: Suc.prems(2) le_funD)
  from this show ?case
    by (simp add: le_fun_def)
qed

lemma IterateMaskA_refin: "mono S \<Longrightarrow> S \<le> T \<Longrightarrow> IterateMaskA S n \<le> IterateMaskA T n"
  by (simp add: IterateMaskA_def power_refin Mask_refinement)


lemma IterateOmegaA_refin: "mono S \<Longrightarrow> S \<le> T \<Longrightarrow> IterateOmegaA S \<le> IterateOmegaA T"
  apply (simp add: IterateOmegaA_def)
  by (simp add: Fusion_refinement IterateMaskA_refin)


  lemma IterateOmega_spec: "IterateOmega ({.p.} o [:r:]) 
    =  {. (\<lambda>x . \<forall> n . \<forall>y. (r ^^ n) x y \<longrightarrow> p y) .} 
       \<circ> [: INF n. (r ^^ n) OO eqtop n :]"
    apply (simp add: IterateOmega_def IterateMask_simp Mask_def power_spec
        Iterate_SkipNext_IterateSkipRel SkipTop_def comp_assoc demonic_demonic)
    apply (simp add: Fusion_spec)
    apply (subgoal_tac "(INF x. (\<lambda>xa. \<forall>i<x. \<forall>y. (r ^^ i) xa y \<longrightarrow> p y)) 
           = (\<lambda>x . \<forall>n y. (r ^^ n) x y \<longrightarrow> p y)")
    by auto

  lemma IterateOmegaA_spec: "IterateOmegaA ({.p.} o [:r:]) 
    =  {. (\<lambda>x . \<forall> n y. (r ^^ n) x y \<longrightarrow> p y) .} 
       \<circ> [: INF n. (r ^^ n) OO eqtop (n-1) :]"
    apply (simp add: IterateOmegaA_def IterateMaskA_simp Mask_def power_spec
        Iterate_SkipNext_IterateSkipRel SkipTop_def comp_assoc demonic_demonic)
    apply (simp add: Fusion_spec)
    apply (subgoal_tac "(INF x. (\<lambda>xa. \<forall>i<x. \<forall>y. (r ^^ i) xa y \<longrightarrow> p y)) 
           = (\<lambda>x . \<forall>n y. (r ^^ n) x y \<longrightarrow> p y)")
    by auto



  lemma IterateOmegaA_demonic: "IterateOmegaA ([:r:]) 
    = [: INF n. (r ^^ n) OO eqtop (n-1) :]"
    apply (cut_tac p = \<top> and r = r in IterateOmegaA_spec)
    by (simp add: assert_true_skip assert_true_skip_a)


  lemma rel_power_a: "\<And> y . ((r :: 'a \<Rightarrow> 'a \<Rightarrow> bool) ^^ n) x y \<Longrightarrow> \<exists> a . x = a 0 \<and> y = a n \<and> (\<forall> i < n . r (a i) (a (Suc i)))"
    proof (induction n, simp_all)
      fix y::'a show "\<exists>a. y = a 0"
        by (rule_tac x = "\<lambda> i . y" in exI, simp)
    next
      fix y::'a and n
      assume " (r ^^ n OO r) x y"
      from this obtain z where [simp]: "(r ^^ n) x z " and [simp]: "r z y"
        by blast
      assume "(\<And>y. (r ^^ n) x y \<Longrightarrow> \<exists>a. x = a 0 \<and> y = a n \<and> (\<forall>i<n. r (a i) (a (Suc i))))"
      from this have "\<exists>a. x = a 0 \<and> z = a n \<and> (\<forall>i<n. r (a i) (a (Suc i)))"
        by simp
      from this obtain a where [simp]: "x = a 0" and [simp]: "z = a n" and [simp]: "(\<forall>i<n. r (a i) (a (Suc i)))"
        by blast
      show "\<exists>a. x = a 0 \<and> y = a (Suc n) \<and> (\<forall>i<Suc n. r (a i) (a (Suc i)))"
        apply (rule_tac x = "\<lambda> i . if i \<le> n then a i else y" in exI)
        apply auto
        using \<open>r z y\<close> less_Suc_eq by auto
  qed

  lemma rel_power_b: "\<And> y . \<exists> a . x = a 0 \<and> y = a n \<and> (\<forall> i < n . r (a i) (a (Suc i))) \<Longrightarrow> ((r :: 'a \<Rightarrow> 'a \<Rightarrow> bool) ^^ n) x y"
    apply (induction n, simp_all)
    apply blast
    apply safe
    apply (simp add: OO_def)
    apply (rule_tac x = "a n" in exI, simp_all)
    using less_Suc_eq by auto

  lemma rel_power: "((r :: 'a \<Rightarrow> 'a \<Rightarrow> bool) ^^ n) x y = (\<exists> a . x = a 0 \<and> y = a n \<and> (\<forall> i < n . r (a i) (a (Suc i))))"
    by (metis relpowp_fun_conv)


  lemma IterateOmega_demonic_spec: "IterateOmega [:r:] = [: INF n. r ^^ n OO eqtop n :]"
    apply (cut_tac p="\<lambda>x. True" and r=r in IterateOmega_spec)
    by (simp add: assert_true_skip_a)

  lemma IterateOmega_func: "IterateOmega [- f -] = [: x \<leadsto> y . \<forall> n. eqtop n ((f ^^ n) x) y :]"
    apply (simp add: IterateOmega_demonic_spec update_def)
    apply (rule_tac f = demonic in arg_cong)
    by (simp add: fun_eq_iff OO_def rel_fun_power)

  lemma IterateOmega_func_aux_a: "(\<forall> n. eqtop n ((f ^^ n) x) y) = (\<forall> n . \<forall> i < n . (f ^^ n) x i = y i)"
    by (simp)

  lemma IterateOmega_func_a: "IterateOmega [- f -] = [: x \<leadsto> y . (\<forall> n . \<forall> i < n . (f ^^ n) x i = y i):]"
   by (simp add: IterateOmega_func_aux_a IterateOmega_func)

  definition "apply x i = ((fst (fst x) i, snd (fst x) i), snd x i)"

  lemma IterateOmega_func_aux_b: "(\<forall> n. eqtop n ((f^^n) x) y) = (\<forall> n::nat . \<forall> i::nat < n . apply ((f ^^ n) x) i = apply y i)"
    by (simp add: eqtop_prod_def apply_def, auto)


  lemma IterateOmega_func_aa: "IterateOmega [- f -] = [: x \<leadsto> y . (\<forall> n . \<forall> i::nat < n . apply ((f ^^ n) x) i = apply y i):]"
    by (simp add: IterateOmega_func_aux_b IterateOmega_func)

  lemma IterateOmega_func_b: "(\<forall> x n . \<forall> i < n . (f ^^ n) x i = (f ^^ (Suc i)) x i) \<Longrightarrow> IterateOmega [- f -] = [-\<lambda> x . (\<lambda> i . (f ^^ (Suc i)) x i)-]"
    apply (simp add: IterateOmega_func_a)
    apply (unfold update_def)
    apply (rule_tac f = demonic in arg_cong)
    apply (simp add: fun_eq_iff, auto)
    by (metis gt_ex)

  lemma IterateOmega_func_bb: "(\<forall> x n . \<forall> i::nat < n . apply (((f::( ((nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b)) \<times> (nat \<Rightarrow> 'c) \<Rightarrow>  ((nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b)) \<times> (nat \<Rightarrow> 'c)) )^^ n) x) i = apply ((f ^^ (Suc i)) x) i) 
       \<Longrightarrow>  
      IterateOmega [- f -] =[- (\<lambda> x . (let z = (\<lambda> i . apply ((f ^^ (Suc i)) x) i) in ((fst o fst o z, snd o fst o z), snd o z))) -]"
    apply (simp add: IterateOmega_func_aa)
    apply (unfold update_def)
    apply (rule_tac f = demonic in arg_cong)
    apply (simp add: fun_eq_iff apply_def Let_def, auto)
    apply (metis gt_ex)
    apply (metis gt_ex)
    by (metis gt_ex)

  lemma IterateOmega_func_c: "\<forall> x . \<not> (\<forall> n . \<forall> i < n . (f ^^ n) x i = (f ^^ (Suc i)) x i) \<Longrightarrow> IterateOmega [- f -] = Magic"
    apply (simp add: IterateOmega_func_a)
    apply (subgoal_tac "(\<lambda> x y . \<forall>n i. i < n \<longrightarrow> (f ^^ n) x i = y i) = \<bottom>")
    apply simp
    apply (simp add: fun_eq_iff, auto)
    apply (drule_tac x = x in spec, safe)
    by (metis comp_apply funpow.simps(2) lessI)


  lemma IterateOmega_assert_update: "IterateOmega ({.p.} o [-f-]) 
    =  {. (\<lambda>x . \<forall> n . p ((f ^^ n) x)) .} 
       \<circ> [: x \<leadsto> y . \<forall> n. eqtop n ((f ^^ n) x) y :]"
    apply (simp add: IterateOmega_def IterateMask_simp Mask_def power_spec
        Iterate_SkipNext_IterateSkipRel SkipTop_def comp_assoc demonic_demonic update_def)
    apply (simp add:  Fusion_spec)
    apply (subgoal_tac "(INF x. (\<lambda>xa. \<forall>i<x. \<forall>y. ((\<lambda>x y. y = f x) ^^ i) xa y \<longrightarrow> p y)) 
           = (\<lambda>x. \<forall>n. p ((f ^^ n) x))")
    apply auto
    apply (subgoal_tac "(INF x. (\<lambda>x y. y = f x) ^^ x OO eqtop x) = (\<lambda>x y. \<forall>n. eqtop n ((f ^^ n) x) y)")
    apply auto
    apply (simp add: fun_eq_iff OO_def rel_fun_power)
    apply (simp add: fun_eq_iff rel_fun_power)
    by blast

  lemma IterateOmega_assert_update_a: "IterateOmega ({.p.} o[- f -]) = {. (\<lambda>x . \<forall> n . p ((f ^^ n) x)) .} o [: x \<leadsto> y . (\<forall> n . \<forall> i < n . (f ^^ n) x i = y i):]"
    by (simp add: IterateOmega_assert_update)

  lemma IterateOmega_assert_update_b: "(\<forall> x n . \<forall> i < n . (f ^^ n) x i = (f ^^ (Suc i)) x i) \<Longrightarrow> IterateOmega ({.p.} o[- f -]) = {.(\<lambda>x . \<forall> n . p ((f ^^ n) x)).} o [-\<lambda> x . (\<lambda> i . (f ^^ (Suc i)) x i)-]"
    apply (simp add: IterateOmega_assert_update_a)
    apply (unfold update_def)
    by (metis gt_ex)

  lemma IterateOmega_assert_update_c: "IterateOmega ({.p.} o [- f -]) = {.(\<lambda>x . \<forall> n . p ((f ^^ n) x)).} o [: x \<leadsto> y . (\<forall> n . \<forall> i::nat < n . apply ((f ^^ n) x) i = apply y i):]"
    by (simp add: IterateOmega_func_aux_b IterateOmega_assert_update)


  thm IterateOmega_spec

  lemma IterateOmega_assert_update_d: "(\<forall> x n . \<forall> i::nat < n . apply (((f::( ((nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b)) \<times> (nat \<Rightarrow> 'c) \<Rightarrow>  ((nat \<Rightarrow> 'a) \<times> (nat \<Rightarrow> 'b)) \<times> (nat \<Rightarrow> 'c)) )^^ n) x) i = apply ((f ^^ (Suc i)) x) i) \<Longrightarrow>  
      IterateOmega ({.p.} o [- f -]) = {.(\<lambda>x . \<forall> n . p ((f ^^ n) x)).} o [- (\<lambda> x . (let z = (\<lambda> i . apply ((f ^^ (Suc i)) x) i) in ((fst o fst o z, snd o fst o z), snd o z))) -]"
    apply (simp add: IterateOmega_assert_update_c)
    apply (unfold update_def)
    apply (rule prec_rel_eq, simp_all)
    apply (simp add: fun_eq_iff apply_def Let_def, auto)
    apply (metis gt_ex)
    apply (metis gt_ex)
    by (metis gt_ex)

  lemma IterateOmega_assert_update_e: "\<forall> x . \<not> (\<forall> n . \<forall> i < n . (f ^^ n) x i = (f ^^ (Suc i)) x i) \<and> (\<forall>n. p ((f ^^n) x)) \<Longrightarrow> IterateOmega ({.p.} o [- f -]) = Magic"
    apply (simp add: IterateOmega_assert_update_a)
    apply (subgoal_tac "(\<lambda> x y . \<forall>n i. i < n \<longrightarrow> (f ^^ n) x i = y i) = \<bottom>")
    apply (simp add: assert_true_skip_a)
    apply (simp add: fun_eq_iff, auto)
    apply (drule_tac x = x in spec, safe)
    by (metis comp_apply funpow.simps(2) lessI)

      
 
definition "defined r = (\<forall> x . \<exists> y . r x y)"
  

fun calcu :: "(nat \<Rightarrow> 'a) \<Rightarrow> (nat \<Rightarrow> 'b) \<Rightarrow> ('a \<times> 'b \<Rightarrow> 'a \<times> 'c \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a" where
  "calcu u x r n i = (if i \<le> n then u i else SOME u' . (\<exists>y. r (calcu u x r n (i-1), x (i-1)) (u', y)))"

  thm choice_iff'

lemma prec_loc_st_defined_simp: "defined r \<Longrightarrow> prec_pre_sts init p r
          = (\<lambda> x . \<forall> u . init (u 0) \<longrightarrow> (\<forall>n . \<exists> y. r (u n, x n) (u (Suc n), y)) \<longrightarrow> (\<forall> n . p (u n, x n)))"
    apply (simp add: prec_pre_sts_simp fun_eq_iff)
  apply (safe)
    apply (simp add: choice_iff)
   apply blast
    apply (subst (asm) defined_def)
    apply (drule_tac x = "(calcu u x r n)" in spec)
  apply safe
    apply simp
   apply (simp del: calcu.simps)
   apply (case_tac "na < n", simp, blast) 
   apply (case_tac "na = n" , simp)
    apply (drule_tac x = "u n" in spec)
    apply (drule_tac x = "x n" in spec, safe)
    apply (rule someI, blast)
    apply (subst (2) calcu.simps)
   apply (simp del: calcu.simps)
    apply (drule_tac x = "(calcu u x r n na)" in spec)
    apply (drule_tac x = "x na" in spec, safe)
   apply (rule someI, blast)
proof -
  fix u n x
  assume " \<forall>na. p (calcu u x r n na, x na)"
  from this have "p (calcu u x r n n, x n)"
    by (simp del: calcu.simps)
  from this show "p (u n, x n)"
    by simp
qed
  
(*
lemma prec_loc_st_defined_simp_aux_a: "(\<forall>n . \<exists> y. r (u n, x n) (u (Suc n), y)) = (\<box> (lft_r_a (inpt_a r))) (u , x) (\<odot> u)"
    by (simp add: always_def at_prod_def at_fun_def lft_r_a_def inpt_a_def next_trace_def)

  lemma prec_loc_st_defined_simp_aux_b: "(\<forall>n::nat. p (u n, x n)) = (\<box> (lft_p p)) (u, x)"
    by (simp add: always_def at_prod_def at_fun_def lft_p_def)

  lemma prec_loc_st_defined_simp_a: "defined r \<Longrightarrow> (\<forall> u . init (u 0) \<longrightarrow> prec_loc_st (\<lambda> u x . p (u,x)) (\<lambda> u u' x y . r (u, x) (u', y)) u x)
          = (\<forall> (u::nat \<Rightarrow> 'a) . init (u 0) \<and> (\<box> (lft_r_a (inpt_a r))) (u , x) (\<odot> u) \<longrightarrow> (\<box> (lft_p p)) (u, x))"
    by (simp add: prec_loc_st_defined_simp prec_loc_st_defined_simp_aux_a prec_loc_st_defined_simp_aux_b, auto)

  lemma DelayFeedback_defined_simp: "defined r \<Longrightarrow> DelayFeedback init ({.p.} o [:r:]) 
                =  {.x . \<forall> (u::nat \<Rightarrow> 'a) . init (u 0) \<and> (\<box> (lft_r_a (inpt_a r))) (u , x) (\<odot> u) \<longrightarrow> (\<box> (lft_p p)) (u, x).} 
                 o [:x \<leadsto> y . \<exists> (u::nat \<Rightarrow> 'a) . init (u 0) \<and> (\<box> (lft_r r)) (u, x) (\<odot> u, y):]"
    by (simp add: DelayFeedback_simp  always_st_simp prec_loc_st_defined_simp_a)
      
*)

lemma DelayFeedback_defined_simp: "defined r \<Longrightarrow> DelayFeedback init ({.p.} o [:r:]) 
              =  {.x . \<forall> (u::nat \<Rightarrow> 'a) . init (u 0) \<and> ((\<forall>n . \<exists> y. r (u n, x n) (u (Suc n), y))) \<longrightarrow> (\<forall> n . p  (u n, x n)).} 
               o [:rel_pre_sts init r :]"
proof (simp add: DelayFeedback_simp prec_loc_st_defined_simp)
  have [simp]: "\<And> x . (\<forall>u. init (u 0) \<longrightarrow> (\<forall>n. \<exists>y. r (u n, x n) (u (Suc n), y)) \<longrightarrow> (\<forall>n. p (u n, x n)))
      = (\<forall>u. init (u 0) \<and> (\<forall>n. \<exists>y. r (u n, x n) (u (Suc n), y)) \<longrightarrow> (\<forall>n. p (u n, x n)))"
    by blast
  show " {.x.\<forall>u. init (u 0) \<longrightarrow> (\<forall>n. \<exists>y. r (u n, x n) (u (Suc n), y)) \<longrightarrow> (\<forall>n. p (u n, x n)).} \<circ> [: rel_pre_sts init r :] =
    {.x.\<forall>u. init (u 0) \<and> (\<forall>n. \<exists>y. r (u n, x n) (u (Suc n), y)) \<longrightarrow> (\<forall>n. p (u n, x n)).} \<circ> [: rel_pre_sts init r :]"
    by simp
qed
  

  
lemma defined_fun[simp]: "defined (\<lambda> x y . y = f x)"
  by (simp add: defined_def)

definition "map_f f x n = f (fst x n, snd x n)"
  
(*todo move inside DelayFeedback_update_simp*)
lemma DelayFeedback_update_simp_aux_b: "(\<forall>n. \<exists>y. (u (Suc n), y) = f (u n, x n))  = ((\<odot> u) = map_f (fst o f) (u, x))"    
  apply (simp add: always_def at_prod_def at_fun_def  fun_eq_iff map_f_def next_trace_def, auto)
  apply (metis eq_fst_iff)
  by (metis prod.exhaust_sel)

lemma DelayFeedback_update_simp_aux_a: "rel_pre_sts init (\<lambda>x y. y = f x) = (\<lambda> x y . \<exists>u. init (u 0) \<and> \<odot> u = map_f (fst \<circ> f) (u, x) \<and> y = map_f (snd \<circ> f) (u, x))"
  apply (simp add: always_def at_prod_def at_fun_def  fun_eq_iff map_f_def next_trace_def rel_pre_sts_simp, safe)
  apply (metis fst_conv snd_conv)
  by auto

lemma DelayFeedback_update_simp: "DelayFeedback init ({.p.} o [-f-]) 
      = {. \<lambda>x. \<forall> (u::nat \<Rightarrow> 'a). init (u 0) \<and> (\<odot> u) = map_f (fst o f) (u, x) \<longrightarrow> (\<forall> n . p (u n, x n)) .} 
      \<circ> [: \<lambda>x y. \<exists> (u:: nat \<Rightarrow> 'a). init (u 0) \<and> (\<odot> u) = map_f (fst o f) (u, x) \<and> y = map_f (snd o f) (u, x) :]"
  apply (simp add: update_def DelayFeedback_simp prec_loc_st_defined_simp DelayFeedback_update_simp_aux_b DelayFeedback_update_simp_aux_a)
  apply (subgoal_tac "\<And> x . (\<forall>u. init (u (0::nat)) \<longrightarrow> \<odot> u = map_f (fst \<circ> f) (u, x) \<longrightarrow> (\<forall>n. p (u n, x n)))
    = (\<forall>u. init (u (0::nat)) \<and> \<odot> u = map_f (fst \<circ> f) (u, x) \<longrightarrow> (\<forall>n. p (u n, x n)))")
   apply simp
  by blast


primrec itr :: "('a \<times> 'b \<Rightarrow> 'a) \<Rightarrow> 'a \<Rightarrow> (nat \<Rightarrow> 'b) \<Rightarrow> nat \<Rightarrow> 'a" where 
  "itr f u0 x 0 = u0" | 
  "itr f u0 x (Suc n) = f (itr f u0 x n, x n)"

lemma map_itr_aux: "((\<odot> u) = map_f f (u, x)) \<Longrightarrow> (u n = itr f (u 0) x n)"
  apply (induction n, simp_all)
  by (simp add: fun_eq_iff next_trace_def map_f_def)

lemma map_itr_simp: "((\<odot> u) = map_f f (u, x)) = (u = itr f (u 0) x)"
  apply safe
  apply (subst fun_eq_iff, safe)
  apply (rule map_itr_aux, simp)
  apply (simp add: fun_eq_iff)
  by (metis AbsNat_nat_def add.left_neutral add_Suc at_trace_def fst_conv id_apply itr.simps(2) map_f_def next_trace_def snd_conv)

lemma DelayFeedback_update_itr_simp: "DelayFeedback init ({.p.} o [-f-]) 
          = {. x. \<forall> a . init a \<longrightarrow> (\<forall> i . p (itr (fst o f) a x i, x i)) .} 
          \<circ> [: \<lambda>x y. \<exists> a. init a \<and> y = map_f (snd o f) (itr (fst o f) a x, x) :]"
proof (simp add: DelayFeedback_update_simp map_itr_simp)
  have [simp]: "\<And> x . (\<forall>u. init (u 0) \<and> u = itr (fst \<circ> f) (u 0) x \<longrightarrow> (\<forall>n. p (u n, x n))) = (\<forall>a. init a \<longrightarrow> (\<forall>i. p (itr (fst \<circ> f) a x i, x i)))"
    by auto
  have [simp]: "\<And> x y . (\<exists>u. init (u 0) \<and> u = itr (fst \<circ> f) (u 0) x \<and> y = map_f (snd \<circ> f) (u, x))
    = (\<exists>a. init a \<and> y = map_f (snd \<circ> f) (itr (fst \<circ> f) a x, x))"
    apply auto
    by (rule_tac x = "itr (fst \<circ> f) a x" in exI, simp_all)
      
  show " {.x.\<forall>u. init (u 0) \<and> u = itr (fst \<circ> f) (u 0) x \<longrightarrow> (\<forall>n. p (u n, x n)).} 
      \<circ> [:x\<leadsto>y.\<exists>u. init (u 0) \<and> u = itr (fst \<circ> f) (u 0) x \<and> y = map_f (snd \<circ> f) (u, x):] =
        {.x.\<forall>a. init a \<longrightarrow> (\<forall>i. p (itr (fst \<circ> f) a x i, x i)).} \<circ> [:x\<leadsto>y.\<exists>a. init a \<and> y = map_f (snd \<circ> f) (itr (fst \<circ> f) a x, x):]"
    by simp
qed
  
definition "DelayFeedbackInit a S = DelayFeedback (\<lambda> u . u = a) S"
  
  
definition "lft_1_2 p = (\<lambda> (x, y) . p (x (0::nat), y (0::nat)))"
definition "lft_2_2 r = (\<lambda> (x, y) (z, t). r (x (0::nat), y (0::nat)) (z (0::nat), t (0::nat)))"

theorem DelayFeedbackInit_update_simp_a: "DelayFeedbackInit u ({.p.} o [-f-]) 
                = {. x. (\<forall> n . p (itr (fst o f) u x n, x n)) .} \<circ> [-\<lambda>x . map_f (snd o f) (itr (fst o f) u x, x)-]"
  apply (simp add: DelayFeedback_update_itr_simp DelayFeedbackInit_def)
  by (simp add: update_def)

lemma [simp]: "(\<box> lft_1_2 \<top>) = \<top>"
  by (simp add: always_def fun_eq_iff lft_1_2_def at_fun_def)
    
theorem DelayFeedbackInit_update_simp_b: "DelayFeedbackInit u [-f-] = [-\<lambda>x . map_f (snd o f) (itr (fst o f) u x, x)-]"
  apply (cut_tac p = \<top> and f = f and u = u in DelayFeedbackInit_update_simp_a)
  by (simp add: assert_true_skip assert_true_skip_a)


lemma prec_itr_simp: "((\<box> lft_1_2 p) (itr f u x, x)) = (\<forall> n. p (itr f u x n, x n))"
  by (simp add: fun_eq_iff always_def lft_1_2_def at_fun_def at_prod_def)

lemma prec_itr_induction_aux: "p (u, x 0) \<Longrightarrow> (\<And> n a . p (a, x n) \<Longrightarrow> p (f (a ,x n), x (Suc n))) \<Longrightarrow> p (itr f u x n, x n)"
  by (induction n, simp_all)

lemma prec_itr_induction: "p (u, x 0) \<Longrightarrow> (\<And> n a . p (a, x n) \<Longrightarrow> p (f (a ,x n), x (Suc n))) \<Longrightarrow> ((\<box> lft_1_2 p) (itr f u x, x))"
  by (simp add: prec_itr_simp prec_itr_induction_aux)

definition "lft_r r x y = r (fst x 0, snd x 0) (fst y 0, snd y 0)"
definition "lft_r_b r x y = r (x 0) (y 0)"

lemma rel_itr_simp: "(\<box> (lft_r_b r)) x (map_f g (itr f u x, x)) = (\<forall> n . r (x n) (g (itr f u x n, x n)))"
  by (simp add: fun_eq_iff always_def  at_fun_def at_prod_def lft_r_b_def map_f_def)

  lemma rel_itr_induction_aux: "r (x 0) (g (u, x 0)) \<Longrightarrow> (\<And> n a . r (x n) (g (a, x n)) \<Longrightarrow> r (x (Suc n)) (g (f (a ,x n), x (Suc n)))) \<Longrightarrow> r (x n) (g (itr f u x n, x n))"
    by (induction n, simp_all)
  
  lemma rel_itr_induction: "r (x 0) (g (u, x 0)) \<Longrightarrow> (\<And> n a . r (x n) (g (a, x n)) \<Longrightarrow> r (x (Suc n)) (g (f (a ,x n), x (Suc n)))) \<Longrightarrow> (\<box> (lft_r_b r)) x (map_f g (itr f u x, x))"
    by (simp add: rel_itr_simp rel_itr_induction_aux)

  lemma rel_bounded_itr_induction_aux: "(0 \<in> b \<Longrightarrow> r (x 0) (g (u, x 0))) \<Longrightarrow> 
      (\<And> n a . (n \<in> b \<Longrightarrow> r (x n) (g (a, x n))) \<Longrightarrow> Suc n \<in> b  \<Longrightarrow> r (x (Suc n)) (g (f (a ,x n), x (Suc n)))) \<Longrightarrow> n \<in> b \<Longrightarrow> r (x n) (g (itr f u x n, x n))"
    by (induction n, simp_all)

  lemma rel_bounded_itr_induction: "(0 \<in> b \<Longrightarrow> r (x 0) (g (u, x 0))) \<Longrightarrow> (\<And> n a . (n \<in> b \<Longrightarrow> r (x n) (g (a, x n))) \<Longrightarrow> Suc n \<in> b \<Longrightarrow> r (x (Suc n)) (g (f (a ,x n), x (Suc n)))) 
      \<Longrightarrow> (\<box>b b (lft_r_b r)) x (map_f g (itr f u x, x))"
    apply (simp add: fun_eq_iff always_bounded_def  at_fun_def at_prod_def lft_r_b_def map_f_def, safe)
    apply (cut_tac g = g and r = r and n = fa and u = u and f = f and x = x  and b = b in rel_bounded_itr_induction_aux)
    by simp_all

  lemma refin_demonic_spec: "([:r:] \<le> {.p.} o [:r':]) = (p = \<top> \<and> r' \<le> r)"
    apply (simp add: fun_eq_iff le_fun_def demonic_def assert_def, auto)
    apply (rule_tac y = "r xa" in order_trans)
    by simp_all
  
lemma spec_delay_feedback_fun_refine: "({.p'.} o [: r :] \<le> DelayFeedbackInit u ({.p.} o [-f-])) = ((p' \<le> (\<lambda>x. (\<box> lft_1_2 p) (itr (fst o f) u x, x)))
    \<and> (\<forall> x . p' x \<longrightarrow>  r x ( map_f (snd o f) (itr (fst o f) u x, x)) ))"
    apply (simp add: DelayFeedbackInit_update_simp_a)
    apply (simp add: update_def)
    by (simp add: assert_demonic_refinement, auto simp add: lft_1_2_def always_def at_fun_def at_prod_def)

lemma prec_itr_inductionA: "(p' x \<Longrightarrow> p (u, x 0)) \<Longrightarrow> (\<And> n a . p' x \<Longrightarrow> p (a, x n) \<Longrightarrow> p (f (a ,x n), x (Suc n))) \<Longrightarrow> p' x \<Longrightarrow> ((\<box> lft_1_2 p) (itr f u x, x))"
  by (simp add: prec_itr_simp prec_itr_induction_aux)

lemma prec_itr_inductionB: "(\<And> x . p' x \<Longrightarrow> p (u, x 0)) \<Longrightarrow> (\<And> x n a . p' x \<Longrightarrow> p (a, x n) \<Longrightarrow> p (f (a ,x n), x (Suc n))) \<Longrightarrow> p'\<le> (\<lambda> x . (\<box> lft_1_2 p) (itr f u x, x))"
  by (simp add: le_fun_def prec_itr_simp prec_itr_induction_aux)

lemma rel_itr_inductionA: "(\<And> x . p' x \<Longrightarrow> r (x 0) (g (u, x 0))) \<Longrightarrow> (\<And> x n a . p' x \<Longrightarrow> r (x n) (g (a, x n)) \<Longrightarrow> r (x (Suc n)) (g (f (a ,x n), x (Suc n)))) 
  \<Longrightarrow> p' x \<Longrightarrow> (\<box> (lft_r_b r)) x (map_f g (itr f u x, x))"
  by (simp add: rel_itr_simp rel_itr_induction_aux)


lemma "{:z \<leadsto> x . x \<noteq> (0::nat):} o [:x \<leadsto> y . x = 0 \<and> y = (0::nat):] = \<top>"
  apply (simp_all add: fun_eq_iff demonic_def assert_def assume_def le_fun_def angelic_def)
  by auto [1]
    
lemma "{:z \<leadsto> x . x \<noteq> (Suc n):} o [:x \<leadsto> y . x = 0 \<and> y = (0::nat):] = \<top>"
  apply (simp_all add: fun_eq_iff demonic_def assert_def assume_def le_fun_def angelic_def)
  apply auto [1]
  by (rule_tac x = "Suc (Suc n)" in exI, simp)


lemma "({.p'.} o [: \<box> (lft_r_b r) :] \<le> DelayFeedbackInit u ({.p.} o [-f-])) = ((p' \<le> (\<lambda>x. (\<box> lft_1_2 p) (itr (fst o f) u x, x)))
  \<and> (\<forall> x . p' x \<longrightarrow>  (\<box> (lft_r_b r)) x ( map_f (snd o f) (itr (fst o f) u x, x)) ))"
  by (simp add: spec_delay_feedback_fun_refine)

  lemma demonic_delay_feedback_fun_refine: "([:r :] \<le> DelayFeedbackInit u ({.p.} o [-f-])) = (((\<lambda>x. (\<box> lft_1_2 p) (itr (fst o f) u x, x)) = \<top>)
    \<and> (\<forall> x . r x (map_f (snd o f) (itr (fst o f) u x, x)) ))"
    apply (simp add: DelayFeedbackInit_update_simp_a)
    apply (simp add: update_def)
    by (simp add: refin_demonic_spec, auto simp add: lft_1_2_def always_def at_fun_def at_prod_def)

lemma "([: \<box> (lft_r_b r) :] \<le> DelayFeedbackInit u ({.p.} o [-f-])) = (((\<lambda>x. (\<box> lft_1_2 p) (itr (fst o f) u x, x)) = \<top>)
  \<and> (\<forall> x . (\<box> (lft_r_b r)) x (map_f (snd o f) (itr (fst o f) u x, x)) ))"
  by (simp add: demonic_delay_feedback_fun_refine)
  
  lemma refin_update_spec: "([: \<box>b b (lft_r_b r) :] \<le> DelayFeedbackInit u ({.p.} o [-f-])) = (((\<lambda>x. (\<box> lft_1_2 p) (itr (fst o f) u x, x)) = \<top>)
    \<and> (\<forall> x y . y = map_f (snd o f) (itr (fst o f) u x, x) \<longrightarrow>  (\<box>b b (lft_r_b r)) x y ))"
    apply (simp add: DelayFeedbackInit_update_simp_a)
    apply (simp add: update_def)
    by (simp add: refin_demonic_spec, auto simp add: lft_1_2_def always_def at_fun_def at_prod_def)

  definition "prec_delay p f_state u = (\<lambda>x. (\<box> lft_1_2 p) (itr (f_state) u x, x))"
  definition "func_delay f_state f_out u = (\<lambda>x . map_f f_out (itr f_state u x, x))"

  theorem DelayFeedbackInit_update_simp_c: "DelayFeedbackInit u ({.p.} o [-f-]) 
                = {.prec_delay p (fst o f) u.} \<circ> [-func_delay (fst o f) (snd o f) u-]"
    apply (simp add: DelayFeedbackInit_update_simp_a prec_delay_def func_delay_def)
    by (auto simp add: lft_1_2_def always_def at_fun_def at_prod_def)

theorem DelayFeedbackInit_update_simp_d: "DelayFeedbackInit u [-f-] =  [-func_delay (fst o f) (snd o f) u-]"
  by (simp add: DelayFeedbackInit_update_simp_b func_delay_def)


lemma always_lft_bot: "(\<box> lft_1_2 (\<bottom>::('a \<times> 'b \<Rightarrow> bool) )) = \<bottom>"
  by (simp add: fun_eq_iff lft_1_2_def always_def at_fun_def at_prod_def)
    

lemma DelayFeedbackInit_bot: "DelayFeedbackInit u ((\<bottom>::(('a \<times> 'b) \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c) \<Rightarrow> bool)) = \<bottom>"
  apply (subst Refinement.fail_assert_update)
  apply (simp add: DelayFeedbackInit_update_simp_a always_lft_bot)
  by (simp add: fun_eq_iff assert_def)


lemma simp_prec: "{. p .} \<circ> [: \<lambda>x y. \<not> p x \<or> r x y :] = {.p.} o [:r:]"
  by (simp add: fun_eq_iff demonic_def assert_def le_fun_def, auto)

lemma inpt_and_rel: "(inpt r x \<and> r x y) = r x y"
  using inpt_def by fastforce
    
lemma [simp]: "inpt (\<lambda>x y. inpt r x \<and> r x y) = inpt r"
  by (simp add: inpt_and_rel)
    
thm DelayFeedback_defined_simp
lemma DelayFeedback_inpt: "DelayFeedback init ({.inpt r.} o [:r:]) 
    =  {.x. \<forall>(u::nat \<Rightarrow> 'a). init (u 0) \<and> ((\<forall>n. \<exists>y. \<not> inpt r (u n, x n) \<or> r (u n, x n) (u (Suc n), y))) \<longrightarrow> (\<forall>n. inpt r (u n, x n)).} \<circ>
    [:  rel_pre_sts init (\<lambda>x y. \<not> inpt r x \<or> r x y) :] "
  apply (cut_tac init = init and p = "inpt r" and r = "\<lambda> x y . \<not> inpt r x \<or> r x y" in DelayFeedback_defined_simp)
  apply (simp add: defined_def inpt_def)
  apply auto [1]
  by (simp only: simp_prec)

  declare comp_skip[simp del]
  declare skip_comp[simp del]
  declare prod_skip_skip[simp del]
declare fail_comp[simp del]

subsection{*Data Refinement*}
definition "data_refin_sts d S S' =  ({:t, x  \<leadsto> s, x' . x = x' \<and> d t s:} o S \<le> S' o {:t', y \<leadsto> s', y' . y = y' \<and> d t' s':})"

lemma data_refin_sts_simp: "data_refin_sts d ({. p .} \<circ> [: r :]) ({. p' .} \<circ> [: r' :]) = 
  ((\<forall>t x s. d t s \<and> p (s, x) \<longrightarrow> p' (t, x)) \<and>
  (\<forall>t x s t' y. d t s \<and> p (s, x) \<and> r' (t, x) (t', y) \<longrightarrow> (\<exists>s'. d t' s' \<and> r (s, x) (s', y))))"
  apply (simp add: data_refin_sts_def fun_eq_iff le_fun_def demonic_def angelic_def assert_def, safe)
  by blast+
    
primrec s_r :: "('a \<Rightarrow>'b \<Rightarrow> bool) \<Rightarrow> ('b \<Rightarrow> bool) \<Rightarrow> ('b \<times> 'c \<Rightarrow> 'b \<times> 'd \<Rightarrow> bool) \<Rightarrow> (nat \<Rightarrow> 'c) \<Rightarrow> (nat \<Rightarrow> 'd) \<Rightarrow> (nat \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> 'b" where
  "s_r d init r x y t 0 = (SOME s . d (t 0) s \<and> init s)" |
  "s_r d init r x y t (Suc n) = (SOME s . d (t (Suc n)) s \<and> r (s_r d init r x y t n, x n) (s, y n))" 
      
theorem data_refinement_sts: "(\<And> t . init' t \<Longrightarrow> \<exists> s . d t s \<and> init s) \<Longrightarrow>
  data_refin_sts d ({.p.} o [:r:]) ({.p'.} o [:r':]) \<Longrightarrow> LocalSystem init p r \<le> LocalSystem init' p' r'"
proof (simp add: LocalSystem_simp)
  assume A: "data_refin_sts d ({. p .} \<circ> [: r :]) ({. p' .} \<circ> [: r' :])"
  have C: "\<And> t x s. d t s \<Longrightarrow> p (s, x) \<Longrightarrow> p' (t, x)"
    and D: "\<And> t x s t' y. d t s \<Longrightarrow> p (s, x) \<Longrightarrow> r' (t, x) (t', y) 
        \<Longrightarrow> (\<exists>s'. d t' s' \<and> r (s, x) (s', y))"
    using A apply (simp add: data_refin_sts_simp, blast)
    using A by (simp add: data_refin_sts_simp)
      
  assume B: "\<And> t . init' t \<Longrightarrow> \<exists> s . d t s \<and> init s"
      
  define s where "s = (\<lambda> x y t . s_r d init r x y t)"
      
  have [simp]: "\<And> t x y . init' (t 0) \<Longrightarrow> init (s x y t 0)"
    apply (simp add: s_def)
    using B  by (rule someI2_ex, simp_all)

  have X[simp]: "\<And> t x y . init' (t 0) \<Longrightarrow> d (t 0) (s_r d init r x y t 0)"
    apply (simp add: s_def)
    using B  by (rule someI2_ex, simp_all)
      
  have Aux: "\<And>t x y n. init' (t 0) \<Longrightarrow>
               \<forall>n. (\<forall>i<n. r (s x y t i, x i) (s x y t (Suc i), y i)) \<longrightarrow> p (s x y t n, x n) 
      \<Longrightarrow> (\<forall>i<n. r' (t i, x i) (t (Suc i), y i)) \<Longrightarrow> (\<forall>i < n. d (t (Suc i)) (s x y t (Suc i)) \<and> r (s x y t i, x i) (s x y t (Suc i), y i))  \<and> p' (t n, x n)"
  proof -
    fix t x y n 
    show " init' (t 0) \<Longrightarrow> \<forall>n. (\<forall>i<n. r (s x y t i, x i) (s x y t (Suc i), y i)) \<longrightarrow> p (s x y t n, x n) 
      \<Longrightarrow> (\<forall>i<n. r' (t i, x i) (t (Suc i), y i)) \<Longrightarrow> (\<forall>i < n. d (t (Suc i)) (s x y t (Suc i)) \<and> r (s x y t i, x i) (s x y t (Suc i), y i))  \<and> p' (t n, x n)"
  proof (induction n)
    case 0
    then show ?case
      apply simp
      apply (drule_tac x = 0 in spec, simp)
      apply (rule_tac s = "s x y t 0" in C, simp_all)
      apply (simp add: s_def)
      using B  by (rule someI2_ex, simp_all)
  next
    case (Suc n)
      
    have [simp]: "d (t n) (s_r d init r x y t n)"
      apply (case_tac n)
      using Suc.prems(1) X apply auto[1]
      using Suc.IH Suc.prems(1) Suc.prems(2) Suc.prems(3) s_def less_SucI by blast
        
    have [simp]: "p (s_r d init r x y t n, x n)"
      using Suc(3) apply (drule_tac x = n in spec)
      apply safe
      using Suc.IH Suc.prems(1) Suc.prems(2) Suc.prems(3) less_SucI apply blast
      by (simp add: s_def)
      
    have "\<And> i . i < Suc n \<Longrightarrow> d (t (Suc i)) (s x y t (Suc i)) \<and> r (s x y t i, x i) (s x y t (Suc i), y i)"
      apply (case_tac "i < n")
      using Suc.IH Suc.prems(1) Suc.prems(2) Suc.prems(3) less_SucI apply blast
      apply (case_tac "i = n", simp_all)
      apply (simp add: s_def)
      apply (rule someI2_ex, simp_all)
      apply (rule_tac t = "t n" in D, simp_all)
      by (simp add: Suc.prems(3))
    from this show ?case
      apply simp
      apply (rule_tac s = "s x y t (Suc n)" in  C, simp)
      using Suc(3) by simp
  qed
qed
  have [simp]: "\<And> x .prec_pre_sts init p r x \<Longrightarrow> prec_pre_sts init' p' r' x"
  proof (unfold prec_pre_sts_simp, safe)
    fix xa u y n
    assume [simp]: "\<forall>u y. init (u 0) \<longrightarrow> (\<forall>n. (\<forall>i<n. r (u i, xa i) (u (Suc i), y i)) \<longrightarrow> p (u n, xa n))"
    assume [simp]: "\<forall>i<n. r' (u i, xa i) (u (Suc i), y i)"
    assume [simp]: "init' (u 0)"
    have "(\<forall>i< n. d (u (Suc i)) (s xa y u (Suc i)) \<and> r (s xa y u i, xa i) (s xa y u (Suc i), y i)) \<and> p' (u ( n), xa ( n))"
      by (rule Aux, simp_all)
    from this show " p' (u n, xa n)"
      by simp
  qed
    
  have [simp]: "\<And> x y . prec_pre_sts init p r x \<Longrightarrow> rel_pre_sts init' r' x y \<Longrightarrow> rel_pre_sts init r x y"
  proof (simp add: prec_pre_sts_simp rel_pre_sts_simp, safe)
    fix x y u
    assume [simp]: "\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n)) "
    assume [simp]: "init' (u (0::nat))"
    assume [simp]: "\<forall>i. r' (u i, x i) (u (Suc i), y i)"

    have "\<And> n . (\<forall>i< n. d (u (Suc i)) (s x y u (Suc i)) \<and> r (s x y u i, x i) (s x y u (Suc i), y i)) \<and> p' (u ( n), x ( n))"
      by (rule Aux, simp_all)
    from this show "\<exists>u. init (u 0) \<and> (\<forall>i. r (u i, x i) (u (Suc i), y i))"
      by (rule_tac x = "s x y u" in exI, simp, blast)
  qed
      
  show "{. prec_pre_sts init p r .} \<circ> [: rel_pre_sts init r :] \<le> {. prec_pre_sts init' p' r' .} \<circ> [: rel_pre_sts init' r' :]"
    by (unfold assert_demonic_refinement, safe, simp_all)
qed

subsection{*Reachability and Refinement*}

definition "reach init r n x y s = (init (s 0) \<and> (\<forall> i < n . r (s i, x i) (s (Suc i), y i)))"

lemma reach_prec_always: "reach init r n x y s \<Longrightarrow> p \<le> inpt r \<Longrightarrow> prec_pre_sts init p r x 
    \<Longrightarrow> \<exists> s' y' . init (s' 0) \<and> (\<forall> i < n . y' i = y i) \<and> (\<forall> i \<le> n . s' i = s i) \<and> (\<box> lift_rel r) (s', x) (s'[1..], y')"
proof -
  assume "reach init r n x y s"
  from this have [simp]: "init (s 0)" and A: "\<And> i . i < n \<Longrightarrow> r (s i, x i) (s (Suc i), y i)"
    by (simp add: reach_def)+
      
  assume "prec_pre_sts init p r x"
  from this have C: "\<And> s n. init (s 0) \<Longrightarrow> (\<exists> y . (\<forall>i<n. r (s i, x i) (s (Suc i), y i))) \<Longrightarrow> p (s n, x n)"
    by (simp add: prec_pre_sts_simp , blast)
      
  assume "p \<le> inpt r"
  from this have B: "\<And> s x . p (s, x) \<Longrightarrow> (\<exists>  sy . r (s,x) (sy))"
    by (simp add: inpt_def le_fun_def)
      
  define nxt where "nxt = (\<lambda> s x  . Eps (r (s, x)))"
    
  define sa where "sa = (\<lambda> i . (if i < n then s i else uu r (s n) (x[n..]) (i - n)))"
  define ya where "ya = (\<lambda> i . (if i < n then y i else yy r (s n) (x[n..]) (i - n)))"
    
  have D: "\<And> n . \<forall> i < n . r (sa i, x i) (sa (Suc i), ya i)"
  proof -
    fix m
    show " \<forall> i < m . r (sa i, x i) (sa (Suc i), ya i)"
    proof (induction m)
      case 0
      then show ?case by simp
    next
      case (Suc m)
      show ?case
        apply safe
        apply (case_tac "i < m")
        using Suc apply (simp)
        apply (case_tac "i = m", simp_all)
        apply (case_tac "m < n", simp add: sa_def ya_def,  safe)
        using A apply blast
         apply (simp add: uu_def)
         apply (case_tac "n = Suc m", simp_all)
        using A apply blast
          apply (case_tac "m = n", simp)
         apply (simp add: sa_def ya_def uu_def yy_def)
         apply (rule_tac someI_ex)
         apply (rule B)
         apply (rule C, simp_all)
         apply (rule_tac x = y in exI)
        using A apply blast
          apply (case_tac "n < m", simp_all)
        apply (simp add: sa_def ya_def uu_def yy_def split: nat.split, safe)
        apply (case_tac "x2a = Suc x2", simp_all)
          apply (case_tac "(Suc (n + x2)) = m", simp_all)
         apply (rule_tac someI_ex)
        apply (rule B)
        apply (cut_tac s = sa and n = m in C)
          apply  (simp add: sa_def uu_def)
        using Suc apply blast
          by (simp add: sa_def uu_def)
      qed
    qed
      
  show "\<exists> s' y' . init (s' 0) \<and> (\<forall> i < n . y' i = y i) \<and>(\<forall> i \<le> n . s' i = s i) \<and> (\<box> lift_rel r) (s', x) (s'[1..], y')"
    apply (rule_tac x = "sa" in exI)
    apply (rule_tac x = "ya" in exI)
    apply (simp_all add: always_def lift_rel_def at_fun_def at_prod_def, safe)
       apply (simp add: sa_def uu_def)
      apply (simp add: ya_def)
     apply (simp add: sa_def uu_def)
      by (cut_tac n = "Suc f" in D, blast)
qed

 
lemma refinemen_reachable_B:
  assumes R: "LocalSystem init p r \<le> LocalSystem init' p' r'"
    and  [simp]: "p' \<le> inpt r'"
  shows "prec_pre_sts init p r x \<Longrightarrow> reach init' r' n x y t \<Longrightarrow> \<exists> s . reach init r n x y s"
    and "prec_pre_sts init p r x \<Longrightarrow> reach init' r' n x y t \<Longrightarrow> p' (t n, x n)"
proof -
  assume prec: "prec_pre_sts init p r x"
  from R and prec have A: "prec_pre_sts init' p' r' x"
    by (unfold LocalSystem_simp assert_demonic_refinement, blast)

  from prec and R have B: "\<And> y . rel_pre_sts init' r' x y \<Longrightarrow> rel_pre_sts init r x y"
    by (unfold LocalSystem_simp assert_demonic_refinement, blast)
      
  assume [simp]: "reach init' r' n x y t"

  from A have "\<exists> s' y' . init' (s' 0) \<and> (\<forall> i < n . y' i = y i) \<and> (\<forall> i \<le> n . s' i = t i) \<and> (\<box> lift_rel r') (s', x) (s'[1..], y')"
    by (rule_tac p = p' in reach_prec_always, simp_all)
      
  from this obtain ta ya where I[simp]: "init' (ta 0)" and X: " (\<And> i . i \<le> n \<Longrightarrow> ta i = t i)" and T: "\<And> i . i < n \<Longrightarrow> ya i = y i" and Y: "(\<box> lift_rel r') (ta, x) (ta[1..], ya)"
    by blast
      
  have "rel_pre_sts init' r' x ya"
    using Y I by (unfold rel_pre_sts_def, blast)

  from this B have "rel_pre_sts init r x ya" 
    by blast
      
  from this obtain sa where "init (sa 0) \<and> (\<box> lift_rel r) (sa, x) (sa[1..], ya)"
    by (simp add: rel_pre_sts_def, blast)
      
  from this have "reach init r n x y sa"
    by (simp add: reach_def always_def at_fun_def lift_rel_def at_prod_def  T [symmetric])
      
  from this show "\<exists> s . reach init r n x y s"
    by blast
      
  from A have Z: "((\<forall>i<n. r' (ta i, x i) (ta (Suc i), ya i)) \<Longrightarrow> p' (ta n, x n))"
    by (simp add: prec_pre_sts_simp)
      
  have [simp]: "\<And> i . r' (ta i, x i) (ta (Suc i), ya i)"
    using Y by (simp add: always_def at_fun_def lift_rel_def at_prod_def)
      
  show "p' (t n, x n)"
    by (simp add: X [symmetric], rule Z, simp)
qed

  
lemma sel_inf_a: "finite X \<Longrightarrow> (\<And> i :: nat . f i \<in> X) \<Longrightarrow> (\<exists> x \<in> X . infinite {i . f i = x})"
proof -
  assume A: "finite X"
  assume "(\<And> i :: nat . f i \<in> X)"
  from this have "(\<Union> x \<in> X . {i . f i = x}) = UNIV"
    by blast
  from this have "infinite  (\<Union> x \<in> X . {i . f i = x})"
    by simp
  from this show "\<exists> x\<in> X . infinite {i . f i = x}"
    using A by auto
qed
  
lemma "X \<noteq> {} \<Longrightarrow> \<exists> (x::'a::wellorder) \<in> X . \<forall> y \<in> X . x \<le> y"
proof -
  assume "X \<noteq> {}"
  then obtain aa :: "'a set \<Rightarrow> 'a" where
    "aa X \<in> X"
    by (metis equals0I)
  then show ?thesis
    by (meson wellorder_Least_lemma(1) wellorder_Least_lemma(2))
qed
  
primrec min_rest :: "nat set \<Rightarrow> nat \<Rightarrow> nat" where
  "min_rest X 0 =  (LEAST x . x \<in> X)" |
  "min_rest X (Suc n) = min_rest (X  - {LEAST x . x \<in> X}) n"
 

lemma sel_inf_fun: "\<And> X . infinite X \<Longrightarrow> min_rest X n \<in> X \<and> min_rest X n < min_rest X (Suc n)"
proof (induction n)
  case 0
  have "\<And> x X . (x::nat) \<in> X \<Longrightarrow>  (LEAST x. x \<in> X) \<le> x"
    by (simp add: Least_le)

  from this have A: "\<And> x X . (x::nat) \<in> X \<Longrightarrow> x \<noteq> (LEAST x. x \<in> X) \<Longrightarrow>  (LEAST x. x \<in> X) < x"
    by (simp add: dual_order.strict_iff_order)
      
  have B: "infinite (X - {LEAST x. x \<in> X})"
    using 0 by (rule infinite_remove)
      
  from this obtain x where C: "x \<in> (X - {LEAST x. x \<in> X})"
    by (cut_tac S  = "(X - {LEAST x. x \<in> X})" in infinite_imp_nonempty, auto)
      
  have D: " (LEAST x. x \<in> X - {LEAST x. x \<in> X}) \<in> (X - {LEAST x. x \<in> X})"
    apply (rule_tac k = x in LeastI)
    using C by simp
    
      
  from 0 show ?case 
    apply simp
    apply safe
     apply (meson LeastI bounded_nat_set_is_finite)
     using D by (rule_tac A, simp_all)
      
next
  case (Suc n)
  from this have U: "\<And> X .  infinite X \<Longrightarrow> min_rest X n \<in> X"
    and V: "\<And> X . infinite X \<Longrightarrow>  min_rest X n < min_rest X (Suc n)"
    by simp_all

  have A: "min_rest X (Suc n) \<in> X"
    apply simp
    apply (cut_tac X = "(X - {LEAST x. x \<in> X})" in U)
    using Suc(2) by simp_all
    
  have B: "min_rest X (Suc n) < min_rest X (Suc (Suc n))"
    apply simp
    apply (cut_tac X = "(X - {LEAST x. x \<in> X})" in V)
    using Suc(2) apply simp
      by simp
  from A and B show ?case
    by simp
qed
  
 
lemma sel_inf: "finite X \<Longrightarrow> (\<And> i :: nat . f i \<in> X) \<Longrightarrow> (\<exists> g x . x \<in> X \<and> (\<forall> i . f (g i) = x) \<and> (\<forall> i . g i < g (Suc i)))"
  apply (drule_tac f = f in  sel_inf_a, simp_all)
  apply safe
  apply (subst ex_comm)
  apply (rule_tac x = x in exI, simp)
  apply (rule_tac x = "min_rest  {i. f i = x}" in exI)
  apply simp
    apply safe
    apply (cut_tac X = "{i. f i = x}" and n = i in sel_inf_fun, simp_all)
    by (cut_tac X = "{i. f i = x}" and n = i in sel_inf_fun, simp_all)

  
definition "sel_inf f X = (SOME g . \<exists> x . x \<in> X \<and> (\<forall> i . f (g i) = x) \<and> (\<forall> i . g i < g (Suc i)))"
  
lemma sel_inf_prop_aux: "finite X \<Longrightarrow> (\<And> i :: nat . f i \<in> X) \<Longrightarrow>  (\<exists> x . x \<in> X \<and> (\<forall> i . f (sel_inf f X i) = x) \<and> (\<forall> i . sel_inf f X i < sel_inf f X (Suc i)))"
  apply (simp add: sel_inf_def)
  apply (rule_tac P = " (\<lambda> g . \<exists> x . x \<in> X \<and> (\<forall> i . f (g i) = x) \<and> (\<forall> i . g i < g (Suc i)))" in someI_ex)
  by (rule sel_inf, simp_all)

lemma sel_inf_prop: 
  assumes A: "finite X" and B: "(\<And> i :: nat . f i \<in> X)" 
  shows "f (sel_inf f X i) = f (sel_inf f X 0) " and "\<And> i . sel_inf f X i < sel_inf f X (Suc i)"
    and "i \<le> sel_inf f X i"
proof -
  from A and B have C: "(\<exists> x . x \<in> X \<and> (\<forall> i . f (sel_inf f X i) = x) \<and> (\<forall> i . sel_inf f X i < sel_inf f X (Suc i)))"
    by (simp add: sel_inf_prop_aux)
      
  from this show "f (sel_inf f X i) = f (sel_inf f X 0)"
    by (safe, frule_tac x = 0 in spec, simp)
      
  from C show D: "\<And> i . sel_inf f X i < sel_inf f X (Suc i)"
    by blast
      
  show "i \<le>  sel_inf f X i"
  proof (induction i)
    case 0
    then show ?case by simp
  next
    case (Suc i)
    then show ?case
      by (metis D Suc_leI le_less less_trans)
  qed
    
qed
  
    
fun SSa :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'b \<Rightarrow> 'a \<times> 'c \<Rightarrow> bool) \<Rightarrow> (nat \<Rightarrow> 'b) \<Rightarrow> (nat \<Rightarrow> nat \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a" where
  "SSa init r x s 0  = (s[Suc 0..] o sel_inf (\<lambda> i . s (Suc i) 0)  {s . init s})" |
  "SSa init r x s (Suc n)  = ((SSa init r x s n[Suc 0..]) o 
      sel_inf (\<lambda> i . SSa init r x s n (Suc i) (Suc n))  {s' . \<exists> y . r ((SSa init r x s n[Suc 0..]) 0 n, x n) (s', y) })"


lemma refinemen_reachable_aux:
  assumes finite_next: "\<And> s x . finite {s' . \<exists> y . r (s, x) (s', y)}"
    and finite_init[simp]: "finite {s . init s}"
   assumes  A: "(\<And> n . reach init r (Suc n) x y (s n))" 
   shows "(\<forall> j . \<forall> k \<le> n . SSa init r x s n j k = SSa init r x s n 0 k) \<and> reach init r n x y ( SSa init r x s n n) 
      \<and> (\<exists> k . \<forall> i . k i \<ge> n \<and> SSa init r x s n i = s (k i) \<and> k i < k (Suc i))
      \<and> (\<forall> j . \<forall> k \<le> n . SSa init r x s (Suc n) j k = SSa init r x s n 0 k)"
    proof (induction n)
      case 0
      show " (\<forall>j k. k \<le> 0 \<longrightarrow> SSa init r x s 0 j k = SSa init r x s 0 0 k) \<and>
          reach init r 0 x y (SSa init r x s 0 0) \<and> (\<exists>k. \<forall>i. 0 \<le> k i 
        \<and> SSa init r x s 0 i = s (k i) \<and> k i < k (Suc i))
        \<and> (\<forall>j k. k \<le> 0 \<longrightarrow> SSa init r x s (Suc 0) j k = SSa init r x s 0 0 k)"
        apply simp
        apply safe
         apply (rule sel_inf_prop, simp)
        using A apply (simp_all add: reach_def)
        apply (rule_tac x = "\<lambda> i . (Suc (sel_inf (\<lambda>i. s (Suc i) 0) (Collect init) i))" in exI, clarsimp)
         apply (rule sel_inf_prop, simp_all)
        apply (rule sel_inf_prop(1))
          by simp_all
    next
      case (Suc n)
      from this have X: "reach init r n x y (SSa init r x s n n)"
        by simp
      have T: "\<And> i k . k \<le> n \<Longrightarrow> SSa init r x s n i k = (SSa init r x s n (Suc 0) k)"
        by (metis Suc.IH)

      have "(\<exists>k. \<forall>i. n \<le> k i \<and> SSa init r x s n i = s (k i) \<and> k i < k (Suc i))"
        using Suc by simp
          
      from this obtain k where VV[simp]: "\<And> i . n \<le> k i" and UU: "\<And> i . SSa init r x s n i = s (k i)" and XX: "\<And> i . k i < k (Suc i)"
        by blast

      have "\<And> i . r  (SSa init r x s n (Suc i) n, x n) (SSa init r x s n (Suc i) (Suc n), y n)"
        apply (simp add: UU)
        apply (cut_tac n = "k (Suc i)" in A)
        apply (simp add: reach_def, safe)
        apply (drule_tac x = n in spec, safe)
          using VV by (simp add: le_imp_less_Suc)
          
      from this have Y[simp]: "\<And> i . r (SSa init r x s n (Suc 0) n, x n) (SSa init r x s n (Suc i) (Suc n), y n)"
        by (cut_tac i = "Suc i" and k = n in T[symmetric], simp_all)
        
          
      have [simp]:"finite {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)}"
        using finite_next by blast
          
      have U: "\<And> j k . k \<le> Suc n \<Longrightarrow> SSa init r x s (Suc n) j k = SSa init r x s (Suc n) 0 k"
        apply simp
        apply (case_tac "k = Suc n", simp_all)
         apply (rule sel_inf_prop(1), simp_all)
        using Y apply blast
        apply (case_tac "k \<le> n", simp_all)
          apply (subst (3) T, simp_all)
        by (subst (4) T, simp_all)
          
      have Z: "SSa init r x s n (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} (Suc n))) n
        = SSa init r x s n (Suc 0) n"
        by (subst (3) T, simp_all)

      have V: "reach init r (Suc n) x y (SSa init r x s (Suc n) (Suc n))"
        apply simp
        apply (simp add: reach_def, safe)
        using X apply (simp add: reach_def)
         apply (subst (3) T, simp_all, clarify)
         apply (subst (asm) T, simp_all)
        apply (case_tac "i = n", simp_all)
         apply (simp add: Z)
        apply (case_tac "i < n", simp_all)
          apply (subst (3) T, simp_all)
        apply (subst (4) T, simp_all)
        using X apply (simp add: reach_def, clarify)
        apply (drule_tac x = i in spec, simp)
         apply (subst (asm) (2) T, simp_all)
        by (subst (asm) (3) T, simp_all)
          
      have mono_k: "\<And> i j . i < j \<Longrightarrow> k i < k j"
      proof -
        fix i j
        have Az: "i < j \<longrightarrow> k i < k j"
        proof (induction j)
          case 0
          then show ?case 
            by simp
        next
          case (Suc j)
          show ?case
            apply(case_tac "i = j", simp_all)
             apply (simp add: XX)
            apply (safe)
            apply (case_tac "i < j", simp_all)
              using Suc and XX apply simp
              using less_trans by auto
          qed
          assume "i < j"
          from this and Az show "k i < k j"
            by simp
        qed
          
          
      have W: "(\<exists>k. \<forall>i. Suc n \<le> k i \<and> SSa init r x s (Suc n) i = s (k i) \<and> k i < k (Suc i))"
        apply simp
        apply (rule_tac x = "\<lambda> i . k (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} i)) " in exI, safe)
        prefer 2
          apply (simp add: UU)
         apply (rule_tac y = "Suc ( k (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} i))" in order_trans)
          apply simp_all
         apply (simp add: Suc_leI \<open>\<And>i. k i < k (Suc i)\<close>)
        apply (rule mono_k, simp)
        apply (rule sel_inf_prop(2))
         apply simp_all
          using Y by blast
          define ZZ where "ZZ = (\<lambda>j . (Suc (sel_inf (\<lambda>i. SSa init r x s n
                                        (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} (Suc i)))
                                        (Suc (Suc n)))
                          {s'. \<exists>y. r (SSa init r x s n
                                       (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} (Suc 0))) (Suc n),
                                      x (Suc n))
                                    (s', y)}
                          j)))"
            
            from this have ZZ_simp: "\<And> j . ZZ j = (Suc (sel_inf (\<lambda>i. SSa init r x s n
                                        (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} (Suc i)))
                                        (Suc (Suc n)))
                          {s'. \<exists>y. r (SSa init r x s n
                                       (Suc (sel_inf (\<lambda>i. SSa init r x s n (Suc i) (Suc n)) {s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)} (Suc 0))) (Suc n),
                                      x (Suc n))
                                    (s', y)}
                          j))"
              by simp
            
        have TT: "\<And> j k . k \<le> Suc n \<Longrightarrow> SSa init r x s (Suc (Suc n)) j k = SSa init r x s (Suc n) 0 k"
          apply simp
          apply (simp add: ZZ_simp[symmetric])
            apply (case_tac "k = Suc n")
          apply (cut_tac f = "(\<lambda>i. SSa init r x s n (Suc i) (Suc n))"
                and X = "{s'. \<exists>y. r (SSa init r x s n (Suc 0) n, x n) (s', y)}" and i = "ZZ j" in  sel_inf_prop(1))
             apply simp_all
          using Y apply blast
          apply (case_tac "k \<le> n", simp_all)
            apply (subst (3) T, simp_all)
            by (subst (4) T, simp_all)
 
            
      from U and V and W and TT show "(\<forall>j k. k \<le> Suc n \<longrightarrow> SSa init r x s (Suc n) j k = SSa init r x s (Suc n) 0 k) \<and>
         reach init r (Suc n) x y (SSa init r x s (Suc n) (Suc n)) \<and>
         (\<exists>k. \<forall>i. Suc n \<le> k i \<and> SSa init r x s (Suc n) i = s (k i) \<and> k i < k (Suc i)) \<and>
         (\<forall>j k. k \<le> Suc n \<longrightarrow> SSa init r x s (Suc (Suc n)) j k = SSa init r x s (Suc n) 0 k)"
        by blast
    qed
  

lemma refinemen_reachable_A:
  assumes finite_next: "\<And> s x . finite {s' . \<exists> y . r (s, x) (s', y)}"
    and finite_init: "finite {s . init s}"
  assumes A: "\<And> n x y t . prec_pre_sts init p r x \<Longrightarrow> reach init' r' n x y t \<Longrightarrow> p' (t n, x n)"
    and B: "\<And> n x y t . prec_pre_sts init p r x \<Longrightarrow> reach init' r' n x y t \<Longrightarrow> \<exists> s . reach init r n x y s"
  shows "LocalSystem init p r \<le> LocalSystem init' p' r'"
proof (simp add: LocalSystem_simp assert_demonic_refinement, simp add: le_fun_def, safe)
  fix x y
  assume [simp]: "prec_pre_sts init p r x"


  show "prec_pre_sts init' p' r' x"
    apply (simp add: prec_pre_sts_simp, safe)
    apply (rule_tac y = y in A, simp)
    by (simp add: reach_def)
      
  assume [simp]: "rel_pre_sts init' r' x y"
   
    
  from this obtain t where "\<And> n . reach  init' r' n x y t"
    by (simp add: rel_pre_sts_simp reach_def, blast)
      
  from this have "\<And> n . \<exists> s . reach init r n x y s"
    by (rule_tac t = t in B, simp_all)

  from this have "\<exists> s . \<forall> n . reach init r n x y (s n)"
    by (subst choice_iff [symmetric], blast)
      
  from this obtain s where X: "\<And> n . reach init r n x y (s n)"
    by blast
      
  define ss where "ss = s[Suc 0..]"
    
  from X have Y: "\<And> n . reach init r (Suc n) x y (ss n)"
    by (simp add: ss_def)
        
  have U: "\<And> n . reach init r n x y (SSa init r x ss n n) " and
     V: "\<And> n j k. k \<le> n \<Longrightarrow> SSa init r x ss n j k = SSa init r x ss n 0 k" and
     W: "\<And> n j k . k \<le> n \<Longrightarrow> SSa init r x ss (Suc n) j k = SSa init r x ss n 0 k"
    apply (cut_tac s = ss and init = init and r = r and x= x and y = y in refinemen_reachable_aux)
    using finite_next apply (simp)
    using finite_init apply (simp)
    using Y apply simp
    apply blast
    apply (cut_tac s = ss and init = init and r = r and x= x and y = y in refinemen_reachable_aux)
    using finite_next apply (simp)
    using finite_init apply (simp)
    using Y apply simp
    apply blast
    apply (cut_tac s = ss and init = init and r = r and x= x and y = y and n = n in refinemen_reachable_aux)
    using finite_next apply (simp)
    using finite_init apply (simp)
    using Y apply simp
    by blast
      
  define sa where "sa = (\<lambda> i . SSa init r x ss (Suc i) (Suc i) i)"
  have [simp]: "\<And> i . r (sa i, x i) (sa (Suc i), y i)"
    apply (simp add: sa_def del: SSa.simps)
    apply (subst (2) W, simp)
    apply (cut_tac n = "Suc i" in U)
    apply (simp add: reach_def del: SSa.simps, safe)
      apply (drule_tac x = i in spec)
    apply (simp del: SSa.simps)
    by (subst (asm) (3) V, simp_all)
      
  have [simp]: "init (sa 0)"
    apply (simp add: sa_def del: SSa.simps)
    apply (cut_tac n = "Suc 0" in U)
    by (simp add: reach_def del: SSa.simps)
      
      
  show "rel_pre_sts init r x y"
    apply (simp add: rel_pre_sts_simp)
    by (rule_tac x = sa in exI, simp)
qed
 
definition "symb_sts_refin init p r init' p' r'
    = 
      ((\<forall> n x y t . prec_pre_sts init p r x \<longrightarrow> reach init' r' n x y t \<longrightarrow> p' (t n, x n))
      \<and> (\<forall> n x y t . prec_pre_sts init p r x \<longrightarrow> reach init' r' n x y t \<longrightarrow> (\<exists> s . reach init r n x y s)))"
  

lemma refinemen_reachable_iff:
  assumes finite_next[simp]: "\<And> s x . finite {s' . \<exists> y . r (s, x) (s', y)}"
    and finite_init[simp]: "finite {s . init s}"
    and [simp]: "p' \<le> inpt r'"
  shows "LocalSystem init p r \<le> LocalSystem init' p' r' = symb_sts_refin init p r init' p' r'"
proof
  assume [simp]: "LocalSystem init p r \<le> LocalSystem init' p' r'"
  show "symb_sts_refin init p r init' p' r'"
    apply (simp add: symb_sts_refin_def)
      apply safe
    apply (cut_tac p = p and p' = p' and init = init and init' = init' and r = r and r' = r' in refinemen_reachable_B(2), simp_all)
    by (cut_tac p = p and p' = p' and init = init and init' = init' and r = r and r' = r' in refinemen_reachable_B(1), simp_all)
next
  assume A: "symb_sts_refin init p r init' p' r'"
  show "LocalSystem init p r \<le> LocalSystem init' p' r'"
    apply (rule refinemen_reachable_A, simp_all)
    using A apply (simp add: symb_sts_refin_def)
    using A by (simp add: symb_sts_refin_def, blast)
qed
  
definition "inv_top n P = (\<forall> u v . eqtop n u v \<longrightarrow> (P u = P v))"
definition "prec_pre_sts_bound init p r N x = ((\<forall>u. init (u 0) \<longrightarrow> (\<forall>y . \<forall> n < N. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))))"
  
lemma replace_variables: "(inv_top (Suc N) (P N)) \<Longrightarrow> (inv_top N (R N)) \<Longrightarrow> (inv_top N (Q' N)) \<Longrightarrow> (\<forall> (x::nat\<Rightarrow>'z) . P N x \<and> (ZZ (Q' N x) (Q N (x[N..]))) \<and> R N x \<longrightarrow> S N (x N))
    =  (\<forall> x xN y . P N (x(N := xN)) \<and> y 0 = xN \<and> (ZZ (Q' N x) ( Q N (y))) \<and> R N x \<longrightarrow> S N (xN))"
   proof safe
     fix x xN y
     assume A: "\<forall>x. P N x \<and> ZZ (Q' N x) (Q N (x[N ..])) \<and> R N x \<longrightarrow> S N (x N)"
     assume B: " inv_top (Suc N) (P (N::nat))"
     assume C: " inv_top N (R N)"
     assume Ca: "inv_top N (Q' N)"
     assume D: "P N (x(N := y (0::nat)))"
     assume E: "ZZ (Q' N x) (Q N y)"
     assume F: "R N x"
       
     define xx where "xx = (\<lambda> i . if i < N then x i else y (i -N))"
     from A have  U: "P N xx \<and> ZZ (Q' N xx) (Q N (xx[N ..])) \<and> R N xx \<longrightarrow> S N (xx N)"
       by blast
     have [simp]: "P N xx"
       apply (cut_tac B, simp add: inv_top_def)
       apply (drule_tac x =  xx in spec)
       apply (drule_tac x =  "(x(N := y (0::nat)))" in spec)
       using D apply simp
       apply safe
       by (simp add: xx_def)
         
     from E have [simp]: "ZZ (Q' N xx) (Q N (xx[N ..]))"
       apply (subgoal_tac "xx[N..] = y")
        apply simp
       apply (cut_tac Ca, simp add: inv_top_def)
       apply (drule_tac x =  xx in spec)
       apply (drule_tac x =  "x" in spec)
       by (simp_all add: xx_def)
         
       
     have [simp]: "R N xx"
       apply (cut_tac C, simp add: inv_top_def)
       apply (drule_tac x =  xx in spec)
       apply (drule_tac x =  "x" in spec)
       using F apply simp
       apply safe
       by (simp add: xx_def)
       
       
     from U show "S N (y 0)"
       apply (simp)
       by (simp add: xx_def)
   next
     fix x::"nat \<Rightarrow> 'z"
     assume A: "\<forall>x xN y. P N (x(N := xN)) \<and> y 0 = xN \<and> ZZ (Q' N x) (Q N y) \<and> R N x \<longrightarrow> S N xN"
     assume "inv_top (Suc N) (P N)"
     assume "inv_top N (R N)"
     assume "inv_top N (Q' N)"
       
     assume [simp]: " P N x"
     assume [simp]: "ZZ (Q' N x) (Q N (x[N ..]))"
     assume [simp]: "R N x"
       
     from A show "S N (x N)"
       apply (drule_tac x = x in spec)
       apply (drule_tac x = "x N" in spec)
       apply (drule_tac x = "x[N..]" in spec)
       by simp
         
   qed
    
lemma prec_pre_sts_reach: "\<And> x . prec_pre_sts init p r x = (\<forall> s n. (\<exists> y . reach init r n x y s) \<longrightarrow> p (s n, x n))"
  by (simp add: prec_pre_sts_simp reach_def, blast)
       
lemma prec_pre_sts_bound_simp: "\<And> N x . prec_pre_sts_bound init p r N x = 
      (\<forall>u n. (n < N \<and> init (u 0) \<and> ((\<exists> y . \<forall>i<n . r (u i, x i) (u (Suc i), y i)))) \<longrightarrow> (\<forall> k \<le> n . p (u k, x k)))"
   proof (unfold  prec_pre_sts_bound_def, safe)
     fix N x u n k y
     assume A: "\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. n < N \<longrightarrow> (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))"
     assume B: "n < N"
     assume C: "init (u (0::nat)) "
     assume D: " \<forall>i<n. r (u i, x i) (u (Suc i), y i)"
     assume E: "k \<le> n"
       
     from D have "(\<forall>i<n. r (u i, x i) (u (Suc i), y i))"
       by blast

     from this and E  have "(\<forall>i < k. r (u i, x i) (u (Suc i), y i))"
       using less_le_trans by blast
         
     from this and  A and C and B and E show "p (u k, x k)"
       by auto
         
   next
     fix N x u y n
     assume A: "\<forall>u n. n < N \<and> init (u 0) \<and> (\<exists>y. \<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> (\<forall>k\<le>n. p (u k, x k))"
     assume B: "init (u (0::nat)) "
     assume C: "n < N"
     assume D: "\<forall>i<n. r (u i, x i) (u (Suc i), y i)"
       
     from A B C D show  "p (u n, x n)"
       by blast
   qed

lemma prec_pre_sts_bound: "\<And> x N . prec_pre_sts init p r x = (prec_pre_sts_bound init p r N x 
    \<and> (\<forall> s y . reach init r N x y s \<longrightarrow>  prec_pre_sts (\<lambda> u . u = s N) p r (x[N..])))"
   proof (simp add: prec_pre_sts_simp prec_pre_sts_bound_def, safe)
     fix x N u y v n
     assume A: "\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))"
     assume C: " init (u (0::nat))"
     assume B: "\<forall>i<n. r (u i, x i) (u (Suc i), y i)"
     from A and C have "(\<forall>y n. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))"
       by blast

     from this and B show "p (u n, x n)"
       by blast
       
   next
     fix x N s y u ya n
     assume A: "\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n)) "
     assume B: "reach init r N x y s"
     assume F: "u (0::nat) = s N"
     assume C: " \<forall>i<n. r (u i, x (N + i)) (u (Suc i), ya i)"
       
         
     from B have E: "init (s 0) \<and> (\<forall> i < N . r (s i, x i) (s (Suc i), y i))"
       by (simp add: reach_def)
         
     define uu where "uu = cat N s u"
       
     define yy where "yy = cat N y ya"
       
     have "init (uu 0)"
       apply (simp add: uu_def, safe)
       using E apply blast
       using F and E by simp
         
     from this and A have G: "((\<forall>i<N + n. r (uu i, x i) (uu (Suc i), yy i)) \<Longrightarrow> p (uu (N + n), x (N + n)))"
       by blast
       
         
     have "(\<forall>i<N + n. r (uu i, x i) (uu (Suc i), yy i))"
       apply (simp add: uu_def yy_def, safe)
       apply (cut_tac E) apply (safe)
         apply (drule_tac x = i in spec, simp)
        apply (case_tac "N = Suc i", simp_all)
         apply (cut_tac E, safe)
        apply (drule_tac x = i in spec, simp)
       using F apply simp
       using C apply (drule_tac x = "i - N" in spec)
       apply safe
        apply simp_all
         by (simp add: Suc_diff_le)
         
         
         
       from this and G have "p (uu (N + n), x (N + n))"
         by simp
         
       from this show "p (u n, x (N + n))"
         by (simp add: uu_def)
           
     next
       fix x N u y n
       assume A: "\<forall>u. init (u 0) \<longrightarrow> (\<forall>y n. n < N \<longrightarrow> (\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<longrightarrow> p (u n, x n))"
       assume B: " \<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> (\<forall>u. u 0 = s N \<longrightarrow> (\<forall>y n. (\<forall>i<n. r (u i, x (N + i)) (u (Suc i), y i)) \<longrightarrow> p (u n, x (N + n)))) "
       assume C: "init (u (0::nat))"
       assume D: "\<forall>i<n. r (u i, x i) (u (Suc i), y i)"
         
       from B have F: "\<And> ua y n . (\<exists>y. reach init r N x y u) \<Longrightarrow> (ua 0 = (u N) \<Longrightarrow> (\<forall>i<n. r (ua i, x (N + i)) (ua(Suc i), y i) \<Longrightarrow> p (ua n, x (N + n))))"
         by blast
         
       from A have E: "\<And> n . n < N \<Longrightarrow> ((\<forall>i<n. r (u i, x i) (u (Suc i), y i)) \<Longrightarrow> p (u n, x n))"
         using C by simp
         
       show "p (u n, x n)"
         proof (cases "n < N")
           case True
           then show ?thesis
             apply (rule_tac E, simp_all)
             using D by simp
               
         next
           case False
           define uu where "uu = u[N..]"
           define yy where "yy = y[N..]"
           from False have H: "N \<le> n"
             by auto
               
           from this have "(\<exists>y. reach init r N x y u)"
             apply (rule_tac x = y in exI)
             using C and D by (simp add: reach_def)

               
           from this and F have U: "(uu 0 = (u N) \<Longrightarrow> (\<forall>i< (n - N). r (uu i, x (N + i)) (uu(Suc i), yy i) \<Longrightarrow> p (uu (n - N), x (N + (n -N)))))"
             by simp
               
           have [simp]: "uu 0 = (u N)"
             by (simp add: uu_def)
           from H have [simp]: "\<forall>i< (n - N). r (uu i, x (N + i)) (uu(Suc i), yy i)"
             using D by (simp add: uu_def yy_def)
               
           from U have " p (uu (n - N), x (N + (n -N)))"
             by simp
             
             
           then show ?thesis
             using H by (simp add: uu_def)
         qed
         
       
       qed

lemma AA: "\<And> t x N y . ((prec_pre_sts init p r x \<and> reach init' r' N x y t) \<longrightarrow> p' (t N, x N))
    = ((prec_pre_sts_bound init p r N x \<and> (\<forall> s y . reach init r N x y s \<longrightarrow>  prec_pre_sts (\<lambda> u . u = s N) p r (x[N..])) \<and> reach init' r' N x y t) \<longrightarrow> p' (t N, x N))"
  by (cut_tac x = x and N = N and p = p and r = r and init = init in prec_pre_sts_bound, simp)

lemma [simp]: "inv_top (Suc N) (prec_pre_sts_bound init p r N)"
  by (simp add: inv_top_def prec_pre_sts_bound_def fun_eq_iff)
    
lemma [simp]: "inv_top N (\<lambda>x. \<exists>y. reach init' r' N x y t)"
  by (simp add: inv_top_def fun_eq_iff reach_def)
    
lemma [simp]: " inv_top N (\<lambda>x s. \<exists>y. reach init r N x y s)"
  by (simp add: inv_top_def fun_eq_iff reach_def)
    
lemma sts_refinement_A_bounded: "(\<forall> x y . (prec_pre_sts init p r x \<and> reach init' r' N x y t) \<longrightarrow> p' (t N, x N))
    = (\<forall> xN . (\<exists> x . prec_pre_sts_bound init p r N (x(N := xN)) 
        \<and> (\<exists> xz . xz 0 = xN \<and> (\<forall> s y . reach init r N x y s \<longrightarrow>  prec_pre_sts (\<lambda> u . u = s N) p r xz)) 
        \<and> (\<exists> y . reach init' r' N x y t)) 
          \<longrightarrow> p' (t N, xN))"
   proof (simp add: AA, safe)
     fix t x y xz
     assume A: "\<forall>x. prec_pre_sts_bound init p r N x \<and> (\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r (x[N ..])) \<and> (\<exists>y. reach init' r' N x y t) \<longrightarrow> p' (t N, x N)"
     assume B: "prec_pre_sts_bound init p r N (x(N := xz (0::nat)))"
     assume C: "\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r xz"
     assume D: "reach init' r' N x y t"
     thm replace_variables
       
     have "(\<forall>x. prec_pre_sts_bound init p r N x \<and> (\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r (x[N ..])) \<and> (\<exists>y. reach init' r' N x y t) \<longrightarrow> p' (t N, x N))
        = ((\<forall>x xN. prec_pre_sts_bound init p r N (x(N := xN)) \<and> (\<exists>y. y 0 = xN \<and> (\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r y) \<and> (\<exists>y. reach init' r' N x y t)) \<longrightarrow> p' (t N, xN)))"
       apply (cut_tac P = "\<lambda> N x . prec_pre_sts_bound init p r N x" and N = N
            and Q = "\<lambda> N y s . prec_pre_sts (\<lambda>u. u = s N) p r (y)" 
            and Q' = "\<lambda> N x s . \<exists>y. reach init r N x y s"
            and ZZ = "(\<lambda> A B . \<forall> s . A s \<longrightarrow> B s)" 
            and R = "\<lambda> N x . (\<exists>y. reach init' r' N x y t)" 
            and S = "\<lambda> N x . p' (t N, x)" in replace_variables)
          by simp_all
         
        from A this have "\<And>x xN. prec_pre_sts_bound init p r N (x(N := xN)) \<Longrightarrow> (\<exists>y. y 0 = xN \<and> (\<forall>s. (\<exists>y. reach init r N x y s) 
            \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r y) \<and> (\<exists>y. reach init' r' N x y t)) \<Longrightarrow> p' (t N, xN)"
          by simp
            
        from this have "prec_pre_sts_bound init p r N (x(N := xz 0)) \<Longrightarrow> (\<exists>y. y 0 = xz 0 \<and> (\<forall>s. (\<exists>y. reach init r N x y s)
            \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r y) \<and> (\<exists>y. reach init' r' N x y t)) \<Longrightarrow> p' (t N, xz 0)"
          by simp
            
        from this have U: " (\<exists>y. y 0 = (xz 0) \<and> (\<forall>s. (\<exists>y. reach init r N x y s) 
            \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r y) \<and> (\<exists>y. reach init' r' N x y t)) \<Longrightarrow> p' (t N, (xz 0))"
          by (simp add: B)

          from C D have "(\<exists>y. y 0 = (xz 0) \<and> (\<forall>s. (\<exists>y. reach init r N x y s) 
            \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r y) \<and> (\<exists>y. reach init' r' N x y t))"
            by (rule_tac x = xz in exI, simp, blast)
              
          from this and U show "p' (t N, (xz 0))"
            by simp
 
   next
     fix t x y
     assume A: "\<forall>xN. (\<exists>x. prec_pre_sts_bound init p r N (x(N := xN)) \<and> 
              (\<exists>xz. xz 0 = xN \<and> (\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r xz)) \<and> (\<exists>y. reach init' r' N x y t)) \<longrightarrow>
                p' (t N, xN)"
     assume B: "prec_pre_sts_bound init p r N x"
     assume C: "\<forall>s. (\<exists>y. reach init r N x y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r (x[N ..])"
     assume D: "reach init' r' N x y t"
         
     from A have  E: "(\<exists>xx. prec_pre_sts_bound init p r N (xx(N := x N)) \<and>
                       (\<exists>xz. xz 0 = x N \<and> (\<forall>s. (\<exists>y. reach init r N xx y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r xz)) \<and> (\<exists>y. reach init' r' N xx y t)) \<Longrightarrow>
                  p' (t N, x N)"
       by blast
           
       have "(\<exists>xx. prec_pre_sts_bound init p r N (xx(N := x N)) \<and>
                       (\<exists>xz. xz 0 = x N \<and> (\<forall>s. (\<exists>y. reach init r N xx y s) \<longrightarrow> prec_pre_sts (\<lambda>u. u = s N) p r xz)) \<and> (\<exists>y. reach init' r' N xx y t))"
         apply (rule_tac x = x in exI, simp, safe)
         using B apply simp
          apply (rule_tac x = "x[N..]" in exI, simp)
         using C apply simp
         using D by blast

       from this show "p' (t N, x N)"
         by (rule E)
     qed

lemma reach_until: "(\<exists> x s y n. reach init r n x y s \<and> s n = t) 
    = (\<exists> sa . init (sa 0) \<and> ((\<lambda> sa . (\<exists> x y . r (sa 0, x) (sa (Suc 0), y))) until (\<lambda> sa . sa 0 = t)) sa)"
  apply (simp add: reach_def until_def at_fun_def, auto)
   apply blast
  apply (simp add: choice_iff', safe)
  by blast

lemma LocalSystem_prec_top: "LocalSystem init \<top> r = [: rel_pre_sts init r:]"
  by (simp add: LocalSystem_simp assert_true_skip skip_comp)


lemma LocalSystem_input_complete: "(LocalSystem init p r = [:rel_pre_sts init r:])
  = ((\<forall> x s . init s \<longrightarrow> p (s,x)) \<and> 
      (\<forall> s s' x x' y n . 
      (\<exists> x y . reach init r n x y s) \<and> p (s n, x) \<and> r (s n,x) (s', y) \<longrightarrow> p (s', x')))"
proof (simp add: LocalSystem_simp assert_demonic_eq_demonic, safe)
  fix s x
  assume [simp]: "init s"
  assume A: "\<forall>x. prec_pre_sts init p r x"
  show "p (s, x)"
    using A apply (drule_tac x = "\<lambda> i . x" in spec)
    apply (simp add: prec_pre_sts_simp)
    apply (drule_tac x = "\<lambda> i . s" in spec)
    apply (simp add: fun_eq_iff)
    by blast

next
  fix s s' x x' y xa ya n
  assume A: " \<forall>x. prec_pre_sts init p r x"
  assume "p (s (n::nat), x)"
  assume W: "r (s n, x) (s', y)"
  assume "reach init r n xa ya s"
  from this have [simp]: "init (s 0) " and V: " \<And> i . i < n \<Longrightarrow> r (s i, xa i) (s (Suc i), ya i)"
    by (simp_all add: reach_def)
  define xb where "xb = (\<lambda> i . if i < n then xa i else (if i = n then x else x'))"
  define yb where "yb = (\<lambda> i . if i < n then ya i else y)"
  define sb where "sb = (\<lambda> i . if i \<le> n then s i else s')"
    
  have [simp]: "init (sb 0)"
    by (simp add: sb_def)
      
  from A have "prec_pre_sts init p r xb"
    by simp

  from this have X: "(\<forall>i < (Suc n). r (sb i, xb i) (sb (Suc i), yb i)) \<Longrightarrow> p (sb (Suc n), xb (Suc n)) "
    by (simp add: prec_pre_sts_simp)
    
  have [simp]: "\<And> i . i < (Suc n) \<Longrightarrow> r (sb i, xb i) (sb (Suc i), yb i)"
    apply (simp add: sb_def xb_def yb_def)
    using V W less_antisym by blast
      
  from X show "p (s', x')"
    apply simp
    by (simp add: sb_def xb_def)

next
  fix x 
  assume A: "\<forall>x s. init s \<longrightarrow> p (s, x)"
  assume B: "\<forall>s s'. (\<exists>x y n. (\<exists>x y. reach init r n x y s) \<and> p (s n, x) \<and> r (s n, x) (s', y)) \<longrightarrow> (\<forall>x'. p (s', x'))"
  show " prec_pre_sts init p r x"
  proof (simp add: prec_pre_sts_simp, safe)
    fix u y n
    assume C: "init (u (0::nat))"
    from A and C have [simp]: "\<And> x . p (u 0, x)"
      by simp
    show "\<forall>i<n. r (u i, x i) (u (Suc i), y i) \<Longrightarrow> p (u n, x n)"
      proof (induction n)
        case 0
        then show ?case by simp
      next
        case (Suc n)
        from this have U: "p (u n, x n)"
          by auto

        from Suc(2) have V: "r (u n, x n) (u (Suc n), y n)"
          by blast

        have W: "(\<exists>x y . reach init r n x y u)"
          apply (rule_tac x = x in exI)
          apply (rule_tac x = y in exI)
          apply (simp add: reach_def)
          using Suc(2) and C by simp
            
        from B have D: "(\<exists>x y . reach init r n x y u) \<Longrightarrow> (p (u n, x n) \<and> (r (u n, x n) (u (Suc n), y n))) \<Longrightarrow>  p (u (Suc n), x (Suc n))"
          by blast

        from U V W D show ?case
          by simp
      qed
  qed
qed

end
  
